
 
PROGRAM SignWriter (input, output);




	(*--------------------------------------*)
	(*				        *)
	(*			        	*)
	(*	    SignWriter 4.3	        *)
	(*	    for the IBM PC	        *)
	(*				        *)
	(*				        *)
	(*      Author: Richard Gleaves      	*)
        (*            				*)
        (*                                    	*)  
	(*     	Date:  7 September 1995        	*)
	(*				        *)
	(*			        	*)
	(*         Copyright 1995 by            *)
	(*			        	*)
	(*	 The Center for Sutton      	*)
	(*	   Movement Writing	        *)
	(*				        *)
	(*			        	*)
	(*--------------------------------------*)









 (* ---------------------  File structures  ------------------ *)


       
CONST  SWversion = 5;	       (* SignWriter file version    *)
       SWhost    = 3;	       (* host machine code = IBM PC *)
       SWdict    = 2;	       (* SW dictionary version      *)	

       
       MsgFile = 'SW.MSG';     (* message file name    *) 
       KeyFile = 'SW.KEY';     (* keymap file name     *)
       SymFile = 'SW.SYM';     (* symbol file name     *)
       DicFile = 'SW.DIC';     (* dictionary file name *)
       DinFile = 'SW.DIN';     (* dict index file name *)
       FileExt = '.SGN';       (* SW file extension    *)
       GrafExt = '.PCX';       (* graphics file ext    *)


       iNoRoom =  8;	       (* insufficient disk space code *)
       iNoFile = 10;	       (* no such file on disk code    *)



TYPE   fName   = LSTRING(8);		  (* file name       *)
       
       blkFile = FILE OF BYTE;  	  (* block file type *)
  





 (* -------------------  File screen structures   ----------------- *)



VAR    FExit:   BOOLEAN;                        (* cmd loop terminator *)

       FScreen: BOOLEAN;			(* file screen mode    *)

       UpdateFileScrn: BOOLEAN;			(* file screen status  *)

       FileName: fName;				(* current file name   *)

       maxName, curName: INTEGER;		(* file list indices   *)

       UpdateFileList: BOOLEAN;			(* file list status    *)







 (* -------------------  Edit screen structures  ------------------- *)


       
CONST  SymYOff = 2;	       (* Y offset of symbol in key box *)

       AlertBoxDelay = 2;      (* alert box display time        *)

       StartDelay    = 30;     (* startup  msg display time     *)		


TYPE   scrInx = INTEGER;		(* screen coordinates   *)

       TypeMode = (Sign,Finger,Alpha);  (* typing modes         *)   


VAR    EExit: BOOLEAN;			(* cmd loop terminator  *)

       TMode: TypeMode;	        	(* current type mode    *)
       
       FHandCode: BYTE;                 (* fspell icon sym code *)
       FHandOpt:  BYTE;                 (* fspell icon opt byte *)
       PartCode:  BYTE;                 (* part mode icon code  *)

       altCmd, UpdateCmdList: BOOLEAN;	(* command line status  *)






 (* -----------------  Edit buffer structures  ------------------ *)



CONST  MaxEBuf = 29696;	       			 (* edit buffer size *)


TYPE   bufInx = INTEGER;			     (* buffer index *)

       buffer = ARRAY [0..MaxEBuf] OF BYTE;
       bufPtr = ^buffer;
	      

VAR    buf:    bufPtr;				(* edit buffer	     *)
       bufEnd: bufInx;				(* 1st open buf byte *)
       Maxbuf: bufInx;				(* buffer high bound *)






 (* -------------------  Alt buffer structures  --------------------- *)



CONST  MaxABuf = 10240;	       			 (* alt buffer size  *)


TYPE   altbuffer = ARRAY [0..MaxABuf] OF BYTE;
       altPtr	 = ^altbuffer;


VAR    altbuf: altPtr;				(* alt buffer	     *)

     




 (* ---------------------  Line structures  --------------------- *)



CONST  MaxLn   = 2;	       		    (* max line (0-based)     *)
       MaxLnHt = 59;			    (* max line height (dots) *)


TYPE   linInx = 0..MaxLn;


VAR    LineStart: ARRAY [linInx] OF bufInx;
       LineBase:  ARRAY [linInx] OF scrInx;	(* Y offsets of lines *)
       Line:	  linInx;			(* current line	      *)





       
 (* -----------------------  Sign structures  --------------------- *)



CONST  SpWidth = 36;	       		(* space sym width (in dots) *)
       SpYOff  = MaxLnHt;	       	(* Y offset of space symbol  *)


TYPE   sgnInx = INTEGER;		(* sign-relative coordinates *)


VAR    SignInx:	  bufInx;			(* cur sign inx	     *)
       SignBase:  scrInx;			(* cur sign base     *)
       SignLen:	  sgnInx;			(* cur sign length   *)
       SignLen2:  sgnInx;			(* sgnlen - lastsym  *)
       SignNxt:	  bufInx;			(* follow sign inx   *)
       
       

       


 (* --------------------  Marker structures  --------------------- *)



CONST  MrkLen   = 24;	       (* marker height (in dots)  *)
       MrkWidth =  7;	       (* marker sym width (dots)  *)
       MrkYOff  =  6;	       (* Y offset of sel marker   *)


VAR    Lmark, Rmark, curmark: bufInx;		(* text select marks *)







 (* --------------------  Cursor structures  --------------------- *)



CONST  CursLMar = 6;	       (* cursor left margin  *)	
       CursRMar = 630;	       (* cursor right margin *)	
       CursWMar = 8;           (* window left margin  *)

       CursHomeY = 35;	       (* cursor std line height *)

       CursYOff  = 4;	       (* Y offset of std cursor    *)
       CursIn    = 16;         (* cursor indent at new sign *)
       CursTab   = 16;	       (* cursor undent for tab pos *)	
       CursARet  = 16;	       (* cursor down on alpha retn *)
	
       CursDelay = 40;	       (* cursor blink delay	*)
       CursOn	 = 15;	       (* cursor blink on time	*)
       CursOff	 = 5;	       (* cursor blink off time *)


TYPE   cursInx = INTEGER;		     (* cursor loc index  *)

       CursPos = RECORD			     (* sym cursor coords *)
		   X, Y: sgnInx;
		 END;


VAR    CursX, CursY: scrInx;			(* cursor position *)
       NewX, NewY:   scrInx;			(* new cursor pos  *)

       CursSelect: BOOLEAN;			(* cursor key status *)
       CursGrid:   ARRAY [0..7] OF CursPos;	(* active curs locs  *)
       CursLoc:	   cursInx;			(* cur active loc    *)







 (* ---------------------  Symbol codes  ----------------------- *)



CONST  CharCurs = 127;	       (* blk cursor char code *)
       CopyRght =  31;	       (* copyright char code  *)
       
       CharCode = 250;	       (* ASCII char sym code	 *)
       AHead    = 251;	       (* arrowhead sym code   	 *)
       Cursor   = 251;	       (* cursor sym code	 *)
       Marker   = 252;	       (* select marker sym code *)
       Spacer   = 252;	       (* intersign space code   *)

       NewSign  = 253;	       (* control codes *)
       NewLine  = 254;
       EOF      = 255;






 (* ---------------------  Symbol structures  ----------------------- *)



CONST  Vface	= 0;	       			     (* symbol types *)
       Face	= 1;
       Hand	= 2;
       Movement = 3;
       Shoulder = 4;
       Contact	= 5;
       Other	= 6;
       Xmvmt	= 7;

       
       SizeBias = 0;	       	       		  (* arrowhead types *)
       Solid	= 1;
       Hollow	= 2;
       Open	= 3;


       SymFreq  = 1;	               (* select symbol shimmer time *)


     
TYPE   Attrs = PACKED RECORD		      	 (* return sym attrs *)
        	 Stype	   : BYTE;
		 Sizes	   : BYTE;
		 Rotations : BYTE;
		 Flops	   : BYTE;
		 Images    : BYTE;
	       END;
       
       Info = PACKED RECORD		      	  (* return img info *)
		Xsize	   : BYTE;
		Ysize	   : BYTE;
		FixPtX	   : BYTE;
		FixPtY	   : BYTE;
		AHeadPtX   : BYTE;
		AHeadPtY   : BYTE;
		AHeadOrient: BYTE;
	      END;


       
VAR   SymSelect:    BOOLEAN;			(* sym select status *)
      SymInx:	    bufInx;			(* selected sym inx  *)
      SymLast:	    BOOLEAN;			(* rghtmost in sign? *)
      SymOpt:       BYTE;			(* sym option byte   *)
      SymX1, SymX2: sgnInx;			(* sym X coords.     *)
      SymY1, SymY2: sgnInx;			(* sym Y coords.     *)
      SymAttrs:     Attrs;			(* sym attributes    *)
      SymInfo:	    Info;			(* sym information   *)

       
      TSymAttr  [Public]: ADS OF ARRAY [WRD(0)..255] OF WORD;   (* Atr tabl *)
      TSymIndex [Public]: ADS OF ARRAY [WRD(0)..255] OF ADSMEM; (* Sym tabl *)
      TChrIndex [Public]: ADS OF ARRAY [WRD(0)..255] OF ADSMEM; (* Chr tabl *)







 (* -------------------  Option byte structures  --------------------- *)



CONST  StockOpt = 8;           		(* vanilla option byte value *)
       CursOpt  = 0;			(* cursor symbol option val  *)
       SCursOpt = CursOpt+#07;		(* standard cursor opt val   *)
       SpaceOpt = StockOpt+#20;		(* intersign space opt val   *)
       mLeft    = StockOpt; 		(* left sel marker opt val   *)
       mRight   = StockOpt+#80;		(* right sel marker opt val  *)


TYPE   bits4 = 0..15;
       bits3 = 0..7;
       bits2 = 0..3;
       bits1 = 0..1;

  
       OptionRec  = PACKED RECORD	      (* option byte format  *)
		      Rotation: bits3;
		      AHead:    bits2;
		      Size:     bits2;
		      Flop:     bits1;
	            END;


       OptByte = BYTE;






 (* ----------------------  Character codes  ------------------------ *)



TYPE   SetOfChar = SET OF CHAR;                (* valid key sets *)



CONST  AllKeys   = SetOfChar [CHR(0)..CHR(255)];


       NameKeys  = SetOfChar ['a'..'z', 'A'..'Z', '0'..'9', 
			      CHR(128)..CHR(165),
                              CHR(181)..CHR(183), CHR(198)..CHR(199),
                              CHR(208)..CHR(212), CHR(214)..CHR(216),
                              CHR(222), CHR(224)..CHR(237), CHR(245)];


       DNameKeys = SetOfChar [' '..'~',
                              CHR(128)..CHR(165),
                              CHR(181)..CHR(183), CHR(198)..CHR(199),
                              CHR(208)..CHR(212), CHR(214)..CHR(216),
                              CHR(222), CHR(224)..CHR(237), CHR(245)];


       NumKeys   = SetOfChar ['0'..'9'];


       PathKeys  = SetOfChar [':', '\', '.', 
                              'a'..'z', 'A'..'Z',
                              '0'..'9', CHR(128)..CHR(165),
                              CHR(181)..CHR(183), CHR(198)..CHR(199),
                              CHR(208)..CHR(212), CHR(214)..CHR(216),
                              CHR(222), CHR(224)..CHR(237), CHR(245)];



TYPE   Ct = ARRAY [128..237] OF INTEGER;  (* multi-ling case conversion *) 

  
CONST  a_z = SetOfChar ['a'..'z'];
     
       AtZ = SetOfChar ['A'..'Z'];

       HiC = SetOfChar [CHR(128)..CHR(237)];



CONST  UpC = Ct(128, 154,
                144, 182, 142, 183, 143, 128, 210, 211, 212, 216,
                215, 222, 142, 143, 144, 146, 146, 226, 153, 227,
                234, 235, 152, 153, 154, 157, 156, 157, 158, 159, 
                181, 214, 224, 233, 165, 165, 166, 167, 168, 169,
                170, 171, 172, 173, 174, 175, 176, 177, 178, 179,
                180, 181, 182, 183, 184, 185, 186, 187, 188, 189,
                190, 191, 192, 193, 194, 195, 196, 197, 199, 199,
                200, 201, 202, 203, 204, 205, 206, 207, 209, 209,
                210, 211, 212, 213, 214, 215, 216, 217, 218, 219,
                220, 221, 222, 223, 224, 225, 226, 227, 229, 229,
                230, 232, 232, 233, 234, 235, 237, 237);


CONST  LoC = Ct(135, 129,
                130, 131, 132, 133, 134, 135, 136, 137, 138, 139,
                140, 141, 132, 134, 130, 145, 145, 147, 148, 149,
                150, 151, 152, 148, 129, 155, 156, 155, 158, 159, 
                160, 161, 162, 163, 164, 164, 166, 167, 168, 169,
                170, 171, 172, 173, 174, 175, 176, 177, 178, 179,
                180, 160, 131, 133, 184, 185, 186, 187, 188, 189,
                190, 191, 192, 193, 194, 195, 196, 197, 198, 198,
                200, 201, 202, 203, 204, 205, 206, 207, 208, 208,
                136, 137, 138, 213, 161, 140, 139, 217, 218, 219,
                220, 221, 141, 223, 162, 225, 147, 149, 228, 228,
                230, 231, 231, 163, 150, 151, 236, 236);



CONST  esc   =	27;	       (* command key codes *)
       del   = 127;
       tab   =	 9;  
       left  =	 8;
       down  =	10;
       up    =	11;
       right =	12;
       ret   =	13; 
       fsp   =	14;
       home  =  15;
       hend  =  16;
       pgup  =  17;
       pgdn  =  18;
       bs    =  19;






 (* -----------------  Character sort structures  ----------------- *)



CONST  MaxPair = 4;			  (* max char pairs *)


TYPE   CharMap = ARRAY [0..255] OF BYTE;  (* char codes to sort values *) 


       PairRec = RECORD
		   pair: LSTRING(2);	  (* character pair *)
                   sval: BYTE;		  (* sort value     *)
		 END;


       PairMap = ARRAY [0..MaxPair] OF PairRec;


       SortMap = RECORD
		   cnty: WORD;		  (* dict country   *)
		   cmap: CharMap;
		   pmap: PairMap;
		 END;


       SMapPtr = ADS OF SortMap;



VAR    Ccurmap: SMapPtr;		  (* current country sort map *)

       Dcurmap: SMapPtr;		  (* current dict sort map    *)






 (* ----------------------  Key scan codes  ------------------------ *)



CONST  SC_Change = 49;         (* sym cmd key scan codes *)
       SC_Cursor = 50;
       SC_Mirror = 51;
       SC_Size   = 52;
       SC_Rotate = 53;


       SC_a = 30;              (* keybox key scan codes *)
       SC_s = 31;
       SC_d = 32;
       SC_f = 33;
       SC_g = 34;
       SC_h = 35;
       SC_j = 36;
       SC_k = 37;
       SC_l = 38;
       SC_x = 39;

       
       sShft =  3;             (* shift byte code values *)
       sCtrl =  4;
       sAlt  =  8;
       sCap  = 64;



CONST  Z     = #2C;             (* debug -- data key scan codes *)
       X     = #2D;
       C     = #2E;
       V     = #2F;
       B     = #30;
       N     = #31;
       M     = #32;
       Comma = #33;
       Dot   = #34;
       Slash = #35;
     
       A     = #1E;
       S     = #1F;
       D     = #20;
       F     = #21;
       G     = #22;
       H     = #23;
       J     = #24;
       K     = #25;
       L     = #26;
       Semi  = #27;
       Rquot = #28;
     
       Q     = #10;
       W     = #11;
       E     = #12;
       R     = #13;
       T     = #14;
       Y     = #15;
       U     = #16;
       I     = #17;
       O     = #18;
       P     = #19;
       Lbrac = #1A;
       Rbrac = #1B;
       Baksl = #2B;
     
       Lquot = #29;
       n1    = #02;
       n2    = #03;
       n3    = #04;
       n4    = #05;
       n5    = #06;
       n6    = #07;
       n7    = #08;
       n8    = #09;
       n9    = #0A;
       n0    = #0B;
       Minus = #0C;
       Equal = #0D;
     
     
       zbs  = #0E;         (* debug -- command key scan codes *)
       zret = #1C;
       zesc = #01;
       ztab = #0F;






 (* ------------------------  Command codes  ------------------------ *)


       
CONST  C_Select   = 0;          (* edit screen cmd codes *)
       C_Copy     = 1;
       C_Erase    = 2;
       C_Find     = 3;
       C_Replace  = 4;
       C_Margin   = 5;
       C_Quit     = 6;
       C_more     = 7;
       
       C_Undo     = 8;
       C_Get      = 9;
       C_Jump     = 10;
       C_Next     = 11;
       C_Previous = 12;
       C_Adjust   = 13;
       C_Type     = 14;
       C_Dict     = 15;
       C_more2    = 16;


       C_Edit     = 0;          (* file screen cmd codes *)
       C_New      = 1;
       C_Print    = 2;
       C_File     = 3;
       C_Dir      = 4;
       C_Setup    = 5;
       C_Quit2    = 6;


       C_Fnext    = 0;          (* Find cmd codes *)
       C_Fpart    = 1;
       C_Fstop    = 2;


       C_Rchange  = 0;          (* Replace cmd codes *)
       C_Rnext    = 1;
       C_Rall     = 2;
       C_Rstop    = 3;


       C_Qsave    = 0;          (* Quit cmd codes *)
       C_Qwrite   = 1;
       C_Qbackup  = 2;
       C_Qexit    = 3;


       C_Jbegin   = 0;          (* Jump cmd codes *)
       C_Jend     = 1;
       C_Jmarker  = 2;


       C_Tsign    = 0;          (* Type cmd codes *)
       C_Tfinger  = 1;
       C_Talfa    = 2;


       C_Fdelete  = 0;          (* File cmd codes *)
       C_Frename  = 1;
       C_Fexport  = 2;


       C_Sprnting = 0;          (* Setup cmd codes *)
       C_Scountry = 1;
       C_Sdiction = 2;


       C_Sprinter = 0;		(* Setup printing cmd codes *)
       C_Sintface = 1;
       C_Sformat  = 2;
       C_Snumbers = 3;
       C_Slines   = 4;


       C_Sepson   = 0;		(* Setup printer cmd codes *)
       C_Shp      = 1;
       C_Sps      = 2;


       C_No       = 0;          (* "(Y/N)?" cmd codes *)
       C_Yes      = 1;



       
       

       
 (* ----------------------  Key structures  ----------------------- *)



CONST  kMaxMap  = 3000;			     (* key map size     *)


TYPE   keyInx	= INTEGER;		     (* key map index    *)

       SymRec = PACKED RECORD                (* key sym entry    *)
                  codeval: BYTE;
		  rotval : BYTE;
	        END;
       
       InxRec = RECORD CASE INTEGER OF       (* key table entry  *)
                  0: (inx: INTEGER);
                  1: (sym: SymRec);
	        END;



       ShiftState = (None, Shft, Alt,        (* keyboard shift state *)   
                     Ctrl, CtrlShft,
                     CtrlAlt, ShftAlt);

       ScnCode    = 0..255;                  (* key scan code *)


       Kbd    = ARRAY [TypeMode,ScnCode,ShiftState] OF InxRec;  (* key tab *)

       KbdPtr = ADS OF Kbd;



VAR    Keyboard:  KbdPtr;                       (* key-to-codes table *)
       keymap:    ARRAY [0..kMaxMap] OF BYTE; 	(* complex-key table  *)






 (* -------------   Fingerspell descender structures  ---------------- *)



CONST  MaxD   = 100;

TYPE   Des    = RECORD
                  fsym, 			   (* target symbol *)
                  fopt,
                  dval: ARRAY [0..MaxD] OF BYTE;   (* descent value *)
                END;

VAR    Descender: Des;






 (* ---------------------  Message structures  -------------------- *)



CONST  MaxMsg   = 102;				(* # msgs in file   *)


TYPE   msgInx	= INTEGER;			(* message index    *)
       cmdVal   = INTEGER;                      (* command value    *)

       cString =  LSTRING(80);			(* std string type  *)
       pString =  ^cString;			(* std string ptr   *)


       msgPtr = ADS OF cString;
       cmdPtr = ^CmdRec;

       cmdRec = RECORD                          (* cmd list entry *)
                  ch:  CHAR;
                  val: cmdVal;
                  nxt: cmdPtr;
                END;

       MsgRec = RECORD                          (* message record *)
                  msg: msgPtr;
                  cmd: cmdPtr;
                END;

       Messages = ARRAY [1..MaxMsg] OF MsgRec;  (* message list   *)


VAR    Msg: Messages;                           (* SW message list *)







 (* ---------------------  Undo structures   ----------------------- *)


       
VAR    unstate: (uncopy, unerase, unsymbol);	(* undo state	   *)
       unmark:	bufInx;				(* undo marker	   *)
       unlen:	INTEGER;			(* # bytes to undo *)
       undif:   INTEGER;			(* space to undo   *)






 (* ---------------------  Window structures   ---------------------- *)



CONST  WndTop   = 29;				(* window position  *)

       WndHt    = MaxLnHt + 1;		        (* window ht (dots) *)

       FWndBase = WndTop + 1;			(* Find window base *)

       RWndBase = WndTop + WndHt + 2;		(* Repl window base *)



CONST  MaxWBuf = 149;	       			(* window buf size  *)


TYPE   winbuffer  = ARRAY [0..MaxWBuf] OF BYTE;
       winPtr	  = ^winbuffer;


VAR    WExit: BOOLEAN;				(* cmd loop terminator *)

       fbuf, rbuf:  winPtr;			(* Find & Repl bufs  *)
       
       Window: BOOLEAN;				(* window edit mode  *)

       wEdbuf: bufPtr;				(* save ed buf ptr   *)

       wLmark, wRmark: bufInx;			(* save text marks   *)

       wTerm:  CHAR;				(* edit term status  *)






 (* --------------  Dictionary file structures   ------------------ *)



CONST  MaxData = 39;

       MaxDRec = MAXWORD - 1;			(* max rec in dict file *)
       DRecEOL = MAXWORD;			(* end-of-list marker   *)


TYPE   DRec = RECORD 				(* dict file record *)	
	        len:  SINT;
		data: ARRAY [0..MaxData] OF BYTE;
	        link: WORD;
	      END;

       DTitle = LSTRING(6);			(* dictionary name *)



VAR    DictF:     FILE OF DRec;			(* dictionary file   *)
       DinxF:     blkFile;      		(* dict index file   *)

       DictFree:  WORD;				(* dict file free list *)
       DictRecs:  WORD;				(* dict file rec count *)






 (* --------------  Dictionary index structures   ----------------- *)



CONST  MaxDV = 1;
       MaxDE = 5399;
       MaxDK = 9;

       
TYPE   DKey      = LSTRING(MaxDK);	       	(* dict index key  *)

       DinxEntry = RECORD
 		     key:    DKey;
		     recnum: WORD;
  	           END;
	
       DinxArray = ARRAY [0..MaxDE] OF DinxEntry;
       
       DinxVol   = RECORD
	             cnt:    INTEGER;
                     dentry: ADS OF DinxArray;
          	   END;

       DIndex = ARRAY [0..MaxDV] OF DinxVol;


VAR    DictIndex: DIndex;			(* dictionary index  *)






 (* -----------------  Dictionary edit structures   ------------------ *)



CONST  DWndBase = WndTop + 1;			(* dict window base    *)

       MaxDN = 228;				(* max entry name len  *)



TYPE   DName = LSTRING(MaxDN);	       		(* dict entry name     *)


VAR    DictAvail:   BOOLEAN;			(* dictionary avail    *)
       DictChanged: BOOLEAN;			(* dictionary changed  *)
       Dict:        BOOLEAN;			(* dict edit mode      *)
	
       DictName:   DName;			(* entered sign name   *)
       DictCurVol: INTEGER;			(* current dict volume *)
       DictCurEnt: INTEGER;			(* current dict entry  *)
       
       DictOnNew:  BOOLEAN;			(* editing new entry   *)
       DictNewVol: INTEGER;			(* new entry vol       *)
       DictNewEnt: INTEGER;			(* new entry num       *)


       dbuf, dbufb: winPtr;			(* dict symcode bufs   *)






 (* --------------  Dictionary text edit structures   ---------------- *)



CONST  DTxtBase = WndTop + WndHt + 5;		(* text window base    *)

       DTxtHt   = 31;				(* text wind ht (dots) *)



CONST  DMaxLine = 3;				(* max text lines      *)

       DMaxCol  = 76;				(* max text width      *)



TYPE   DlsT =  ARRAY [1..DMaxLine] OF INTEGER;  (* line start array    *)



VAR    DTxtWnd: BOOLEAN;			(* in text window      *)

       dtxt, dtxtb: DName;		        (* text bufs           *)

       Dls: DlsT;                               (* line start array    *)

       DCursLine, DCursCol: INTEGER;		(* text cursor         *)

       DSaveCh: CHAR;				(* undo "buffer"       *)






 (* --------------  Setup configuration structures  ------------------ *)



TYPE   SymHeader = PACKED RECORD		(* Symbol file header      *)
                     PrPort [00] : BYTE;
                     ImgCnt [01] : WORD;
                     CCode  [03] : WORD;
                     PgLen  [05] : BYTE;
	             PrType [06] : BYTE;
		     PgForm [07] : BYTE;
		     PgNum  [08] : WORD;
                     DicTit [10] : DTitle;
                   END;


VAR    PrnType:   BYTE;                         (* printer type            *)       

       PrnPort:   BYTE;                         (* printer port config     *)       

       PageForm:  BYTE;                         (* page format on HP print *)

       PageLen:   BYTE;                         (* # lines printed on page *)

       PageNum:   WORD;                         (* starting page number    *)

       Country:   WORD;                         (* current country code    *)

       DictTitle: DTitle;			(* current dictionary      *)



CONST  NoPgNm  =  10000;			(* suppress page numbers   *)






 (* -------------------  File export structures  -------------------- *)



TYPE GrafData   = STRING(80);		    (* 1 screen line of graf data *)


     GrafHeader = PACKED RECORD             (* 128-byte PCX file header   *)
		    Manuf   [00] : BYTE;
		    Hard    [01] : BYTE;
		    Encod   [02] : BYTE;
		    BitPx   [03] : BYTE;
                    X1      [04] : WORD;
                    Y1      [06] : WORD;
                    X2      [08] : WORD;
                    Y2      [10] : WORD;
                    Hres    [12] : WORD;
                    Vres    [14] : WORD;
		    clrma   [16] : ARRAY [0..47] OF BYTE;
		    Vmode   [64] : BYTE;
		    NPlanes [65] : BYTE;
                    Bplin   [66] : WORD;
		    Xtra    [68] : ARRAY [0..59] OF BYTE;
		  END;






 (* ---------------------  Printing structures --------------------- *)



CONST  ef =  4;	  { ASCII eof       }
       lf = 10;	  { ASCII line feed }
       ff = 12;	  { ASCII form feed }
       cr = 13;   { ASCII return    }



TYPE   PNstr    = ARRAY [0..3] OF LSTRING(4);

CONST  PortName = PNstr ('COM1', 'COM2', 'LPT1', 'LPT2');






 (* --------------------  Page format structures -------------------- *)



CONST  MaxPF = 7;

TYPE   PFnum = ARRAY [1..MaxPF] OF INTEGER;

       PFst5 = ARRAY [1..MaxPF] OF LSTRING(5);

       PFst4 = ARRAY [1..MaxPF] OF LSTRING(4); 

       PFst3 = ARRAY [1..MaxPF] OF LSTRING(3);

       PFst2 = ARRAY [1..MaxPF] OF LSTRING(2);

       PFst1 = ARRAY [1..MaxPF] OF LSTRING(1);




CONST  PFleft = PFst3('260',  '580',  '150',  '150',  '135',  '338',   '75');

       PFtop  = PFst3('165',  '260',  '120',  '115',  '115',  '150',  '115');

       PFcol  = PFst4(  'X',    'X', '1640',  '750',  '768',    'X', '1088');

       PFline = PFst2(  '12',   '8',    '8',    '4',    '4',   '16',    '8');

       PFnlin = PFnum(    7 ,   10 ,     8 ,    22 ,    16 ,     4 ,    11 );

       PFncol = PFnum(    1 ,    1 ,     2 ,     3 ,     4 ,     1 ,     2 );



CONST  PFpgnmLW =  765;     (* page # X coord -- left   (wide format) *)

       PFpgnmRW = 2420;	    (*		         right  (wide format) *)

       PFpgnmCW = 1595;	    (*		         center (wide format) *)

       PFpgnmCT = 1202;	    (*                   center (tall format) *)

       PFpgnmW  = 2300;     (* page # Y coord --        (wide format) *)

       PFpgnmT  = 3020;     (*                          (tall format) *)



CONST  PFcharbias = 14;	    (* 1/2 char width -- for centering pg #'s *)






 (* -----------------  Epson printing structures ------------------ *)



CONST  PrEpson  = C_Sepson;   			 { ID code     }

       
       EpsInit  = CHR(esc) * '@' *               { prn reset   }	
                  CHR(esc) * 'l' * CHR(5)  *     { margin = 1" }	
		  CHR(esc) * 'A' * CHR(8)  *     { line = 1/9" }
                  CHR(esc) * 'J' * CHR(140);	 { top margin  }


       EpsPrnt  = CHR(esc) * '*' * CHR(6) *      { graf mode 6 }	
		  CHR(128) * CHR(2);	         { 640 dots    }	


       EpsLine  = CHR(esc) * 'J' * CHR(12);	 { line space  }


       EpsPage  = CHR(ff)  *			 { form feed   }
                  CHR(esc) * 'J' * CHR(140);	 { top margin  }



       EpsNlin  = 5;				 { lines/page  }


       EpsRet   = CHR(cr)  * CHR(lf);		 { new line    }


       EpsPgnm  = 35;				 { pg # indent }






 (* -----------  Hewlett-Packard printing structures -------------- *)



CONST  PrHP    = C_Shp;				 	   { ID code    }


       HPinit  = CHR(esc) * 'E' *		           { prn reset  }
	         CHR(esc) * '&' * 'l' * '2A'  *	   	   { paper size }
	         CHR(esc) * '&' * 'l' * '@O'  *	           { pg orient  }
	         CHR(esc) * '(' * 's' * '3T'  *            { type face  }
	         CHR(esc) * '(' * 's' * '0S'  *            { type style }
	         CHR(esc) * '(' * 's' * '12V' *            { type size  }
	         CHR(esc) * '(' * 's' * '10H' *            { type pitch }
                 CHR(esc) * '*' * 't' * '?R'  *	           { graf size  }
		 CHR(esc) * '*' * 'r' * '0F'  *	           { graf mode  }
		 CHR(esc) * '*' * 'b' * '2M'  *	           { comp mode  }
	         CHR(esc) * '*' * 'p' * '!x'  * '%Y' *     { set margin }
	         CHR(esc) * '*' * 'r' * '1A';	           { start graf }


       HPprnt  = CHR(esc) * '*' * 'b' * '#W';	           { set count  }	
       

       HPline  = CHR(esc) * '*' * 'p' * '+%Y';	           { line space }


       HPpage  = CHR(ff)  *				   { form feed  }
		 CHR(esc) * '*' * 'r' * 'bC'  * 	   { end graf   }
		 CHR(esc) * '*' * 'r' * '0F'  *	           { graf mode  }
		 CHR(esc) * '*' * 'b' * '2M'  *	           { comp mode  }
                 CHR(esc) * '*' * 'p' * '!x'  * '%Y' *     { set margin }
	         CHR(esc) * '*' * 'r' * '1A';	           { start graf }


       HPcol   = CHR(esc) * '*' * 'r' * 'bC'  * 	   { end graf   }
		 CHR(esc) * '*' * 'p' * '+!x' * '%Y' *     { new margin }
	         CHR(esc) * '*' * 'r' * '1A';	           { start graf }


       HPrule  = CHR(esc) * '*' * 'p' * '1216x' * '75Y' *  { set coords }
	         CHR(esc) * '*' * 'c' * '2850b' * '4A'  *  { set dimens }
	         CHR(esc) * '*' * 'c' * '0P';              { draw line  }


       HPpgnm  = CHR(esc) * '*' * 'p' * '!x' * '%Y';       { pg # coord }




CONST  HPsize = PFst3('100',  '150',  '150',  '300',  '300',   '75',  '150');

       HPpgor = PFst1(  '0',    '0',    '1',    '0',    '1',    '1',    '0');






 (* --------------  PostScript printing structures ---------------- *)



CONST  PrPS    = C_Sps;				   	   { ID code    }


       PSinit  = 'save '		      *		   { job start	}
		 '72 300 div dup neg scale '  *		   { coord sys	}
		 '-@ rotate '	              *		   { pg orient  }
	 	 '# ? translate '	      *		   { sys origin }
	         '/Courier findfont '         *		   { type face  }
		 '[50 0 0 -50 0 0] makefont ' *	           { type size  }
                 'setfont '		      *		   { type font	}
                 'save '		      *		   { first page }
	         '! % moveto ';                            { set margin }


       PSprnt  = 'gsave '		      *		   { save state }
                 'currentpoint translate '    *		   { set origin }
		 '! ? scale '		      *		   { set size   }
		 '# 1 1 [1 0 0 -1 0 0] {< '; 		   { set count  }


       PSpend  = '>} image grestore ';		   	   { print dots }


       PSline  = '0 % rmoveto ';	           	   { line space }


       PSpage  = 'restore gsave showpage '    *		   { print page }
	         'grestore save ! % moveto ';		   { set margin }


       PScol   = 'currentpoint pop % moveto ' *		   { top margin }
		 '! 0 rmoveto ';			   { lft margin }


       PSrule  = 'newpath '	              *		   { set path   }
		 '1216 75 moveto '    	      *		   { set coords }
	         '0 2850 rlineto '    	      *		   { set length }
		 '3 setlinewidth '    	      *		   { set width  }
		 'stroke ';				   { draw line  }


       PSpgnm  = '! % moveto ';       			   { pg # coord }


       PStext  = '(#) show ';	           		   { print text }


       PSterm  = 'restore '* CHR(ef);	           	   { job finish }


       PSesc   = 'restore '* PSterm;			   { job exit   }




CONST PSsize = PFnum(     3,       2,     2,       1,     1,     4,       2 );

      PSpgor = PFst2(    '0',     '0',  '90',     '0',  '90',  '90',     '0');
						   
      PSleft = PFst2(   '75',    '75',  '55',    '75',  '55',  '55',    '75');

      PStop  = PFst5('-3150', '-3150', '150', '-3150', '150', '150', '-3150');






	    
 (* ---------------  Forward procedure declarations  ----------------- *)

       
       
PROCEDURE Command; FORWARD;


PROCEDURE Error (CONSTS errmsg: LSTRING); FORWARD;


PROCEDURE FGetCommand; FORWARD;


PROCEDURE WriteLine (CONSTS line: LSTRING); FORWARD;


PROCEDURE DrawText (x,y: scrInx; CONSTS msg: LSTRING); FORWARD;


PROCEDURE DrawGauge (bitval: INTEGER); FORWARD;


PROCEDURE DrawFileBox; FORWARD;


FUNCTION  SWOpen (VAR f: FCBFQQ; CONST name: LSTRING): BOOLEAN; FORWARD;







 (* ------------------  External procedures  -------------------- *)



PROCEDURE KeyInit; EXTERNAL;


PROCEDURE KeyTerm; EXTERNAL;


PROCEDURE CriticalInit; EXTERNAL;


FUNCTION  KeyAvail: BOOLEAN; EXTERNAL;


FUNCTION  KeyRead (VAR code: INTEGER; VAR shift: WORD): BOOLEAN; EXTERNAL;




PROCEDURE ClearScreen; EXTERNAL;


PROCEDURE GrafMode; EXTERNAL;


PROCEDURE TextMode; EXTERNAL;
  

PROCEDURE Draw (sym, opt: BYTE; x, y: INTEGER); EXTERNAL;


PROCEDURE Erase (sym, opt: BYTE; x, y: INTEGER); EXTERNAL;


PROCEDURE QDraw (sym, opt: BYTE; x, y: INTEGER; exor: BOOLEAN); EXTERNAL;


PROCEDURE QErase; EXTERNAL;




PROCEDURE GetAttrs (code: BYTE; VAR attr: Attrs); EXTERNAL;


PROCEDURE GetInfo (code, opt: BYTE; VAR inf: Info); EXTERNAL;




PROCEDURE DrawLine (x1, y1, x2, y2, bitval: INTEGER); EXTERNAL;


PROCEDURE Scrollup; EXTERNAL;


PROCEDURE Scrolldown (line: linInx); EXTERNAL;


PROCEDURE EraseLine (line: linInx); EXTERNAL;


PROCEDURE SaveBoxArea (save: BOOLEAN); EXTERNAL;




FUNCTION ScanSym (sym, opt: BYTE;
		  buf:ADRMEM; inx: bufInx): INTEGER; EXTERNAL;


FUNCTION ScanCode (stopsigns: BOOLEAN; dir: INTEGER; 
		   buf:ADRMEM; inx: bufInx): INTEGER; EXTERNAL;




PROCEDURE EpsonPrep (scrline: scrInx; printbuf: ADRMEM;	  
	             VAR nonblank: BOOLEAN); EXTERNAL;


PROCEDURE GrafPrep (scrline: scrInx; grafbuf: ADRMEM); EXTERNAL;



PROCEDURE RenameFile (CONST oldname, newname: LSTRING); EXTERNAL;


PROCEDURE GetCurPath (VAR Path : LSTRING); EXTERNAL;



FUNCTION GetDTA: ADSMEM; EXTERNAL;


FUNCTION DOSXQQ (command, parameter: WORD): BYTE; EXTERNAL;


FUNCTION TICS: WORD; EXTERNAL;



PROCEDURE ENDXQQ; EXTERNAL;

 
FUNCTION  GETUQQ (VAR f:FCBFQQ; len: WORD; dst: ADSMEM): WORD; EXTERNAL;


FUNCTION  PUTUQQ (VAR f:FCBFQQ; len: WORD; dst: ADSMEM): WORD; EXTERNAL;




FUNCTION  ALLMQQ (wants: WORD): ADSMEM; EXTERNAL;


FUNCTION  ALLHQQ (wants: WORD): WORD; EXTERNAL;

 
PROCEDURE DISMQQ (block: ADSMEM); EXTERNAL;




  

 
 (* ---------- Keyboard driver emulation procedures  ----------- *)


                  !!!for testing purposes only!!!


 
!  PROCEDURE AsmInit; EXTERNAL;
!  
!  PROCEDURE AsmTerm; EXTERNAL;
!  
!  FUNCTION  AGotChar: Boolean; EXTERNAL;
!  
!  PROCEDURE AGetChar (VAR ch: CHAR; VAR opt: BOOLEAN); EXTERNAL;
!  
!
!  PROCEDURE KeyInit;
!  BEGIN AsmInit END;
!  
!  PROCEDURE KeyTerm;
!  BEGIN AsmTerm END;
!  
!  FUNCTION KeyAvail: BOOLEAN;
!  BEGIN KeyAvail := AGotChar END;
!
!  FUNCTION KeyRead (VAR code: INTEGER; VAR shift: WORD): BOOLEAN;
!  
!  VAR ch:   CHAR;
!      opt: BOOLEAN;
!  
!  BEGIN
!    code  := 255;
!    shift := 0;
!    KeyRead := FALSE;
!  
!    IF AGotChar THEN
!    BEGIN
!      KeyRead := TRUE;
!      AGetChar(ch, opt);
!  
!      CASE ch OF
!        '`': code := Lquot;
!        '1': code := n1;
!        '2': code := n2;
!        '3': code := n3;
!        '4': code := n4;
!        '5': code := n5;
!        '6': code := n6;
!        '7': code := n7;
!        '8': code := n8;
!        '9': code := n9;
!        '0': code := n0;
!        '-': code := Minus;
!        '=': code := Equal;
!  
!        '~': BEGIN code := Lquot; shift := 1 END;
!        '!': BEGIN code := n1;    shift := 1 END;
!        '@': BEGIN code := n2;    shift := 1 END;
!        '#': BEGIN code := n3;    shift := 1 END;
!        '$': BEGIN code := n4;    shift := 1 END;
!        '%': BEGIN code := n5;    shift := 1 END;
!        '^': BEGIN code := n6;    shift := 1 END;
!        '&': BEGIN code := n7;    shift := 1 END;
!        '*': BEGIN code := n8;    shift := 1 END;
!        '(': BEGIN code := n9;    shift := 1 END;
!        ')': BEGIN code := n9;    shift := 1 END;
!        '_': BEGIN code := Minus; shift := 1 END;
!        '+': BEGIN code := Equal; shift := 1 END;
!      
!        'q': code := Q;
!        'w': code := W;
!        'e': code := E;
!        'r': code := R;
!        't': code := T;
!        'y': code := Y;
!        'u': code := U;
!        'i': code := I;
!        'o': code := O;
!        'p': code := P;
!        '[': code := Lbrac;
!        ']': code := Rbrac;
!        '\': code := Baksl;
!    
!        'Q': BEGIN code := Q;     shift := 1 END;
!        'W': BEGIN code := W;     shift := 1 END;
!        'E': BEGIN code := E;     shift := 1 END;
!        'R': BEGIN code := R;     shift := 1 END;
!        'T': BEGIN code := T;     shift := 1 END;
!        'Y': BEGIN code := Y;     shift := 1 END;
!        'U': BEGIN code := U;     shift := 1 END;
!        'I': BEGIN code := I;     shift := 1 END;
!        'O': BEGIN code := O;     shift := 1 END;
!        'P': BEGIN code := P;     shift := 1 END;
!        '{': BEGIN code := Lbrac; shift := 1 END;
!        '}': BEGIN code := Rbrac; shift := 1 END;
!        '|': BEGIN code := Baksl; shift := 1 END;
!    
!        'a': code := A;
!        's': code := S;
!        'd': code := D;
!        'f': code := F;
!        'g': code := G;
!        'h': code := H;
!        'j': code := J;
!        'k': code := K;
!        'l': code := L;
!        ';': code := Semi;
!       '''': code := Rquot;
!    
!        'A': BEGIN code := A;     shift := 1 END;
!        'S': BEGIN code := S;     shift := 1 END;
!        'D': BEGIN code := D;     shift := 1 END;
!        'F': BEGIN code := F;     shift := 1 END;
!        'G': BEGIN code := G;     shift := 1 END;
!        'H': BEGIN code := H;     shift := 1 END;
!        'J': BEGIN code := J;     shift := 1 END;
!        'K': BEGIN code := K;     shift := 1 END;
!        'L': BEGIN code := L;     shift := 1 END;
!        ':': BEGIN code := Semi;  shift := 1 END;
!        '"': BEGIN code := Rquot; shift := 1 END;
!    
!        'z': code := Z;
!        'x': code := X;
!        'c': code := C;
!        'v': code := V;
!        'b': code := B;
!        'n': code := N;
!        'm': code := M;
!        ',': code := Comma;
!        '.': code := Dot;
!        '/': code := Slash;
!             
!        'Z': BEGIN code := Z;     shift := 1 END;
!        'X': BEGIN code := X;     shift := 1 END;
!        'C': BEGIN code := C;     shift := 1 END;
!        'V': BEGIN code := V;     shift := 1 END;
!        'B': BEGIN code := B;     shift := 1 END;
!        'N': BEGIN code := N;     shift := 1 END;
!        'M': BEGIN code := M;     shift := 1 END;
!        '<': BEGIN code := Comma; shift := 1 END;
!        '>': BEGIN code := Dot;   shift := 1 END;
!        '?': BEGIN code := Slash; shift := 1 END;
!  
!        chr(8) : code := zbs;
!        chr(13): code := zret;
!        chr(27): code := zesc;
!        chr(9) : code := ztab;
!      
!        OTHERWISE code := 255;
!      
!      END;
!      IF opt THEN shift := shift + 8;
!
!    END;
!  
!  END;  (* KeyRead *)






  (* -------------------  Time procedure  ------------------- *)



PROCEDURE Pause (keybreak: BOOLEAN; ticks: INTEGER);

VAR oldtick, newtick: WORD;

BEGIN
  newtick := TICS;
  oldtick := newtick;

  WHILE ticks > 0 DO
  BEGIN
    IF keybreak AND THEN KeyAvail THEN BREAK;

    newtick := TICS;

    IF newtick <> oldtick THEN
    BEGIN
      ticks := ticks - 1;
      oldtick := newtick;
    END;
  
  END;

END;  (* Pause *)






  (* -------------  Case conversion procedures  ------------- *)



PROCEDURE UpCaseStr (VAR name: LSTRING);

VAR i:  INTEGER;
    ch: CHAR;

BEGIN
  FOR i := 1 TO ORD(name.Len) DO
  BEGIN
    ch := name[i];
  
    IF ch IN a_z THEN ch := CHR(ORD(ch) - 32) ELSE
    IF ch IN HiC THEN ch := CHR(UpC[ORD(ch)]);

    name[i] := ch;

  END;

END;  (* UpCaseStr *)



FUNCTION UpCase (ch: CHAR): CHAR;

BEGIN
  IF ch IN a_z THEN UpCase := CHR(ORD(ch) - 32) ELSE
  IF ch IN HiC THEN UpCase := CHR(UpC[ORD(ch)]) ELSE
		    UpCase := ch;

END; (* UpCase *)



FUNCTION LoCase (ch: BYTE): BYTE;

BEGIN
  IF CHR(ch) IN AtZ THEN LoCase := WRD(ORD(ch) + 32) ELSE
  IF CHR(ch) IN HiC THEN LoCase := WRD(LoC[ORD(ch)]) ELSE
		         LoCase := ch;

END; (* LoCase *)






  (* --------------  Country code procedures  --------------- *)



PROCEDURE StrToCC (VAR code: WORD; CONST s: LSTRING);

BEGIN
  IF NOT Decode(s, code) THEN code := 1;
END;



PROCEDURE CCToStr (VAR s: LSTRING; code: WORD);

VAR cstr: LSTRING(3);

BEGIN
  IF NOT Encode(cstr, code:3) THEN
    cstr := '001';

  IF cstr[1] = ' ' THEN cstr[1] := '0';
  IF cstr[2] = ' ' THEN cstr[2] := '0';

  CopyLST(cstr, s);

END;



PROCEDURE BuildCCName (VAR name: LSTRING;
                       CONST base: LSTRING; code: WORD);

VAR cstr: LSTRING(3);

BEGIN
  CCToStr(cstr, code);
  CopyLST(base, name);
  Insert(cstr, name, Positn('.', name, 1));
END;
  





  (* -------------  Dictionary title procedure  ------------- *)



PROCEDURE BuildDTName (VAR name: LSTRING;
  		       CONST base, dtitle: LSTRING);

BEGIN
  CopyLST(base, name);
  Insert(dtitle, name, Positn('.', name, 1));
END;
  





  (* --------------  Initialization procedures  --------------- *)



  (* NOTE - The following messages appear (regardless of  *)
  (*        country) if SignWriter dies before it can     *)
  (*        load the message file (done in proc InitB).   *)


CONST err1 =  'Not enough memory to start SignWriter';
      err2 =  'Symbol file not found';
      err3 =  'Error reading symbol file';
      err4 =  'Message file not found';
      err5 =  'Error reading message file';
      err6 =  'Invalid message';


VAR memerr: cString;  (* current out-of-memory message *)



PROCEDURE Terminate (CONSTS msg: LSTRING);

BEGIN
  TextMode;
  Writeln(msg);
  KeyTerm;
  ENDXQQ;

END;   (* Terminate *)



PROCEDURE Warning (CONSTS msg: LSTRING);

BEGIN
  TextMode;
  Writeln(msg);
  Pause(TRUE, StartDelay);

END;   (* Warning *)



PROCEDURE CheckIO (VAR f: FCBFQQ; fileopen: BOOLEAN; CONSTS msg: LSTRING);

BEGIN
  IF (f.errs > 0) OR NOT fileopen THEN
    Terminate(msg);
END;



FUNCTION TestIO (VAR f: FCBFQQ; fileopen: BOOLEAN; CONSTS msg: LSTRING): BOOLEAN;

VAR rslt: BOOLEAN;

BEGIN
  rslt := (f.errs = 0) AND fileopen;
  TestIO := rslt;

  IF NOT rslt THEN Warning(msg);
END;



FUNCTION AllHeap (size: WORD): WORD;

VAR temp: WORD;

BEGIN
  temp := ALLHQQ(size);

  IF temp <= 1 THEN Terminate(memerr);

  AllHeap := temp;
END;



FUNCTION AllMem (size: WORD): ADSMEM;

VAR temp: ADSMEM;

BEGIN
  temp := ALLMQQ(size);

  IF temp.R <= 1 THEN Terminate(memerr);

  AllMem := temp;
END;



PROCEDURE GetMsg (linenum: INTEGER; VAR f: FCBFQQ; 
                  VAR msg: LSTRING; VAR prompt: BOOLEAN);

VAR i:  INTEGER;
    ch: CHAR;

  PROCEDURE BadMsg;
    BEGIN Terminate(err6) END;

BEGIN
  prompt := FALSE;
  
  Read(f,i);
  CheckIO(f, TRUE, err5);
  
  IF i <> linenum THEN BadMsg;

  REPEAT 
    Read(f,ch);
    CheckIO(f, TRUE, err5);
  UNTIL (ch <> ' ') OR EOLN(f);

  IF NOT ((ch = '!') OR (ch = '''')) THEN BadMsg;

  IF ch = '!' THEN   (* following msg is prompt *)
    BEGIN
      prompt := TRUE;
    
      REPEAT 
        Read(f, ch);
        CheckIO(f, TRUE, err5);
      UNTIL (ch = '''') OR EOLN(f);

      IF ch <> '''' THEN BadMsg;
    END;

  Readln(f, msg);
  CheckIO(f, TRUE, err5);

  i := Positn('''', msg, 1);   (* zap end quote and trailing blanks *)
  IF i <= 1 THEN BadMsg;

  msg.Len := WRD(i - 1);

END;   (* GetMsg *)
  
  

PROCEDURE AddCmd (inx: msgInx; newch: CHAR; newval: INTEGER);

VAR temp: cmdPtr;

BEGIN
  WITH Msg[inx] DO
  BEGIN
    temp := cmd;
    cmd  := ReType(cmdPtr, AllHeap(Sizeof(cmd^)));

    WITH cmd^ DO
    BEGIN
      ch  := newch;
      val := newval;
      nxt := temp;
    END;
  END;

END;   (* AddCmd *)



PROCEDURE InitSymFile;

VAR i: WORD;
    f: blkFile;
    SymSize, ChrSize: ARRAY [WRD(0)..255] OF INTEGER;

BEGIN
  memerr := err1;                             (* hard error message     *)

  TSymIndex := AllMem(Sizeof(TSymIndex^));    (* Alloc space for tables *)
  TChrIndex := AllMem(Sizeof(TChrIndex^));
  TSymAttr  := AllMem(Sizeof(TSymAttr^));


  CheckIO(f, SWOpen(f, SymFile), err2);       (* open symbol file     *)

  PrnPort := f^;                              (* read printer config  *)

  f.Errs := GETUQQ(f, 2, ADS(i));             (* skip image count...  *)

  f.Errs := GETUQQ(f, 2, ADS(Country));       (* read country code    *)
  f.Errs := GETUQQ(f, 1, ADS(PageLen));       (* read page length     *)
  f.Errs := GETUQQ(f, 1, ADS(PrnType));       (* read printer type    *)
  f.Errs := GETUQQ(f, 1, ADS(PageForm));      (* read page format     *)
  f.Errs := GETUQQ(f, 2, ADS(PageNum));       (* read page number     *)

  f.Errs := GETUQQ(f, Sizeof(DictTitle), ADS(DictTitle));  (* read dt *)


  Reset(f);
  Seek(f, 513);    (* skip header *)
  
  f.Errs := GETUQQ(f, Sizeof(ChrSize),   ADS(ChrSize));  (* read tables *)
  f.Errs := GETUQQ(f, Sizeof(SymSize),   ADS(SymSize));
  f.Errs := GETUQQ(f, Sizeof(TSymAttr^), ADS(TSymAttr^));
  CheckIO(f, TRUE, err3);

  FOR i := 0 TO 255 DO                        (* read char images  *)
  BEGIN
    TChrIndex^[i] := AllMem(WRD(ChrSize[i]));
    f.Errs := GETUQQ(f, WRD(ChrSize[i]), ADS(TChrIndex^[i]^));
  END;

  FOR i := 0 TO 255 DO                        (* read sym images   *)
  BEGIN
    TSymIndex^[i] := AllMem(WRD(SymSize[i]));
    f.Errs := GETUQQ(f, WRD(SymSize[i]), ADS(TSymIndex^[i]^));
  END;
  
  CheckIO(f, TRUE, err3);
  Close(f);                                   (* close symbol file *)
  
END;   (* InitSymFile *)



PROCEDURE InitMsgFile;

VAR i, inx, cnt: INTEGER;
    str:  cString;
    cmds: BOOLEAN;
    f:    TEXT;

BEGIN
  memerr := err1;                             (* hard error message *)

  BuildCCName(str, MsgFile, Country);
  CheckIO(f, SWOpen(f, str), err4);

  FOR i := 1 TO MaxMsg DO
  BEGIN
    GetMsg(i, f, str, cmds); 

    WITH Msg[i] DO
    BEGIN
      cmd := NIL;

      msg := ReType(msgPtr, AllMem(WRD(1 + str.Len)));
      MoveSL(ADS(str), ADS(msg^), WRD(1 + str.Len));

      IF cmds THEN
      BEGIN
        inx := Positn('(', str, 1);
        
        IF inx > 1 THEN   (* "(Y/N)" prompt *)
          BEGIN
            AddCmd(i, str[inx+1], C_Yes);
            AddCmd(i, str[inx+3], C_No );
          END
        
        ELSE              (* command list *)
          BEGIN
            cnt := 0;

            WHILE str.Len > 0 DO
            BEGIN
              WHILE (str.Len > 0) AND THEN
                    (str[1] = ' ') DO Delete(str, 1, 1);

              IF str.Len = 0 THEN BREAK;

              AddCmd(i, str[1], cnt);
              cnt := cnt + 1;

              WHILE (str.Len > 0) AND THEN
                    (str[1] <> ' ') DO Delete(str, 1, 1);
            END;
          
          END;  (* else *)
  
      END;  (* if *)

    END  (* with *)
 
  END;  (* for *)

  Close(f);

END;   (* InitMsgFile *)



PROCEDURE InitBuffers;

VAR i: INTEGER;

BEGIN
  memerr := Msg[29].msg^;      (* soft memerr message *)

  FScreen := TRUE;
  Window  := FALSE;
  
  buf    := Retype(bufPtr, AllHeap(SizeOf(buf^)));
  maxBuf := MaxEBuf;
  bufEnd := 0;

  altbuf := Retype(altPtr, AllHeap(SizeOf(altbuf^)));

  Keyboard := Retype(KbdPtr, AllMem(SizeOf(Keyboard^)));

  Ccurmap := Retype(SMapPtr, AllMem(SizeOf(Ccurmap^)));
  Dcurmap := Retype(SMapPtr, AllMem(SizeOf(Dcurmap^)));

  fbuf := Retype(winPtr, AllHeap(SizeOf(fbuf^)));
  rbuf := Retype(winPtr, AllHeap(SizeOf(rbuf^)));
  
  fbuf^[0] := EOF;  fbuf^[1] := 0;
  rbuf^[0] := EOF;  rbuf^[1] := 0;

  dbuf  := Retype(winPtr, AllHeap(SizeOf(dbuf^)));
  dbufb := Retype(winPtr, AllHeap(SizeOf(dbufb^)));

  FOR i := 0 TO MaxDV DO
    WITH DictIndex[i] DO
      dentry := AllMem(Sizeof(dentry^));

  LineBase[0] := 0;
  
  FOR i := 1 TO MaxLn DO
    LineBase[i] := LineBase[i - 1] + MaxLnHt + 2;

  
END;   (* InitBuffers *)


  
PROCEDURE InitKeyFile;

VAR f:   blkFile;
    str: cString;

BEGIN
  BuildCCName(str, KeyFile, Country);
  CheckIO(f, SWOpen(f, str), Msg[30].msg^);

  FHandCode := f^;                                    (* fspell icon codes *)
  f.Errs := GETUQQ(f, 1, ADS(FHandOpt));
  
  f.Errs := GETUQQ(f, 1, ADS(PartCode));              (* "Part" icon code  *)


  f.Errs := GETUQQ(f, Sizeof(Ccurmap^), ADS(Ccurmap^));    (* sort table   *)


  Reset(f);        					   (* skip header  *)
  Seek(f, 513);


  f.Errs := GETUQQ(f, Sizeof(Keyboard^), ADS(Keyboard^));  (* keybd tables *)
  f.Errs := GETUQQ(f, Sizeof(Descender), ADS(Descender));
  f.Errs := GETUQQ(f, Sizeof(keymap),    ADS(keymap));
  
  CheckIO(f, TRUE, Msg[31].msg^);
  Close(f);

END;   (* InitKeyFile *)



PROCEDURE InitDict;

VAR i:       INTEGER;
    str:     cString;
    dcode:   BYTE;
    dheader: ARRAY [0..7] OF WORD;
    
BEGIN
  DictAvail := FALSE;


  BuildDTName(str, DicFile, DictTitle);
  IF NOT TestIO(DictF, SWOpen(DictF, str), Msg[88].msg^) THEN RETURN;


  BuildDTName(str, DinFile, DictTitle);
  IF NOT TestIO(DinxF, SWOpen(DinxF, str), Msg[88].msg^) THEN RETURN;


  dcode := DinxF^;    		      	      (* read index file header    *)

  IF dcode <> SWdict THEN 
    BEGIN Terminate(Msg[102].msg^); RETURN END;


  FOR i := 0 TO 7 DO
    DinxF.Errs := GETUQQ(DinxF, 2, ADS(dheader[i]));

  DictFree := WRD(dheader[0]);			
  DictRecs := WRD(dheader[1]);


  DinxF.Errs := GETUQQ(DinxF, Sizeof(Dcurmap^), ADS(Dcurmap^));  {sort table}

  IF NOT TestIO(DinxF, TRUE, Msg[87].msg^) THEN RETURN;

  
  Reset(DinxF);				      (* skip header     	   *)	
  Seek(DinxF, 513);
  

  FOR i := 0 TO MaxDV DO		      (* read dict index 	   *)	

    WITH DictIndex[i] DO
    BEGIN
      cnt := ORD(dheader[i+2]);
      DinxF.Errs := GETUQQ(DinxF, WRD(cnt)*Sizeof(dentry^[0]), ADS(dentry^));
    END;

  IF NOT TestIO(DinxF, TRUE, Msg[87].msg^) THEN RETURN;


  Close(DictF);
  Close(DinxF);


  DictAvail := TRUE;	     		      (* dict loaded and available *)


END;   (* InitDict *)



PROCEDURE InitScreen;    (* welcome message *)

VAR name: cString;

BEGIN
  ClearScreen;
  GrafMode;

  CopyLST('S i g n W r i t e r', name);
  DrawText(320, 72, name);

  CopyLST('4.3   ? 1995', name);
  name[7] := CHR(CopyRght);
  DrawText(344, 104, name);
  CopyLST('The Center for Sutton', name);
  DrawText(312, 120, name);

  CopyLST('Movement Writing', name);
  DrawText(328, 128, name);
  
  DrawGauge(1);
  DrawFileBox;

  UpdateFileScrn := FALSE;
  UpdateFileList := TRUE;
  UpdateCmdList  := TRUE;  

END;   (* InitScreen *)					   



PROCEDURE SWInit;

BEGIN
  InitSymFile;
  InitMsgFile;
  InitBuffers;
  InitKeyFile;
  InitDict;
  InitScreen;
END; 






  (* -------------  Dictionary handling procedures  -------------- *)



FUNCTION OpenDict: BOOLEAN;

VAR s1, s2: cString;

BEGIN
  BuildDTName(s1, DicFile, DictTitle);
  BuildDTName(s2, DinFile, DictTitle);

  OpenDict := SWOpen(DictF, s1) AND SWOpen(DinxF, s2);

END;  (* OpenDict *)



PROCEDURE CloseDict;

BEGIN
  Close(DictF);
  Close(DinxF);

END;  (* CloseDict *)



PROCEDURE ResetDict;

BEGIN
  DictAvail := OpenDict;
  CloseDict;

END;  (* ResetDict *)






  (* ------------  Option byte handling procedures  -------------- *)



PROCEDURE ReadRot (opt: OptByte; VAR val: INTEGER);

 { Get value from Rotation field (first 3 bits) of optbyte }

BEGIN
  val := ORD(opt) AND 7;    { Set val to lower 3 bits of Rotation field }
END;   



PROCEDURE ReadAhead (opt: OptByte; VAR val: BYTE);

 { Get value from Ahead field (4th and 5th bits) of optbyte }

VAR x: BYTE;

BEGIN
  x := opt DIV 8;           { Right justify 4th bit                  }
  val := x AND 3;           { Set val to lower 2 bits of Ahead field }
END;   



PROCEDURE ReadSize (opt: OptByte; VAR val: INTEGER);

 { Get value from Size field (6th and 7th bits) of optbyte.  Also get }
 { value from Ahead field.  If Ahead = SizeBias, increase Size value. }

VAR x, ahd: BYTE;

BEGIN
  x := opt DIV 32;          { Right justify 6th bit                 }
  x := x AND 3;        	    { Set val to lower 2 bits of Size field }

  ReadAhead(opt, ahd);	    { Check size bias from Ahead field      }

  IF ahd = SizeBias THEN
    x := x + 4;

  val := ORD(x);
END;   



PROCEDURE ReadFlop (opt: OptByte; VAR val: INTEGER);

 { Get value from Flop field (8th bit) of optbyte }

VAR x: BYTE;

BEGIN
  x := opt DIV 128;         { Right justify 8th bit               }
  val := ORD(x) AND 1;      { Set val to lowest bit of Flop field }
END;   





PROCEDURE WriteRot (VAR opt: OptByte; val: INTEGER);

 { Put value into Rotation field (first 3 bits) of optbyte }

VAR x: BYTE;

BEGIN
  x := WRD(val) AND 7;      { Get 3 lower bits		        }
  opt := opt AND #F8;       { Clear first 3 bits                }
  opt := opt OR x;          { 3 bits from val now in 1st 3 bits }
END;   



PROCEDURE WriteAhead (VAR opt: OptByte; val: BYTE);

 { Put value into Ahead field (4th and 5th bits) of optbyte }

BEGIN
  val := val AND 3;         { Get 2 lower bits 			    }
  val := val * 8;           { Shift left 3 			    }  
  opt := opt AND #E7;       { Clear 4th and 5th bits 		    }
  opt := opt OR val;        { 2 bits from val now in 4th & 5th bits }
END;   



PROCEDURE WriteSize (VAR opt: OptByte; val: INTEGER);

 { Put value into Size field (6th and 7th bits) of optbyte.  If too  }
 { large to fit, decrease Size value and set Ahead field = SizeBias. }

VAR x, ahd: BYTE;

BEGIN
  ReadAhead(opt, ahd);	    { set Ahead to reflect Size value       }

  IF val >= 4 THEN
    BEGIN WriteAhead(opt, SizeBias); val := val - 4 END

  ELSE IF ahd = SizeBias THEN WriteAhead(opt, Solid);

  x := WRD(val) AND 3;      { Get 2 lower bits			    }
  x := x * 32;              { Shift left 5 			    }  
  opt := opt AND #9F;       { Clear 6th and 7th bits		    }
  opt := opt OR x;          { 2 bits from val now in 6th & 7th bits }
END;   



PROCEDURE WriteFlop (VAR opt: OptByte; val: INTEGER);

 { Put value into Flop field (8th bit) of optbyte }

VAR x: BYTE;

BEGIN
  x := WRD(val) AND 1;      { Get lower bit                 }
  x := x * 128;             { Shift left 7                  }  
  opt := opt AND #7F;       { Clear upper bit               }
  opt := opt OR x;          { 1 bit from val now in 8th bit }
END;   





PROCEDURE KeyToOpt (val: BYTE; VAR opt: OptByte);

VAR size: INTEGER;

BEGIN
  opt := StockOpt;
  size := 0;

  WHILE val >= 144 DO
  BEGIN
    val := val - 16;
    size := size + 1;
  END;

  WriteSize(opt, size);

  IF val >= 136 THEN
  BEGIN
    WriteFlop(opt, 1);
    val := val - 8;
  END;

  WriteRot(opt, ORD(val - 128));

END;   (* KeyToOpt *)






  (* -----------  Cursor location procedures  -------------- *)

	    

PROCEDURE GetSign (VAR sgnbase: scrInx; VAR sgninx, sgnnxt: bufInx);

VAR sinx,  lstinx:  bufInx;
    sbase, lstbase: scrInx;
    code:  BYTE;

BEGIN
  lstbase := -1;  lstinx := 0;
  sbase	  :=  8;  sinx	 := LineStart[Line];
  
  code := NewSign;
  
  WHILE (CursX >= sbase) AND (code < NewLine) DO
    BEGIN
      lstbase := sbase;
      lstinx  := sinx;  

      sbase := sbase + ORD((buf^[sinx + 1] * 2)) + SpWidth;
      sinx  := sinx + ScanCode(TRUE, 1, ADR(buf^), sinx);
      code  := buf^[sinx];

    END;
    
  sgnbase := lstbase;
  sgninx  := lstinx;
  sgnnxt  := sinx;
  
END;   (* GetSign *)
	    
	    

PROCEDURE GetSignBound (VAR xpos: scrInx; VAR sgninx: bufInx);

VAR sinx:  bufInx;
    sbase: scrInx;
    code:  BYTE;

BEGIN
  sinx  := LineStart[Line];
  sbase := 8;
  code  := NewSign;
  
  IF buf^[sinx] <> EOF THEN
    
    WHILE (CursX >= sbase) AND (code < NewLine) DO
      BEGIN
	sbase := sbase + ORD((buf^[sinx + 1] * 2)) + SpWidth;
	sinx  := sinx + ScanCode(TRUE, 1, ADR(buf^), sinx);
	code  := buf^[sinx];

      END;
    
  sgninx := sinx;

  IF code < NewLine THEN xpos := sbase
		    ELSE xpos := sbase - SpWidth + 2;
  
END;   (* GetSignBound *)
	    
	    

PROCEDURE GetLineEnd (VAR xpos: scrInx);

VAR junk: bufInx;
    endpos, temp: scrInx;

BEGIN
  temp  := CursX;
  CursX := CursRMar;

  GetSignBound(endpos, junk);

  xpos  := endpos - 2;
  CursX := temp;

END;

	    
	    


	    
  (* --------  Sign/select marker location procedures  --------- *)


	    
PROCEDURE UpdateLineStart (base: bufInx);

VAR i:	    INTEGER;
    linx:   bufInx;
    lineok: BOOLEAN;
    
BEGIN
  linx   := base;
  lineok := TRUE;
  
  FOR i := 0 TO MaxLn DO
  BEGIN
    LineStart[i] := linx;
    
    IF lineok THEN 
      lineok := buf^[linx] <> EOF;
    
    IF lineok THEN
      linx := linx + ScanCode(FALSE, 1, ADR(buf^), linx)
    ELSE
      linx := -1;
  
  END;

END;  (* UpdateLineStart *)



FUNCTION OnScreen (mark: BOOLEAN; inx: bufInx): BOOLEAN;

VAR linx, hinx: bufInx;
    i: INTEGER;

BEGIN
  linx := LineStart[0];
  
  i := MaxLn;

  WHILE (i > 0) AND THEN (LineStart[i] = -1) DO
    i := i - 1;
  
  hinx := LineStart[i];
  
  IF (hinx > -1) AND THEN (buf^[hinx] <> EOF) THEN
    hinx := hinx + ScanCode(FALSE, 1, ADR(buf^), hinx);
  
  IF mark THEN 
    OnScreen := ((inx = linx) AND (inx = Lmark)) OR
		((inx = hinx) AND (inx = Rmark)) OR
		((inx > linx) AND (inx < hinx))

  ELSE 
    OnScreen := (inx >= linx) AND (inx < hinx);
  
END;   (* OnScreen *)

		
		
PROCEDURE GetSignPos (mark: BOOLEAN; inx: bufInx; 
		      VAR xpos, ypos: scrInx; VAR line: linInx);

VAR linx: bufInx;
    lpos: scrInx;
    ln:	  INTEGER;
    
{ Assert: OnScreen(inx) }

BEGIN
  ln := 0;
  
  WHILE (inx > LineStart[ln]) AND (ln < MaxLn) DO
    ln := ln + 1;
    
  IF (inx < LineStart[ln]) OR ((inx = LineStart[ln]) AND
			      mark AND (inx = Rmark)) THEN
    ln := ln - 1;
  
  lpos := 0;
  linx := LineStart[ln];
  
  WHILE linx < inx DO
  BEGIN
    lpos := lpos + ORD((buf^[linx + 1] * 2)) + SpWidth;
    linx := linx + ScanCode(TRUE, 1, ADR(buf^), linx);

  END;
  
  IF (buf^[linx] >= NewLine) AND (lpos > 0) THEN
    lpos := lpos - (SpWidth - 10);
  
  xpos := lpos;
  ypos := LineBase[ln];
  line := ln;
  
END;   (* GetSignPos *)






  (* --------------  Display procedures	 ------------------- *)



(*
      Screen layout:
		
	vertical:	 0..59	     text line 1
			 60	     line boundary 1
			 61..120     text line 2
			 121	     line boundary 2
			 122..181    text line 3
			 182	     line boundary 3

	horizontal:	 8	     line boundary left margin
			 630	     line boundary right margin
*)



PROCEDURE DrawBlock (x1, y1, xSize, ySize, bitval: INTEGER);

VAR x2, y: INTEGER;

BEGIN
  x2 := x1 + xSize - 1;

  FOR y := y1 TO (y1 + ySize - 1) DO 
    DrawLine(x1, y, x2, y, bitval);

END;


       
PROCEDURE QCursDraw;

BEGIN
  QDraw(Cursor, SCursOpt, CursX, CursY - CursYOff, TRUE);
END;



PROCEDURE SCursDraw;

VAR opt: OptByte;
    im: Info;

BEGIN
  opt := CursOpt;
  WriteRot(opt, CursLoc);

  GetInfo(Cursor, opt, im);
  
  QDraw (Cursor, opt, 
	 CursX - ORD(im.FixPtX),
	 CursY - ORD(im.FixPtY),
	 TRUE);
	 
END;   (* SCursDraw *)



PROCEDURE QSymDraw (code, opt: BYTE; xpos, ypos: scrInx);

VAR atr:    Attrs;
    infAr,  infHd: Info;
    ahdopt: OptByte;
    temp:   INTEGER;
    ahd:    BYTE;
    
BEGIN
  QDraw(code, opt, xpos, ypos, FALSE);
  
  GetAttrs(code, atr);
  ahdopt := opt;
  
  ReadAHead(ahdopt, ahd);

  IF (atr.Stype = Movement) AND (ahd <> Open) THEN
  BEGIN
    GetInfo(code, opt, infAr);

    WriteRot(ahdopt, ORD(infAr.AHeadOrient));
    WriteFlop(ahdopt, 0);

    ReadSize(ahdopt, temp);
    IF temp > 1 THEN temp:= 0;
    IF ahd = Hollow THEN temp := temp + 2;
    WriteSize(ahdopt, temp);

    
    GetInfo(AHead, ahdopt, infHd);
    
    Draw (AHead,
	  ahdopt,
	  xpos + ORD(infAr.AHeadPtX) - ORD(infHd.AHeadPtX),
	  ypos + ORD(infAr.AHeadPtY) - ORD(infHd.AHeadPtY));
  
  END;
  
END;   (* QSymDraw *)



PROCEDURE SymDraw (code, opt: BYTE; xpos, ypos: scrInx);

VAR atr:    Attrs;
    infAr,  infHd: Info;
    ahdopt: OptByte;
    temp:   INTEGER;
    ahd:    BYTE;
    
BEGIN
  Draw(code, opt, xpos, ypos);
  
  GetAttrs(code, atr);
  ahdopt := opt;

  ReadAHead(ahdopt, ahd);

  IF (atr.Stype = Movement) AND (ahd <> Open) THEN
  BEGIN
    GetInfo(code, opt, infAr);
    
    WriteRot(ahdopt, ORD(infAr.AHeadOrient));
    WriteFlop(ahdopt, 0);

    ReadSize(ahdopt, temp);
    IF temp > 1 THEN temp:= 0;
    IF ahd = Hollow THEN temp := temp + 2;
    WriteSize(ahdopt, temp);

    
    GetInfo(AHead, ahdopt, infHd);
    
    Draw (AHead,
	  ahdopt,
	  xpos + ORD(infAr.AHeadPtX) - ORD(infHd.AHeadPtX),
	  ypos + ORD(infAr.AHeadPtY) - ORD(infHd.AHeadPtY));
  
  END;
  
END;   (* SymDraw *)



PROCEDURE SymErase (code, opt: BYTE; xpos, ypos: scrInx);

VAR atr:    Attrs;
    infAr,  infHd: Info;
    ahdopt: OptByte;
    temp:   INTEGER;
    ahd:    BYTE;
    
BEGIN
  GetAttrs(code, atr);
  ahdopt := opt;

  ReadAHead(ahdopt, ahd);
  
  IF (atr.Stype = Movement) AND (ahd <> Open) THEN
  BEGIN
    GetInfo(code, opt, infAr);
    
    WriteRot(ahdopt, ORD(infAr.AHeadOrient));
    WriteFlop(ahdopt, 0);

    ReadSize(ahdopt, temp);
    IF temp > 1 THEN temp:= 0;
    IF ahd = Hollow THEN temp := temp + 2;
    WriteSize(ahdopt, temp);

    GetInfo(AHead, ahdopt, infHd);
    
    Erase (AHead,
	   ahdopt,
	   xpos + ORD(infAr.AHeadPtX) - ORD(infHd.AHeadPtX),
	   ypos + ORD(infAr.AHeadPtY) - ORD(infHd.AHeadPtY));
  
  END;
  
  Erase(code, opt, xpos, ypos);

END;   (* SymErase *)



PROCEDURE DrawMark (opt: BYTE; ln: linInx; xpos: scrInx);

BEGIN
  SymDraw(Marker, opt,	 xpos, LineBase[ln] + MrkYOff);
  SymDraw(Marker, opt+1, xpos, LineBase[ln] + MrkYOff + MrkLen);

END;



PROCEDURE ClearMarks;

VAR xpos, ypos: scrInx;
    junk: linInx;

BEGIN
  curmark := -1;
  unmark  := -1;
  
  IF Lmark <> -1 THEN
  BEGIN
    QErase;
  
    IF OnScreen(TRUE, Lmark) THEN
    BEGIN
      GetSignPos(TRUE, Lmark, xpos, ypos, junk);
      DrawBlock(xpos, ypos, MrkWidth, MaxLnHt + 1, 0);

    END;
    
    IF OnScreen(TRUE, Rmark) THEN
    BEGIN
      GetSignPos(TRUE, Rmark, xpos, ypos, junk);
      DrawBlock(xpos, ypos, MrkWidth, MaxLnHt + 1, 0);

    END;
    
    QCursDraw;
  
    Lmark := -1;
    Rmark := -1;
  
  END;
  
END;   (* ClearMarks *)

	    
	    
PROCEDURE DrawBorders;

VAR ln: linInx;
    y:	scrInx;

BEGIN
  FOR ln := 0 TO MaxLn DO
  BEGIN
    y := LineBase[ln] + MaxLnHt + 1;
    DrawLine(7, y, 631, y, 2);
  END;
END;



PROCEDURE DrawText {x, y: scrInx; CONSTS msg: LSTRING};

VAR i:  INTEGER;

BEGIN
  FOR i := 1 TO ORD(msg.len) DO
  BEGIN
    SymDraw(CharCode, WRD(msg[i]), x, y);
    x := x + 8;
  END;
END;



PROCEDURE DispLine (ln: INTEGER);

VAR linx: bufInx;
    code: BYTE;
    sgnbase, sgnend, ybase: scrInx;

BEGIN
  linx := LineStart[ln];

  IF linx = -1 THEN RETURN;
  IF buf^[linx] = EOF THEN RETURN;
  
  sgnbase := 8;
  ybase   := LineBase[ln];
  
  IF linx = Lmark THEN DrawMark(mLeft, ln, 0);
  
  sgnend := sgnbase + ORD(buf^[linx + 1] * 2);
  linx   := linx + 2;
  
  REPEAT
    code := buf^[linx];
    
    IF code < NewSign THEN   (* sym code *)
      BEGIN
	SymDraw(code,
		buf^[linx+1],
		sgnbase + ORD(buf^[linx + 3]) * 2,
		ybase + ORD(buf^[linx + 2]));

	linx := linx + 4;

      END
      
    ELSE IF code = NewSign THEN	  (* put space *)
      BEGIN
	IF NOT FScreen THEN {not print/export}
	  SymDraw(Spacer,
		  SpaceOpt,
		  sgnend,
		  ybase + SpYOff);
	
	sgnbase := sgnend + SpWidth;
	
	IF linx = Lmark THEN DrawMark(mLeft,  ln, sgnbase - 8);
	IF linx = Rmark THEN DrawMark(mRight, ln, sgnbase - 8);
	
	sgnend := sgnbase + ORD(buf^[linx + 1] * 2);

	linx := linx + 2;

      END;
  
  UNTIL code >= NewLine;
  
  IF linx = Rmark THEN DrawMark(mRight, ln, sgnend + 2);
  
END;   (* DispLine *)



PROCEDURE EraseEOL (xpos: scrInx);

BEGIN
  DrawBlock(xpos, LineBase[Line], 640 - xpos, MaxLnHt + 1, 0);
END;



PROCEDURE FloatSym;

VAR linx: bufInx;
    sinfo: Info;
    x1, x2, y1, y2: sgnInx;

BEGIN
  SymErase (buf^[SymInx], buf^[SymInx + 1],
	    SignBase + SymX1, LineBase[Line] + SymY1);
  
  SignLen  := 0;
  SignLen2 := 0;
  linx := SignInx + 2;
  
  WHILE buf^[linx] < NewSign DO
  BEGIN
    
    GetInfo(buf^[linx], buf^[linx + 1], sinfo);
    
    x1 := ORD(buf^[linx + 3]) * 2;
    y1 := ORD(buf^[linx + 2]);
    x2 := x1 + ORD(sinfo.Xsize);
    y2 := y1 + ORD(sinfo.Ysize);
  
    IF x2 > SignLen THEN 
    BEGIN 
      SignLen2 := SignLen; 
      SignLen  := x2;
    END
    ELSE IF x2 > SignLen2 THEN 
      SignLen2 := x2;
     
    IF linx <> SymInx THEN
      
      IF (x1 < SymX2) AND (x2 > SymX1) AND
	 (y1 < SymY2) AND (y2 > SymY1) THEN
	
	 SymDraw (buf^[linx], buf^[linx + 1],
		  SignBase + x1, LineBase[Line] + y1);
    
    linx := linx + 4;
  
  END;
  
  SymLast := SymX2 = SignLen;
  
END;   (* FloatSym *)



PROCEDURE SplitLine (wrap: BOOLEAN; sinx: bufInx; xpos: scrInx);

VAR i, newLn: linInx;
    nonblank: BOOLEAN;

BEGIN
  IF NOT SymSelect THEN QErase;
  
  EraseEOL(xpos);
  IF buf^[sinx] = NewSign THEN buf^[sinx] := NewLine;
  
  newLn := Line + ORD(Line <> MaxLn);
  nonblank := buf^[LineStart[newLn]] <> EOF;
  
  IF Line = MaxLn THEN
    
    BEGIN
      Scrollup;

      FOR i := 0 TO (MaxLn - 1) DO 
	LineStart[i] := LineStart[i + 1];

      Line := MaxLn - 1;
    END
    
  ELSE IF Line < (MaxLn - 1) THEN
    
    BEGIN
      IF nonblank THEN Scrolldown(newLn);

      FOR i := MaxLn DOWNTO (newLn + 1) DO
	LineStart[i] := LineStart[i - 1];
    END
  
  ELSE IF nonblank THEN 
    EraseLine(newLn);
    
  LineStart[newLn] := sinx;
  DispLine(newLn);
  
  IF wrap THEN
  BEGIN
    Line := newLn;
    SignBase := 8;
  END;
  
  IF SymSelect THEN FloatSym;
  
  QCursDraw;
  QErase;
    
END;   (* SplitLine *)
	    
	    




  (* -----------  Command line display procedures  ------------ *)

	    

PROCEDURE DrawPart (bitval: INTEGER);

BEGIN
  IF ODD(bitval) THEN
    IF PartCode <> 255 THEN
      SymDraw(CharCode, PartCode, 0, 192)

    ELSE
      {nothing}

  ELSE
    DrawBlock(0, 192, 7, 8, 0);

END;
	    


PROCEDURE DrawType (bitval: INTEGER);

VAR im: Info;

BEGIN
  IF ODD(bitval) THEN

    IF TMode = Finger THEN 
      IF FHandCode <> 255 THEN
        BEGIN
          GetInfo(FHandCode, FHandOpt, im);  
          SymDraw(FHandCode, FHandOpt, 14, 200 - ORD(im.Ysize));
        END
      ELSE  {nothing}
    
    ELSE IF TMode = Alpha THEN 
      SymDraw(CharCode, ORD('A'), 28, 192)
    
    ELSE  {nothing}
 
  ELSE DrawBlock(14, 184, 42, 16, 0);

END;   (* DrawType *)



PROCEDURE UpdateType;

BEGIN
  DrawType(0);
  DrawType(1);
END;



PROCEDURE DrawGauge {bitval: INTEGER};

VAR needle: INTEGER;

BEGIN
  IF ODD(bitval) THEN
    BEGIN
      DrawLine(67, 189, 67, 199, 1);
      
      IF NOT FScreen THEN
      BEGIN
	needle := 189 + (BufEnd + 50) DIV (MaxBuf DIV 10);
	DrawLine(65, needle, 66, needle, 1);
      END
      
    END
    
  ELSE DrawBlock(65, 189, 3, 11, 0);

END;   (* DrawGauge *)



PROCEDURE UpdateGauge;

BEGIN
  DrawGauge(0);
  DrawGauge(1);
END;



PROCEDURE DrawAlertBox (bitval: INTEGER);

BEGIN
  DrawBlock(0, 189, 61, 11, bitval);
END;



PROCEDURE DrawKBAlertBox (bitval: INTEGER);

BEGIN
  DrawBlock(0, 189, 25, 11, bitval);
END;



PROCEDURE WriteCmdName (CONSTS name: LSTRING);

BEGIN
  DrawBlock(0, 186, 65, 14, 0);
  DrawText(7, 192, name);
END;



PROCEDURE WriteLine {CONSTS line: LSTRING};

BEGIN
  DrawBlock(78, 192, 562, 8, 0);
  DrawText(78, 192, line);
END;





	    
  (* ---------------  Key box display procedures  ---------------- *)



(*
      Key box layout:
		
	vertical:	 175	     box top border
			 176..199    box ysize (= 24)

	horizontal:	 25	     0th  box left border
			 26..73      0th  box xsize (= 48)
			 25+(N*49)   Nth  box left border
			 564	     last box left border
			 613	     last box right border
		       
	NOTE: 10 boxes total; no box at N = 4 or N = 7

*)



PROCEDURE EraseKeys;

VAR box: INTEGER;

BEGIN
  QErase;				     (* zap cursor *)

  box := 0;
  
  WHILE box <= 11 DO
  BEGIN
    DrawBlock(26 + (box * 49), 176, 48, 24, 0);
    box := box + 1 + ORD( (box = 3) OR (box = 6) );
  END;
  
END;   (* EraseKeys *)



PROCEDURE DrawKeyBoxes (bitval: INTEGER);

VAR i:    INTEGER;
    xpos: scrInx;

BEGIN
  QErase;				      (* zap cursor       *)	

  IF ODD(bitval) THEN
    
    BEGIN
      SaveBoxArea(TRUE);		      (* save screen area *)
      EraseKeys;

      DrawBlock( 0, 192, 640,   8, 0);	      (* zap cmd line     *)

      DrawLine( 22, 182, 224, 182, 0);	      (* trim line border *)
      DrawLine(267, 182, 371, 182, 0);	    
      DrawLine(414, 182, 616, 182, 0);	    

      
      DrawLine( 25, 175, 221, 175, 2);	      (* draw box tops    *)
      DrawLine(270, 175, 368, 175, 2);
      DrawLine(417, 175, 613, 175, 2);
      
      FOR i := 0 TO 12 DO		      (* draw box sides   *)
      BEGIN
	xpos := 25 + (i * 49); 
	DrawLine(xpos, 176, xpos, 199, 1);
      END;
	
    END
  
  ELSE SaveBoxArea(FALSE);		      (* restore screen   *)


END;   (* DrawKeyBoxes *)



PROCEDURE DispKeys (kinx: keyInx);

VAR i, j, box, nkeys, nsyms: INTEGER;
    sym,  opt: BYTE;
    sinx: keyInx;
    im:	  Info;

BEGIN
  nkeys := ORD(keymap[kinx]);
  kinx  := kinx + 1;
  box   := 0;
  
  FOR i := 1 to nkeys DO
  BEGIN
    IF keymap[kinx + 1] > 127 THEN   (* unit symbol *)
      
      BEGIN
	sym := keymap[kinx];
	KeyToOpt(keymap[kinx + 1], opt);
	
	GetInfo(sym, opt, im);
	
         j := 24 - ORD(SymYOff + im.Ysize);
         IF j < 0 THEN j := 0;

	SymDraw (sym, opt,
		 26 + (box * 49) + ORD(48 - im.Xsize) DIV 2, 176 + j);
      END
      
    ELSE     (* compound symbol *)
      
      BEGIN
	MoveL(ADR(keymap[kinx]), ADR(sinx), 2);

	nsyms := ORD(keymap[sinx]);
	sinx  := sinx + 1;
	
	FOR j := 1 TO nsyms DO
	BEGIN
	  SymDraw (keymap[sinx], keymap[sinx + 1],
		   26 + (box * 49) + ORD(keymap[sinx + 3]), 
		   176 + ORD(keymap[sinx + 2]) );
	
	  sinx := sinx + 4;
	
	END;
	
      END;  (* else *)
  
    box	 := box + 1;
    kinx := kinx + 2;
    
    IF nkeys > 8 THEN { fill boxes left-to-right }
      BEGIN
	IF box = 4 THEN box := 5
	ELSE IF box = (nkeys - 3) THEN box := 8;
    
	IF i = 4 THEN kinx := kinx + 8
	ELSE IF i = (nkeys - 4) THEN 
	  kinx := kinx - (1 + nkeys + ORD(nkeys = 10));
      END
    
    ELSE { skip middle two boxes }
      IF box = 4 THEN box := 8;
  
  END;	(* for *)


  IF CursSelect THEN SCursDraw		(* restore cursor *)
		ELSE QCursDraw;


END;   (* DispKeys *)



PROCEDURE DispRots (sym, fopt: BYTE; nrots: INTEGER);

VAR box, rot, flop, size: INTEGER;
    opt: OptByte;
    im: Info;
      
BEGIN
  opt := StockOpt;

  ReadFlop(fopt, flop);
  WriteFlop(opt, flop);

  ReadSize(fopt, size);
  WriteSize(opt, size);

  box := 0;
  
  FOR rot := 0 TO nrots DO
  BEGIN
    WriteRot(opt, rot);
    GetInfo(sym, opt, im);

    SymDraw (sym, opt,
	     26 + (box * 49) + ORD(48 - im.Xsize) DIV 2,
             176 + ORD(24 - im.Ysize) DIV 2);

    box := box + 1;

    IF box = 4 THEN box := 8
    ELSE IF box = 12 THEN box := 5;
  
  END;


  IF CursSelect THEN SCursDraw		(* restore cursor *)
		ELSE QCursDraw;


END;   (* DispRots *)






  (* ---------------  Window display procedures	 ---------------- *)



PROCEDURE DrawFWnd (bitval: INTEGER);

CONST wntp = WndTop;
      wnht = WndHt;
      wnbt = wntp + wnht + 1;

BEGIN
  IF ODD(bitval) THEN
    
    BEGIN
      QErase;
    
      DrawBlock( 7, wntp, 625, wnht+2, 0);     (* clear window *)
      
      DrawLine(	 7, wntp, 631, wntp, 1);       (* draw borders *)
      DrawLine(	 7, wnbt, 631, wnbt, 1);
      DrawLine(	 7, wntp,   7, wnbt, 1);
      DrawLine(631, wntp, 631, wnbt, 1);
      
    END
    
  ELSE	(* bitval = 0 *)
  
    BEGIN
      DrawBlock( 7, wntp, 626, wnht+2, 0);    (* clear window	*)
      
      DrawBorders;			      (* restore screen *)
      DispLine(0);
      DispLine(1);
      
      QCursDraw;
    END;
    
END;   (* DrawFWnd *)



PROCEDURE DrawRWnd (bitval: INTEGER);

CONST wntp = WndTop;
      wnht = WndHt;
      wnmd = wntp + wnht + 1;
      wnbt = wnmd + wnht + 1;

BEGIN
  IF ODD(bitval) THEN
    
    BEGIN
      QErase;
    
      DrawBlock( 7, wntp, 625, wnht*2+3, 0);  (* clear window   *)
      
      DrawLine(	 7, wntp, 631, wntp, 1);      (* draw borders   *)
      DrawLine(	 7, wnmd, 631, wnmd, 1);
      DrawLine(	 7, wnbt, 631, wnbt, 1);
      DrawLine(	 7, wntp,   7, wnbt, 1);
      DrawLine(631, wntp, 631, wnbt, 1);
      
    END
    
  ELSE	(* bitval = 0 *)
  
    BEGIN
      DrawBlock( 7, wntp, 626, wnht*2+3, 0);  (* clear window	*)
      
      DrawBorders;			      (* restore screen *)
      DispLine(0);
      DispLine(1);
      DispLine(2);
      
      QCursDraw;
    END;
    
END;   (* DrawRWnd *)



PROCEDURE DrawDWnd (bitval: INTEGER);

CONST wntp = WndTop;
      wnht = WndHt;
      txht = DTxtHt;
      wnmd = wntp + wnht + 1;
      wnbt = wnmd + txht + 1;

BEGIN
  IF ODD(bitval) THEN
    
    BEGIN
      QErase;
    
      DrawBlock( 7, wntp, 625, wnht+txht+3, 0);	 (* clear window   *)
      
      DrawLine(	 7, wntp, 631, wntp, 1);	 (* draw borders   *)
      DrawLine(	 7, wnmd, 631, wnmd, 2);
      DrawLine(	 7, wnbt, 631, wnbt, 1);
      DrawLine(	 7, wntp,   7, wnbt, 1);
      DrawLine(631, wntp, 631, wnbt, 1);
      
    END
    
  ELSE	(* bitval = 0 *)
  
    BEGIN
      DrawBlock( 7, wntp, 626, wnht+txht+3, 0);	 (* clear window   *)
      
      DrawBorders;			         (* restore screen *)
      DispLine(0);
      DispLine(1);
      DispLine(2);
      
      QCursDraw;
    END;
    
END;   (* DrawDWnd *)



PROCEDURE EraseWnd;

BEGIN
  QErase;
  DrawBlock(8, LineBase[Line], 623, WndHt, 0);

END;






  (* -------------  Dictionary text display procedures  ------------ *)



PROCEDURE DWrapText;

VAR inx, ln, lnbrk, txtleft: INTEGER;

BEGIN
  FOR inx := 1 TO DMaxLine DO		      (* init line start array  *)
    Dls[inx] := -1;

  inx := 1;
  ln  := 1;

  REPEAT		      		      (* search for line starts *)
    txtleft := ORD(dtxt.len) - inx + 1;

    IF txtleft > 0 THEN Dls[ln] := inx;

    IF txtleft <= DMaxCol THEN BREAK;


    inx := inx + DMaxCol;

    lnbrk := -ScanEQ(-DMaxCol, ' ', dtxt, inx - 1);

    IF lnbrk < DMaxCol THEN inx := inx - lnbrk;


    ln := ln + 1;

  UNTIL ln > DMaxLine;
  
END;  (* DWrapText *)



FUNCTION DLineLen (line: INTEGER): INTEGER;

VAR curls, nxtls, lnlen: INTEGER;

BEGIN
  curls := Dls[line];

  IF curls < 0 THEN
    BEGIN DLineLen := 0; RETURN END;


  IF (line = DMaxLine) OR ELSE (Dls[line + 1] < 0) THEN
    nxtls := ORD(dtxt.len) + 1

  ELSE 
    nxtls := Dls[line + 1];


  lnlen := nxtls - curls;

  IF lnlen > DMaxCol THEN lnlen := DMaxCol;

  DLineLen := lnlen;


END;  (* DLineLen *)



PROCEDURE DClipText;

BEGIN
  IF DLineLen(DMaxLine) = DMaxCol THEN            (* zap any invisi-text *)
    dtxt.len := WRD(Dls[DMaxLine] + DMaxCol - 1);

  WHILE (dtxt.len > 0) AND THEN (dtxt[ORD(dtxt.len)] = ' ') DO
    dtxt.len := dtxt.len - 1;

END;  (* DClipText *)



PROCEDURE DDrawLine (line: INTEGER);

VAR lstr: cString;

BEGIN
  lstr.len := WRD(DLineLen(line));

  MoveL(ADR(dtxt[Dls[line]]), ADR(lstr[1]), lstr.len);

  DrawText(13, DTxtBase + (line - 1) * 9, lstr);

END;  (* DDrawLine *)



PROCEDURE DEraseLine (line: INTEGER);

BEGIN
  DrawBlock(13, DTxtBase + (line - 1) * 9, 616, 8, 0);

END;  (* DEraseLine *)



PROCEDURE DEraseEOL (line, col: INTEGER);

BEGIN
  DrawBlock(13       + (col  - 1) * 8,
            DTxtBase + (line - 1) * 9, 
	    616      - (col  - 1) * 8,
	    8,
            0);

END;  (* DEraseEOL *)



PROCEDURE DWriteText;

VAR i: INTEGER;

BEGIN
  FOR i := 1 TO DMaxLine DO
  BEGIN
    DEraseLine(i);
    DDrawLine(i);
  END;

END;  (* DWriteText *)



PROCEDURE DWriteWndText;

BEGIN
  DSaveCh := CHR(0);

  DWrapText;
  DWriteText;

END;  (* DWriteWndText *)






    (* ------------  File screen display procedures  ------------- *)


       
PROCEDURE DrawFileBox;

BEGIN
  DrawLine( 67,	 15, 204,  15, 2);
  DrawLine( 67,	 29, 204,  29, 2);
  DrawLine( 67, 175, 204, 175, 2);
  DrawLine( 66,	 15,  66, 175, 1);
  DrawLine(204,	 15, 204, 175, 1);

END;



PROCEDURE DispFileScreen;

BEGIN
  ClearScreen;
  DrawFileBox;
  DrawGauge(1);

  UpdateFileScrn := FALSE;

END;



PROCEDURE WritePathName;

VAR Path: LSTRING(67);

BEGIN
  DrawBlock(68, 16, 136, 13, 0);

  GetCurPath(Path);
  IF Path = NULL THEN Error(Msg[60].msg^);
  
  IF Path.Len > 15 THEN Path.Len := 15;
  DrawText(75, 19, Path);

END;   (* WritePathName *)



PROCEDURE WriteFileNames (base: INTEGER);

VAR inx, adrP: INTEGER;
    fnp: pString;

BEGIN
  DrawBlock(68, 30, 136, 145, 0);
  
  inx := 0;
  adrP := ORD(buf) + (9 * base);
  
  WHILE (inx < 15) AND ((base + inx) <= maxName) DO
  BEGIN
    fnp := Retype(pString,adrP);
    DrawText(75, 35 + (9 * inx), fnp^);
    inx := inx + 1;
    adrP := adrP + 9;
  END;

END;   (* WriteFileNames *)



PROCEDURE DispFileList;

Const FindFirstFile = 78;
      FindNextFile  = 79;
      NameOff       = 30;
      WildCard      = '*' * FileExt * Chr (0);
      Dot           = '.';

VAR I: WORD;
    FileData: ADSMEM;
    CRCXQQ [Extern] : WORD;
    Status : Byte;

BEGIN
  curName :=  0;
  maxName := -1;

  FileData := GetDTA;
  CRCXQQ := 0;
  Status := DOSXQQ(FindFirstFile, (ADS WildCard).R);
  
  WHILE Status = 0 DO
  BEGIN
    maxName := maxName + 1;

    FOR I := 1 TO 8 DO
      IF CHR(FileData^[I + NameOff - 1]) <> Dot THEN
        BEGIN
          buf^[maxName * 9] := I;
          buf^[(maxName * 9) + ORD(I)] := LoCase(FileData^[I + NameOff - 1]);
        END
      ELSE
        Break;
      
    Status := DOSXQQ(FindNextFile, (ADS WildCard).R);

  END;  (* while *)

  WritePathName;
  WriteFileNames(0);

  UpdateFileList := FALSE;

END;   (* DispFileList *)




    
    
  (* ----------  Command interpretation procedures  ----------- *)



FUNCTION NoCharVal (r: InxRec): BOOLEAN;

BEGIN
  NoCharVal := (r.sym.rotval <= 127) OR     (* keymap index *)
               (r.sym.codeval = 255);       (* or undefined *)
END;



FUNCTION SearchCmd (inx: msgInx; sch: CHAR): cmdVal;

VAR c: cmdPtr;

BEGIN
  SearchCmd := -1;
  IF (inx < 1) OR (inx > MaxMsg) THEN RETURN;

  c   := Msg[inx].cmd;
  sch := UpCase(sch);

  WHILE (c <> NIL) AND THEN (c^.ch <> sch) DO
    c := c^.nxt;

  IF c <> NIL THEN SearchCmd := c^.val;

END;  (* SearchCmd *)



FUNCTION GetCmdVal (ch: CHAR): cmdVal;

VAR val: cmdVal;

BEGIN
  val := SearchCmd(1, ch);

  IF val < 0 THEN   (* check alt cmds *)
  BEGIN
    val := SearchCmd(2, ch);
    IF val >= 0 THEN val := val + 8;
  END;

  GetCmdVal := val;
END;






  (* -------------  Keyboard input procedures  ------------- *)



PROCEDURE GetKey (VAR code: INTEGER; VAR shift: ShiftState);

VAR s: WORD;

BEGIN
  WHILE NOT KeyRead(code, s) DO;

  IF (s AND sShft) <> 0 THEN
    
    IF      (s AND sCtrl) <> 0 THEN shift := CtrlShft
    ELSE IF (s AND sAlt ) <> 0 THEN shift := ShftAlt
    ELSE IF (s AND sCap ) <> 0 THEN shift := None
    ELSE                            shift := Shft

  ELSE IF (s AND (sCtrl+sAlt)) = (sCtrl+sAlt) THEN 
    shift := CtrlAlt
    
  ELSE IF (s AND sCtrl) <> 0 THEN shift := Ctrl
  ELSE IF (s AND sAlt ) <> 0 THEN shift := Alt
  ELSE IF (s AND sCap ) <> 0 THEN shift := Shft
  ELSE                            shift := None;
  
END;  (* GetKey *)



PROCEDURE GetDead (r: InxRec; VAR getch: CHAR);

VAR i, kinx, nkeys: INTEGER;
    state: ShiftState;
    code:  INTEGER;
    gotch: BOOLEAN;
    ch:    CHAR;

BEGIN
  getch := CHR(255);   (* undefined *)

  nkeys := ORD(keymap[r.inx]);
  kinx  := r.inx + 1;

  GetKey(code, state);
  
  FOR i := 1 to nkeys DO
  BEGIN
    IF (code  = ORD(keymap[kinx])) AND
       (state = Retype(ShiftState, keymap[kinx + 1])) THEN
    BEGIN
      getch := CHR(keymap[kinx + 2]);
      RETURN;
    END;
      
    kinx := kinx + 4;
  END;

END;   (* GetDead *)



FUNCTION CtrlCmd (code: INTEGER; state: ShiftState): BOOLEAN;

BEGIN
  CtrlCmd := (code IN [#47, #48, #4B, #4D, #50]) AND (state = Ctrl);
END;



PROCEDURE GetChar (cmd: msgInx; okset: SetOfChar; 
                   VAR getch: CHAR; VAR code: INTEGER);

VAR gotch: BOOLEAN;
    state: ShiftState;
    ch:    CHAR;
    r:     InxRec;

BEGIN
  gotch := FALSE;

  REPEAT
    GetKey(code, state);
    r := Keyboard^[Alpha, code, state];

    IF r.sym.rotval > 127 THEN  (* live key *)
   
      IF r.sym.codeval < 255 THEN  (* defined *)
        BEGIN
          ch := CHR(r.sym.codeval);
          gotch := TRUE;
        END
      ELSE {nothing}

    ELSE     (* dead key *)
      BEGIN
        GetDead(r, ch);
        gotch := ch <> CHR(255);
      END;

    IF gotch THEN
      gotch := (ch IN okset) OR (SearchCmd(cmd, ch) >= 0);

  UNTIL gotch;

  getch := ch;

END;   (* GetChar *)



PROCEDURE GetString (CONSTS msg: LSTRING; CONST okset: SetOfChar;
                     maxlen: INTEGER; upper: BOOLEAN; VAR s: LSTRING);
VAR ch:	  CHAR;
    cs:	  LSTRING(1);
    sinx: scrInx;
    junk: INTEGER;

BEGIN
  WriteLine(msg);
  UpdateCmdList := TRUE;
  
  sinx := 85 + ORD(msg.len * 8);
  SymDraw(CharCode, CharCurs, sinx, 192);

  cs    := ' ';
  s.len := 0;

  REPEAT
    GetChar(-1, okset + [CHR(bs),CHR(esc),CHR(ret)], ch, junk);
    
    IF (ch IN okset) AND (ORD(s.len) < maxlen) THEN
      BEGIN
	SymErase(CharCode, CharCurs, sinx, 192);
	SymDraw(CharCode, WRD(ch), sinx, 192);
	
	IF upper THEN ch := UpCase(ch);
	
	cs[1] := ch;
	Concat(s,cs);
	sinx := sinx + 7;
	
	IF ORD(s.len) < maxlen THEN
	  SymDraw(CharCode, CharCurs, sinx, 192);
      
      END
      
    ELSE IF (ch = CHR(bs)) AND (s.len > 0) THEN
      BEGIN
	SymErase(CharCode, CharCurs, sinx, 192);
	sinx := sinx - 7;

	SymDraw(CharCode, CharCurs, sinx, 192);
	Delete(s, ORD(s.len), 1);
      END 
      
    ELSE IF ch = CHR(esc) THEN
      BEGIN
	ch    := CHR(ret);
	s.len := 0;
      END;
    
  UNTIL ch = CHR(ret);


END;   (* GetString *)



PROCEDURE Prompt (inx: msgInx; okset: SetOfChar; CONSTS msg: LSTRING;
                                        VAR ch: CHAR; VAR cmd: cmdVal);
VAR junk: INTEGER;

BEGIN
  WriteLine(msg);
  UpdateCmdList := TRUE;

  GetChar(inx, okset + [CHR(esc)], ch, junk);
  cmd := SearchCmd(inx, ch);
END;






  (* ----------------  Error procedures	 ------------------- *)

	    

PROCEDURE FlashAlertBox (boxup: BOOLEAN);

BEGIN
  IF boxup THEN DrawKBAlertBox(1)
           ELSE DrawAlertBox(1);

  Pause(FALSE, AlertBoxDelay);

  IF boxup THEN DrawKBAlertBox(0)
           ELSE BEGIN DrawAlertBox(0); DrawType(1) END;
  
END;   (* FlashAlertBox *)



PROCEDURE Error {CONSTS errmsg: LSTRING};

VAR s:     cString;
    junk:  CHAR;
    junk2: INTEGER;

BEGIN
  DrawAlertBox(1);
  CopyLST(errmsg, s);
  Concat(s, Msg[33].msg^);

  Prompt(-1, [CHR(ret)], s, junk, junk2);
  DrawAlertBox(0);

END;  (* Error *)



PROCEDURE BuildMsg (VAR msg: LSTRING; CONSTS base, name: LSTRING);

VAR linx, rinx: INTEGER;

BEGIN
  CopyLST(base, msg);

  linx := Positn('<', msg, 1);
  rinx := Positn('>', msg, 1);

  IF (linx > 0) AND (rinx > linx) THEN
  BEGIN
    Delete(msg, linx, rinx - linx + 1);
    Insert(name, msg, linx);
  END;

END;  (* BuildMsg *)




	    
	    
  (* -----------  Cursor movement procedures  -------------- *)

	    

PROCEDURE LMoveX (dx: INTEGER);

VAR nx: scrInx;

BEGIN
  nx := CursX + dx;

  IF nx < CursWMar THEN nx := CursWMar;
  IF nx > CursRMar THEN nx := CursRMar;

  CursX := nx;
  QErase;
  QCursDraw;
  
END;   (* LMoveX *)



PROCEDURE LMoveY (dy: INTEGER);

VAR ny, ybase: scrInx;

BEGIN
  ny := CursY + dy;
  ybase := LineBase[Line];
  
  IF ny < ybase           THEN ny := ybase;
  IF ny > (ybase+MaxLnHt) THEN ny := ybase + MaxLnHt;

  CursY := ny;
  QErase;
  QCursDraw;
  
END;   (* LMoveY *)



PROCEDURE CScrollup (smallmove: BOOLEAN);

VAR linx: bufInx;
    i:	  linInx;

BEGIN
  linx := LineStart[MaxLn];
  
  IF buf^[linx] = EOF THEN FlashAlertBox(FALSE)

  ELSE
    BEGIN
      FOR i := 0 TO (MaxLn - 1) DO
        LineStart[i] := LineStart[i + 1];
    
      LineStart[MaxLn] := linx + ScanCode(FALSE, 1, ADR(buf^), linx);
    
      QErase;
      Scrollup;
      DispLine(MaxLn);
    
      IF smallmove THEN CursY := LineBase[MaxLn];
      CursX := NewX;

      QCursDraw;
    
    END;

END;   (* CScrollup *)



PROCEDURE CScrolldown (smallmove: BOOLEAN);

VAR linx: bufInx;
    i:	  linInx;
    
BEGIN
  linx := LineStart[0];

  IF linx = 0 THEN FlashAlertBox(FALSE)

  ELSE
    BEGIN
      FOR i := MaxLn DOWNTO 1 DO
        LineStart[i] := LineStart[i - 1];
  
      LineStart[0] := linx + ScanCode(FALSE, -1, ADR(buf^), linx);
   
      QErase;
      Scrolldown(0);
      DispLine(0);
  
      IF smallmove THEN CursY := LineBase[0] + MaxLnHt;
      CursX := NewX;

      QCursDraw;
  
    END;

END;   (* CScrolldown *)



PROCEDURE CMoveX (dx: INTEGER);

BEGIN
  NewY := CursY;
  NewX := CursX + dx;
  
  IF NewX < CursLMar THEN
    BEGIN
      NewX := CursRMar;

      IF Line = 0 THEN
	BEGIN
	  CScrolldown(FALSE);
	  RETURN;
	END

      ELSE
	BEGIN
	  NewY := CursY - (MaxLnHt + 2);
	  Line := Line - 1;
	END

    END
    
  ELSE IF NewX > CursRMar THEN
    BEGIN
      NewX := CursLMar;

      IF Line = MaxLn THEN
	BEGIN
	  CScrollup(FALSE);

	  RETURN;

	END

      ELSE IF LineStart[Line + 1] = -1 THEN
	RETURN

      ELSE
	BEGIN
	  NewY := CursY + (MaxLnHt + 2);
	  Line := Line + 1;
	END;

    END;


  QErase;
  CursX := NewX;
  CursY := NewY;

  QDraw(Cursor, SCursOpt, CursX, CursY - CursYOff, TRUE);

  
END;   (* CMoveX *)



PROCEDURE CMoveY (dy: INTEGER);

VAR NewLin: linInx;

BEGIN
  NewX := CursX;
  NewY := CursY + dy;
  
  IF (NewY = (LineBase[0] + MaxLnHt + 1)) OR
     (NewY = (LineBase[1] + MaxLnHt + 1)) THEN  {cross border}
    NewY := NewY + dy;

  
  IF NewY <= (LineBase[0] + MaxLnHt) THEN  {on top line}
  
    IF NewY < LineBase[0] THEN	{off the top}
      BEGIN
	CScrolldown(ABS(dy) <= 4);

	RETURN;

      END

    ELSE NewLin := 0
    
  ELSE	{below top line}
    
    IF NewY > (LineBase[2] + MaxLnHt) THEN  {off the bottom}
      BEGIN
	CScrollup(ABS(dy) <= 4);

	RETURN;

      END

    ELSE NewLin := 1 + ORD(NewY >= LineBase[2]);

    
  IF LineStart[NewLin] = -1 THEN FlashAlertBox(FALSE)

  ELSE
    BEGIN
      Line := NewLin;

      QErase;
      CursX := NewX;
      CursY := NewY;

      QDraw(Cursor, SCursOpt, CursX, CursY - CursYOff, TRUE);

    END;
  

END;   (* CMoveY *)



PROCEDURE CEndLine;

VAR xpos: scrInx;

BEGIN
  GetLineEnd(xpos);
  CMoveX(xpos - CursX);

END;



PROCEDURE CTab;

VAR sinx: bufInx;
    xpos: scrInx;

BEGIN
  CursX := CursX + CursTab;

  GetSignBound(xpos, sinx);

  IF buf^[LineStart[Line]] = EOF THEN 	(* move to line start *)
    xpos := CursLMar + CursTab

  ELSE IF CursX >= xpos THEN 	  	(* wrap to next line  *)
    xpos := 10000

  ELSE IF (CursX > CursLMar) AND        (* move to line end   *)
          (buf^[sinx] >= NewLine) THEN
    xpos := xpos + CursTab - 2; 

  CMoveX(xpos - CursX - CursTab);

END;   (* CTab *)



PROCEDURE CBackTab;

VAR base,  next: bufInx;
    xpos:  scrInx;
    front: BOOLEAN;

BEGIN
  front := LineStart[Line] = 0;

  GetSign(xpos, base, next);

  IF xpos < 0 THEN xpos := 0	  	(* wrap to prev line  *)

  ELSE IF (buf^[base] >= NewLine) THEN	(* move to line start *)
    xpos := xpos + CursTab - 2; 

  CMoveX(xpos - CursX - CursTab);

  IF (xpos = 0) AND NOT front THEN	(* move to line end   *)
    CEndLine;

END;   (* CBackTab *)



PROCEDURE CNewPage (base: bufInx);

VAR i: INTEGER;
    
BEGIN
  UpdateLineStart(base);

  QErase;
  
  FOR i := 0 TO MaxLn DO
    EraseLine(i);
    
  FOR i := 0 TO MaxLn DO
    DispLine(i);
    
  QCursDraw;
    
END;   (* CNewPage *)



PROCEDURE CBeginning;

BEGIN
  Line := 0;

  CursX := CursLMar;
  CursY := CursHomeY;

  CNewPage(0);
END;



PROCEDURE CEnd;

VAR linx: bufInx;
    cl: linInx;

BEGIN
  linx := bufEnd - 2;	    (* points to EOF code *)
  cl := 0;
  
  WHILE (linx <> 0) AND (cl < MaxLn) DO
  BEGIN
    linx := linx + ScanCode(FALSE, -1, ADR(buf^), linx);
    cl := cl + 1;
  END;
  
  Line := cl - ORD(cl <> 0);

  CursX := CursLMar + 2;
  CursY := LineBase[Line] + CursHomeY;

  CNewPage(linx);
  CEndLine;
  
END;   (* CEnd *)



PROCEDURE CPageUp;

VAR linx: bufInx;
    i:	  INTEGER;

BEGIN
  linx := LineStart[0];
  i := 0;
  
  WHILE (linx <> 0) AND (i < (MaxLn + 1)) DO
  BEGIN
    linx := linx + ScanCode(FALSE, -1, ADR(buf^), linx);
    i := i + 1;
  END;
  
  IF i = 0 THEN		       (* already on 1st page *)
    BEGIN
      FlashAlertBox(FALSE);

      QErase;

      Line  := 0;
      CursX := CursLMar;
      CursY := CursHomeY;

      QCursDraw;
    END
  
  ELSE
    CNewPage(linx);

END;   (* CPageUp *)



PROCEDURE CPageDown;

VAR linx: bufInx;
    last: BOOLEAN;
    cl:	  INTEGER;

BEGIN
  linx := LineStart[MaxLn];
  last := TRUE;
  
  IF linx <> -1 THEN
    IF buf^[linx] <> EOF THEN
    BEGIN
      last := FALSE;
      linx := linx + ScanCode(FALSE, 1, ADR(buf^), linx);

      CNewPage(linx);
    END;
    
  cl := 0;
  WHILE (cl < (MaxLn + 1)) AND (buf^[LineStart[cl]] <> EOF) DO 
    cl := cl + 1;
  
  IF last OR (Line > cl) THEN    (* already on 1st page *)
  BEGIN
    FlashAlertBox(FALSE);

    QErase;

    Line := cl;
    CursX := CursLMar + 2;
    CursY := LineBase[Line] + CursHomeY;

    QCursDraw;
  END;
  
END;   (* CPageDown *)



PROCEDURE CSign (sinx: bufInx);

VAR linx: bufInx;
    cnt:  INTEGER;

BEGIN
  IF NOT OnScreen(FALSE, sinx) THEN
  BEGIN
    linx := sinx;
    
    cnt  := 1 + ORD(buf^[linx] < NewLine);
    
    WHILE (linx > 0) AND (cnt > 0) DO	  (* back up from sign *)
    BEGIN
      linx := linx + ScanCode(FALSE, -1, ADR(buf^), linx);
      cnt := cnt - 1;
    END;
    
    CursX := CursLMar;		(* transitory cursor pos *)
    CursY := CursHomeY;

    CNewPage(linx);
    
  END;
  
  QErase;

  GetSignPos(FALSE, sinx, CursX, CursY, Line);

  IF CursX < CursLMar THEN CursX := CursLMar;
  CursY := CursY + CursHomeY;

  QCursDraw;
  
END;   (* CSign *)




	    
	    
  (* -----------  Dictionary text cursor procedures  ------------ *)



PROCEDURE DDrawCurs;

BEGIN
  QDraw(CharCode, CharCurs, 13       + (DCursCol  - 1) * 8,
        		    DTxtBase + (DCursLine - 1) * 9, 
			    TRUE);
END;  (* DDrawCurs *)



PROCEDURE DShowCurs;

VAR ln: INTEGER;

BEGIN
  ln := DMaxLine;

  WHILE (ln > 1) AND (DLineLen(ln) = 0) DO ln := ln - 1;

  DCursLine := ln;

  DCursCol  := 1 + DLineLen(DCursLine);
   
  DDrawCurs;

END;  (* DShowCurs *)



PROCEDURE DMoveCursX (dx: INTEGER);

VAR nline, ncol: INTEGER;

BEGIN
  nline := DCursLine;
  ncol  := DCursCol + dx;
  
  IF ncol < 1 THEN

    IF DCursLine = 1 THEN RETURN
    ELSE
      BEGIN
        nline := DCursLine - 1;
        ncol  := 1 + DLineLen(nline);
      END
    
  ELSE IF ncol > (1 + DLineLen(DCursLine)) THEN

    IF (DCursLine = DMaxLine) OR ELSE
       (DLineLen(DCursLine + 1) = 0) THEN RETURN
    ELSE
      BEGIN
      	nline := DCursLine + 1;
        ncol  := 1;
      END;

  QErase;

  DCursLine := nline;
  DCursCol  := ncol;

  DDrawCurs;
  
END;   (* DMoveCursX *)



PROCEDURE DMoveCursY (dy: INTEGER);

VAR nline, ncol: INTEGER;

BEGIN
  nline := DCursLine + dy;
  ncol  := DCursCol;

  IF ((nline < 1) OR (nline > DMaxLine)) OR ELSE
     (DLineLen(nline) = 0) THEN RETURN

  ELSE IF ncol > (1 + DLineLen(nline)) THEN
    ncol := 1 + DLineLen(nline);

  QErase;

  DCursLine := nline;
  DCursCol  := ncol;

  DDrawCurs;
  
END;   (* DMoveCursY *)



PROCEDURE DGetCursInx (VAR inx: INTEGER);

BEGIN
  IF DLineLen(DCursLine) = 0 THEN
    inx := 1 + ORD(dtxt.len) 

  ELSE
    inx := Dls[DCursLine] + DCursCol - 1;

END;  (* DGetCursInx *)



PROCEDURE DPutCursInx (inx: INTEGER);

VAR ln, col: INTEGER;

BEGIN
  ln  := 1;
  col := 1;

  WHILE (ln <= DMaxLine) AND THEN
        (Dls[ln] <> -1)  AND THEN
        (inx >= Dls[ln])  DO ln := ln + 1;

  IF ln > 1 THEN ln := ln - 1;

  IF Dls[ln] <> -1 THEN col := Dls[ln];

  DCursLine := ln;
  DCursCol  := inx - col + 1;

END;  (* DPutCursInx *)


	    



  (* ------------  Cursor/symbol blink procedures  -------------- *)

  

PROCEDURE BlinkCursor (wait: BOOLEAN);
	
VAR on:    BOOLEAN;
    ticks: INTEGER;
    
BEGIN
  on := TRUE;
  
  IF wait THEN ticks := CursDelay ELSE ticks := 0;

  REPEAT
    Pause(TRUE, ticks);

    on := NOT on;
  
    IF on THEN
      BEGIN
        ticks := CursOn;
        IF CursSelect THEN SCursDraw
      		      ELSE QCursDraw;
      END

    ELSE 
      BEGIN
        ticks := CursOff;
        QErase;
      END;
    
  UNTIL KeyAvail;
  
  IF NOT on THEN
    IF CursSelect THEN SCursDraw
		  ELSE QCursDraw;

END;   (* BlinkCursor *)


  
PROCEDURE BlinkSymbol;
	
VAR sym, opt:   BYTE;
    xpos, ypos: scrInx;
    on:	        BOOLEAN;

BEGIN
  sym  := buf^[SymInx]; 
  opt  := buf^[SymInx + 1];
  xpos := SignBase + SymX1; 
  ypos := LineBase[Line] + SymY1;
  
  on := TRUE;
  
  REPEAT
    Pause(FALSE, SymFreq);

    on := NOT on;
      
    IF on THEN QSymDraw(sym, opt, xpos, ypos)
    ELSE QErase;
    
  UNTIL KeyAvail;
  
  IF NOT on THEN QSymDraw(sym, opt, xpos, ypos);
  
END;   (* BlinkSymbol *)






  (* --------------  SignWriter file procedure  ---------------- *)



FUNCTION SWOpen {VAR f: FCBFQQ; CONST name: LSTRING): BOOLEAN};

CONST maxpath     = 80;
      find1stfile = 16#4E;

VAR i:                WORD;  
    path:             ADSMEM;
    CRCXQQ[Extern]:   WORD;
    fndfile, fndpath: BOOLEAN;
    fullname:         LSTRING(maxpath + 12);

BEGIN
  fndpath := FALSE;

  path   := GetDTA;
  path.r := 44;         (* 44 = offset into program segment of segment    *)

  path.s := WRD(path^[0]) + (WRD(path^[1]) * 16#100);  (* address of path *)

  path.r := 0;          (* Path now points to start of environment space  *)
  
  WHILE NOT fndpath DO
  BEGIN
    IF (path^[0] = 0) And (path^[1] = 0) THEN BREAK;
  
    IF path^[0] = WRD('P') THEN
      IF path^[1] = WRD('A') THEN
        IF path^[2] = WRD('T') THEN
          IF path^[3] = WRD('H') THEN
            IF path^[4] = WRD('=') THEN
              BEGIN
                fndpath := TRUE;
                path.r := path.r + 5;
              END
            ELSE
              path.r := path.r + 4
          ELSE
            path.r := path.r + 3
        ELSE
          path.r := path.r + 2
      ELSE
        path.r := path.r + 1
    ELSE
      path.r := path.r + 1;

  END;  (* while *)


  fndfile := FALSE;

  CopyLST(name, fullname);

  WHILE NOT fndfile DO
  BEGIN
    CRCXQQ := 0;

    fullname.Len := fullname.Len + 1;
    fullname[fullname.Len] := CHR(0);

    fndfile := (DOSXQQ(find1stfile, (ADS fullname[1]).r) = 0);

    IF NOT fndpath THEN BREAK;

    IF NOT fndfile THEN
    BEGIN
      FOR i := 0 TO maxpath - 1 DO
      BEGIN
        IF path^[0] = 0 THEN
        BEGIN
          fndpath := FALSE;
          BREAK;
        END;

        IF path^[0] = WRD(';') THEN
        BEGIN
          path.r := path.r + 1;
          BREAK;
        END;

        fullname.Len := i + 1;
        fullname[i + 1] := CHR(path^[0]);
        path.r := path.r + 1;

      END;  (* for *)

      IF fullname[fullname.Len] <> '\' THEN
        Concat(fullname, '\');
      
      Concat(fullname, name);

    END;  (* if *)

  END;  (* while *)

  IF fndfile THEN
  BEGIN
    f.Mode := Direct;
    f.Errs := 0;
    f.Trap := TRUE;
    fullname.Len := fullname.Len - 1;

    Assign(f, fullname);
    Reset(f);

    fndfile := (f.Errs = 0);

  END;
  
  SWOpen := fndfile;


END;   (* SWOpen *)






  (* ---------------  Command loop procedures  --------------- *)



PROCEDURE FGetCommand;
    
BEGIN
  FExit := FALSE;

  REPEAT
    IF UpdateFileScrn THEN DispFileScreen;
    IF UpdateFileList THEN DispFileList;
    
    IF UpdateCmdList THEN
    BEGIN
      WriteCmdName(NULL);
      WriteLine(Msg[3].msg^);
      UpdateCmdList := FALSE;
    END;

    Command;

  UNTIL FExit;  

END;   (* FGetCommand *)

  
 
PROCEDURE EGetCommand;
    
BEGIN
  EExit := FALSE; 
  
  REPEAT
    IF SymSelect THEN BlinkSymbol
		 ELSE BlinkCursor(TRUE);

    Command;   

    IF UpdateCmdList THEN
    BEGIN
      WriteCmdName(NULL);	
      DrawType(1);
      WriteLine(Msg[1].msg^);
      altCmd := FALSE;
      UpdateCmdList := FALSE;
    END;

  UNTIL EExit;

END;   (* EGetCommand *)



PROCEDURE WGetCommand;
    
BEGIN
  WExit := FALSE; 

  REPEAT
    IF SymSelect THEN BlinkSymbol
    ELSE IF NOT DTxtWnd THEN BlinkCursor(TRUE);

    Command;    
    
    IF UpdateCmdList THEN
    BEGIN
      WriteCmdName(NULL);	
      DrawType(1);
      WriteLine(Msg[1].msg^);
      altCmd := FALSE;
      UpdateCmdList := FALSE;
    END;
  
  UNTIL WExit;

END;   (* WGetCommand *)




PROCEDURE Command;


LABEL EndOfCmd;


PROCEDURE CmdExit; FORWARD;


PROCEDURE DictUpdate; FORWARD;


FUNCTION DLookup (CONST name: DName;
                  VAR vol, entry: INTEGER): BOOLEAN; FORWARD;


PROCEDURE DAddChar (ch: CHAR); FORWARD;






     (* --------------	Buffer manipulation procedures	------------ *)



PROCEDURE MakeSpace (binx: bufInx; nbytes: INTEGER);

VAR ln: linInx;

BEGIN
  IF nbytes > (MaxBuf - bufEnd) THEN
  BEGIN
    Error(Msg[12].msg^);
    CmdExit;
  END;
  
  IF binx < bufEnd THEN
    MoveR(ADR(buf^[binx]), ADR(buf^[binx + nbytes]),
          WRD(bufEnd - binx));
  
  bufEnd := bufEnd + nbytes;
  
  FOR ln := (Line + 1) TO MaxLn DO
    IF LineStart[ln] <> -1 THEN
      LineStart[ln] := LineStart[ln] + nbytes;
    
END;   (* MakeSpace *)



PROCEDURE DeleteCode (binx: bufInx; nbytes: INTEGER);

VAR ln: linInx;

BEGIN
  MoveL(ADR(buf^[binx + nbytes]), ADR(buf^[binx]),
	WRD(bufEnd - (binx + nbytes)));
  
  bufEnd := bufEnd - nbytes;
  
  FOR ln := (Line + 1) TO MaxLn DO
    IF LineStart[ln] <> -1 THEN
      LineStart[ln] := LineStart[ln] - nbytes;

END;   (* DeleteCode *)






  (* ----------------  Disk I/O procedures  ---------------- *)



PROCEDURE OpenFile (VAR f: FCBFQQ; CONST name: LSTRING; new: BOOLEAN);

VAR s:    cString;
    rslt: WORD;

BEGIN
  CopyLST(name, s);

  IF Positn('.', name, 1) = 0 THEN
    Concat(s, FileExt);

  f.mode := Direct;
  f.trap := TRUE;
  Assign(f, s);

  IF new THEN
    BEGIN
      Rewrite(f);
      rslt := f.errs;

    END
  ELSE 
    BEGIN
      Reset(f);
      rslt := f.errs;

      Seek(f,1);

    END;
  
  IF rslt > 0 THEN
  BEGIN
    IF rslt = iNoFile THEN
      BEGIN
        BuildMsg(s, Msg[34].msg^, name);
        Error(s);
      END

    ELSE IF rslt = iNoRoom THEN
      Error(Msg[35].msg^)
      
    ELSE
      BEGIN
        BuildMsg(s, Msg[13].msg^, name);
        Error(s);
      END;    

    CmdExit;
  END;

END;   (* OpenFile *)



PROCEDURE CloseFile (VAR f: FCBFQQ; CONSTS name: LSTRING; save: BOOLEAN);

VAR s: cString;

BEGIN
  IF save THEN Close(f)
  ELSE Discard(f);
  
  IF f.errs > 0 THEN
  BEGIN
    BuildMsg(s, Msg[36].msg^, name);
    Error(s);
    CmdExit;
  END;
  
END;   (* CloseFile *)



PROCEDURE DeleteFile (CONST name: fName);

VAR f: blkFile;

BEGIN
  OpenFile (f, name, TRUE);
  CloseFile(f, name, FALSE);

END;   (* DeleteFile *)



PROCEDURE CheckFile (CONST name: fName);

VAR s:    cString;
    f:    blkFile;
    junk: CHAR;
    cmd:  cmdVal;

BEGIN
  CopyLST(name, s);
  Concat(s, FileExt);

  f.trap := True;
  Assign(f, s);
  Reset(f);
  
  IF f.errs > 0 THEN

    IF f.errs <> iNoFile THEN
      BEGIN
        BuildMsg(s, Msg[13].msg^, name);
	Error(s);
	CmdExit;
      END
    ELSE  {no file}

  ELSE	 (* file exists *)
 
    BEGIN
      BuildMsg(s, Msg[14].msg^, name);
      Prompt(14, [], s, junk, cmd);

      IF cmd <> C_Yes THEN
        BEGIN
          CloseFile(f, name, TRUE);
      	  CmdExit;
        END
      ELSE CloseFile(f, name, FALSE);

    END;

END;   (* CheckFile *)



PROCEDURE IOCheck (CONSTS name: LSTRING; read: BOOLEAN; VAR f:FCBFQQ);

VAR s: cString;

BEGIN
  IF f.errs > 0 THEN
    BEGIN
      Close(f);

      IF read THEN BuildMsg(s, Msg[37].msg^, name)
              ELSE BuildMsg(s, Msg[38].msg^, name);

      Error(s);

      CmdExit;
    END;

END;   (* IOCheck *)



FUNCTION BadVersion (code: BYTE): BOOLEAN;

BEGIN
  BadVersion := (code <> SWversion) AND     {4.1 reads 4.0 files}
     		(code <> (SWversion - 1));

END;  (* BadVersion *)



PROCEDURE GetFileInfo (VAR f: blkFile; CONST name: fName; 
		       VAR mrk1, mrk2, len: bufInx);

VAR i: INTEGER;

BEGIN
  i := ORD(GETUQQ(f, 512, ADS(altbuf^)));
  IOCheck(name, TRUE, f);
  
  IF BadVersion(altbuf^[0]) OR (altbuf^[1] <> SWhost) THEN
  BEGIN
    Error(Msg[39].msg^);
    Close(f); CmdExit;
  END;
  
  MoveL(ADR(altbuf^[3]), ADR(mrk1), 2);
  MoveL(ADR(altbuf^[5]), ADR(mrk2), 2);
  MoveL(ADR(altbuf^[7]), ADR(len),  2);
  
  IF len > Maxbuf THEN
  BEGIN
    Error(Msg[40].msg^);
    Close(f); CmdExit;
  END;
  
END;   (* GetFileInfo *)






  (* -------------  Symbol & text selection  --------------- *)
	    


PROCEDURE SelSym (sginx: bufInx; sgbase: scrInx);

VAR cx, cy: sgnInx;
    x1, x2, y1, y2: sgnInx;
    ssize, minsize: INTEGER;
    sinfo: Info;
    linx: bufInx;
    
BEGIN
  cx := CursX - sgbase;
  cy := CursY - LineBase[Line];
  
  minsize := 10000;
  linx := sginx + 2;
  
  WHILE buf^[linx] < NewSign DO
  BEGIN
    
    GetInfo(buf^[linx], buf^[linx + 1], sinfo);
      
    x1 := ORD(buf^[linx + 3]) * 2;
    y1 := ORD(buf^[linx + 2]);
    x2 := x1 + ORD(sinfo.Xsize);
    y2 := y1 + ORD(sinfo.Ysize);
    
    IF (cx >= x1) AND (cx < x2) AND
       (cy >= y1) AND (cy < y2) THEN
	 
	BEGIN
	  ssize := ORD(sinfo.Xsize * sinfo.Ysize);
	 	  IF ssize < minsize THEN
	  BEGIN
	    minsize := ssize;
	    
	    SymInx  := linx;
	    SymInfo := sinfo;
	    SymOpt  := buf^[linx + 1];
	    SymX1 := x1;   SymX2 := x2;
	    SymY1 := y1;   SymY2 := y2;

	  END;
	
	END;
	
    linx := linx + 4;
  
  END;
  
  IF minsize < 10000 THEN	(* got one *)
  BEGIN
    SymSelect := TRUE;
    
    GetAttrs(buf^[SymInx], SymAttrs);
    
    SignInx  := sginx;
    SignBase := sgbase;
    SignNxt  := linx;
    
    QErase;
    FloatSym;
  
    QSymDraw (buf^[SymInx], buf^[SymInx + 1],
	      SignBase + SymX1, LineBase[Line] + SymY1);
    
    ClearMarks;
    
  END;
  
END;  (* SelSym *)



PROCEDURE SelText (sinx: bufInx; mrkpos: scrInx);

VAR opt: BYTE;

BEGIN
  IF (sinx = Lmark) OR (sinx = Rmark) THEN
  BEGIN
    ClearMarks;
    CmdExit;
  END;
  
  IF (sinx > Lmark) AND (Lmark > Rmark) THEN
    opt := mRight
  ELSE 
    BEGIN
      ClearMarks;
      opt := mLeft;
    END;
    
  IF buf^[sinx] >= NewLine THEN
    IF (mrkpos > 0) = (opt = mLeft) THEN
    BEGIN
      FlashAlertBox(FALSE);
      CmdExit;
    END;
  
  IF opt = mRight THEN Rmark := sinx
		  ELSE Lmark := sinx;
    
  QErase;
  DrawMark(opt, Line, mrkpos);
  QCursDraw;
  
END;  (* SelText *)



PROCEDURE Select;

VAR linx, lastinx, curinx: bufInx;
    sgbase, sgend, lastbase: scrInx;
    
BEGIN
  linx := LineStart[Line];
  
  IF buf^[linx] = EOF THEN 
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;
  
  sgbase := 8;
  sgend  := sgbase + ORD(buf^[linx + 1] * 2);
  
  curinx := linx;
  linx   := linx + ScanCode(TRUE, 1, ADR(buf^), linx);

  
  WHILE (sgbase <= CursX) AND (buf^[linx] < NewLine) DO
  BEGIN
    lastinx := curinx;
    curinx  := linx;
    
    lastbase := sgbase;
    sgbase   := sgend  + SpWidth;
    sgend    := sgbase + ORD(buf^[linx + 1] * 2);

    linx := linx + ScanCode(TRUE, 1, ADR(buf^), linx);

  END;

  
  IF sgbase > CursX THEN
    
    IF (sgbase - SpWidth) > CursX THEN
      SelSym(lastinx, lastbase)

    ELSE
      SelText(curinx, sgbase - 8)
  
  ELSE 
    
    IF sgend > CursX THEN    
      SelSym(curinx, sgbase)

    ELSE
      SelText(linx, sgend + 2);
  
END;   (* Select *)






  (* ---------------  Symbol modification  ----------------- *)



PROCEDURE CheckBound (spos: sgnInx);

BEGIN
  IF spos > 510 THEN
    
    BEGIN
      Error(Msg[41].msg^);
      CmdExit;
    END
  
  ELSE IF (SignBase + spos) > CursRMar THEN
    
    SplitLine(TRUE, SignInx, SignBase - 2)
  
  ELSE IF buf^[SignNxt] = NewSign THEN
    
    IF spos > SignLen THEN
      SplitLine(FALSE, SignNxt, SignBase + SignLen);

END;   (* CheckBound *)



PROCEDURE UpdateBound;

VAR newlen: sgnInx;
    update: BOOLEAN;

BEGIN
  update := FALSE;
  
  IF SymX2 > SignLen THEN
    
    BEGIN
      newlen := SymX2;
      
      IF buf^[SignNxt] = NewSign THEN
	SplitLine(FALSE, SignNxt, SignBase + newlen);
    END
    
  ELSE IF (SymX2 < SignLen) AND SymLast THEN
    
    BEGIN
      IF SymX2 > SignLen2 THEN newlen := SymX2
			  ELSE newlen := SignLen2;
    
      update := buf^[SignNxt] = NewSign;
    END
    
  ELSE newlen := SignLen;
  
  
  buf^[SignInx + 1] := WRD((newlen + 1) DIV 2);
  
  IF update THEN
  BEGIN
    EraseEOL(SignBase + newlen);
    DispLine(Line);
  END;
       
END;   (* UpdateBound *)



PROCEDURE SMoveX (dx: INTEGER);

VAR nx1, nx2: sgnInx;

BEGIN
  nx1 := SymX1 + dx;
  nx2 := SymX2 + dx;
  
  IF nx1 >= 0 THEN
  BEGIN
    CheckBound(nx2);
    
    QErase;
    
    QSymDraw (buf^[SymInx], buf^[SymInx + 1],
	      SignBase + nx1, LineBase[Line] + SymY1);
    
    SymX1 := nx1;
    SymX2 := nx2;
    
    buf^[SymInx + 3] := WRD(SymX1 DIV 2);
  
  END;
  
END;   (* SMoveX *)



PROCEDURE SMoveY (dy: INTEGER);

VAR ny1, ny2: sgnInx;

BEGIN
  ny1 := SymY1 + dy;
  ny2 := SymY2 + dy;
  
  IF (ny1 >= 0) AND (ny2 <= (MaxLnHt + 1)) THEN
  BEGIN
    QErase;
    
    QSymDraw (buf^[SymInx], buf^[SymInx + 1],
	      SignBase + SymX1, LineBase[Line] + ny1);
    
    SymY1 := ny1;
    SymY2 := ny2;
    
    buf^[SymInx + 2] := WRD(SymY1);
 
  END;

END;   (* SMoveY *)



PROCEDURE SRedraw (opt: BYTE);

VAR nx1, nx2, ny1, ny2: sgnInx;
    sym:    BYTE;
    ninfo:  Info;
    
BEGIN
  sym := buf^[SymInx];
  GetInfo(sym, opt, ninfo);
  
  nx1 := SymX1 + ORD(SymInfo.FixPtX - ninfo.FixPtX);
  ny1 := SymY1 + ORD(SymInfo.FixPtY - ninfo.FixPtY);
  
  nx2 := nx1 + ORD(ninfo.Xsize);
  ny2 := ny1 + ORD(ninfo.Ysize);
  
  IF (nx1 < 0) OR (ny1 < 0) OR (ny2 > (MaxLnHt + 1)) THEN
   BEGIN
     FlashAlertBox(FALSE);
     CmdExit;
   END;
  
  CheckBound(nx2);
  
  SymX1 := nx1;	    SymX2 := nx2;
  SymY1 := ny1;	    SymY2 := ny2;
  SymInfo := ninfo;
  
  buf^[SymInx + 1] := opt;
  buf^[SymInx + 2] := WRD(SymY1);
  buf^[SymInx + 3] := WRD(SymX1 DIV 2);
  
  QErase;
  
  QSymDraw (sym, opt, 
	    SignBase + nx1, LineBase[Line] + ny1);

END;   (* SRedraw *)



PROCEDURE SRotate (dr: INTEGER);

VAR opt: OptByte;
    rot, maxrot, size: INTEGER;
    
BEGIN
  maxrot := ORD(SymAttrs.Rotations);

  IF (maxrot = 0) AND (SymAttrs.SType <> Hand) THEN RETURN;


  opt := buf^[SymInx + 1];
  ReadRot(opt, rot);


  ReadSize(opt, size);
  IF (SymAttrs.SType = Hand) AND (maxrot <= 3) THEN   (* palm-facing kludge *)
    IF (size = 1) OR (size >= 3) THEN maxrot := 7;


  rot := rot + dr;
  IF rot < 0 THEN rot := maxrot;
  rot := rot MOD (maxrot + 1);
  WriteRot(opt, rot);

  SymOpt := opt;

  SRedraw(opt);

END;   (* SRotate *)



PROCEDURE SMirror;

VAR opt: OptByte;
    rot, nrots, flop, size: INTEGER;
    dead, zombie: BOOLEAN;
    
BEGIN
  opt := buf^[SymInx + 1];

  nrots := ORD(SymAttrs.Rotations) + 1;
  dead	:= SymAttrs.Flops  = 0;
  zombie:= SymAttrs.Images = 3;

  ReadSize(opt, size);


  IF SymAttrs.SType = Hand THEN    (* palm-facing kludge     *)

    IF nrots = 8 THEN
      IF ((size = 1) OR (size = 4)) AND NOT zombie THEN
        dead := FALSE ELSE {nothing}

    ELSE
      BEGIN
        IF (size = 1) OR (size >= 3) THEN nrots := 8;
	dead   := size <> 4;
        zombie := size = 1;
      END;


  IF SymAttrs.SType = Vface THEN   (* face-flopping kludge   *)
    dead := NOT ODD(size);


  IF (nrots = 8) AND zombie THEN   (* mirror-vertical kludge *)
  BEGIN  
    ReadRot(opt, rot);
    rot := (rot + 4) MOD 8;
    WriteRot(opt, rot)
  END;


  IF (nrots > 3) AND NOT((SymAttrs.SType <= Face) AND (nrots < 8)) THEN 
  BEGIN
    ReadRot(opt, rot);
    rot := (nrots - rot) MOD nrots;
    WriteRot(opt, rot);
    SymOpt := opt;

    IF dead THEN SRedraw(opt);
  END;

   
  IF NOT dead THEN
  BEGIN
    ReadFlop(opt, flop);
    flop := (flop + 1) MOD 2; 
    WriteFlop(opt, flop);
    SymOpt := opt;

    SRedraw(opt);
  END;
 
END;   (* SMirror *)



PROCEDURE SSize (kludge: BOOLEAN; ds: INTEGER);

VAR rot, maxrot, flop, size, maxsize: INTEGER;
    opt: OptByte;
    
BEGIN
  maxsize := ORD(SymAttrs.Sizes) - 1;
  
  IF (maxsize > 0) AND
     (kludge OR (SymAttrs.Stype = Movement)) THEN

  BEGIN
    opt := buf^[SymInx + 1];
    
    ReadSize(opt, size);
    size := size + ds;
    IF size < 0 THEN size := maxsize;
    size := size MOD (maxsize + 1);
    WriteSize(opt, size);   

    ReadRot(opt, rot);
    maxrot := ORD(SymAttrs.Rotations);


    IF (SymAttrs.SType = Vface) THEN  (* face-flopping kludge *)
      WriteFlop(opt, 0);
    
      
    IF (SymAttrs.SType = Hand) THEN   (* palm-facing kludge *)
    BEGIN

      IF (SymAttrs.Flops = 0) AND (size <> 4) THEN
        WriteFlop(opt, 0);

      IF (maxrot <= 3) THEN CASE size OF

	   0,2:   WriteRot(opt, rot MOD (maxrot + 1));

	   1,3,5: BEGIN ReadRot(SymOpt, rot); WriteRot(opt, rot) END;

	   4:     BEGIN ReadFlop(SymOpt, flop); WriteFlop(opt, flop) END;

      END;

    END;


    SRedraw(opt);
    
  END;

END;   (* SSize *)



PROCEDURE SChange (dc: INTEGER);

VAR sym: BYTE;
    opt: OptByte;
    ahd: Byte;
    
BEGIN
  sym := buf^[SymInx];
  
  IF SymAttrs.Stype <> Movement THEN     (* multi-size change kludge *)
      SSize(TRUE, dc)
  
  ELSE
    BEGIN
      opt := buf^[SymInx + 1];

      ReadAHead(opt, ahd);
      ahd := ahd + WRD(dc);
      IF ahd < 1 THEN ahd := 3;
      IF ahd > 3 THEN ahd := 1;
      
      WriteAHead(opt, ahd);
      buf^[SymInx + 1] := opt;
      
      QErase;
      
      QSymDraw (sym, opt, 
		SignBase + SymX1, LineBase[Line] + SymY1);
  
    END;
  
END;   (* SChange *)



PROCEDURE SCursInit (shift: BOOLEAN);

VAR xoff, yoff: sgnInx;
    rotval:     INTEGER;
    opt:        OptByte;

BEGIN
  xoff := 0;
  yoff := 0;
  CursLoc := 5;
  
  IF SymAttrs.Stype <= Face THEN 
  BEGIN
    xoff := 6;
    yoff := 3;
    
    IF (SymAttrs.Rotations = 7) AND 
       (SymAttrs.Images    < 5) AND 
        NOT shift THEN
    BEGIN
      opt := buf^[SymInx + 1];
      ReadRot(opt, rotval);
      CursLoc := (rotval + 3) MOD 8;
    END;
    
  END;
    
  CursGrid[0].X := SymX1 + ORD(SymInfo.Xsize) DIV 2;
  CursGrid[0].Y := SymY1 - 2;
  
  CursGrid[1].X := SymX1 - 4 + xoff;
  CursGrid[1].Y := SymY1 - 2 + yoff;

  CursGrid[2].X := SymX1 - 4;
  CursGrid[2].Y := SymY1 + ORD(SymInfo.Ysize) DIV 2;

  CursGrid[3].X := CursGrid[1].X;
  CursGrid[3].Y := SymY2 + 1 - yoff;

  CursGrid[4].X := CursGrid[0].X;
  CursGrid[4].Y := SymY2 + 1;
  
  CursGrid[5].X := SymX2 + 3 - xoff;
  CursGrid[5].Y := SymY2 + 1 - yoff;

  CursGrid[6].X := SymX2 + 3;
  CursGrid[6].Y := CursGrid[2].Y;

  CursGrid[7].X := CursGrid[5].X;
  CursGrid[7].Y := CursGrid[1].Y;

END;   (* SCursInit *)



PROCEDURE SCursor (dc: INTEGER);

VAR ok: BOOLEAN;

BEGIN
  
  IF NOT CursSelect THEN
    BEGIN
      UpdateBound;
      SCursInit (dc < 0);
    END
    
  ELSE QErase;
  
  SymSelect := FALSE;
  CursSelect := TRUE;
      
  REPEAT
    CursLoc := CursLoc + dc;
    
    IF CursLoc < 0 THEN CursLoc := 7;
    CursLoc := CursLoc MOD 8;
  
    WITH CursGrid[CursLoc] DO
    BEGIN
      IF (SignBase + X) >= CursRMar THEN
	SplitLine(TRUE, SignInx, SignBase - 2);
      
      ok := (X >= 6) AND (Y >= 1) AND (Y <= (MaxLnHt - 1));
    
    END;
  
  UNTIL ok;
    
  CursX := SignBase + CursGrid[CursLoc].X;
  CursY := LineBase[Line] + CursGrid[CursLoc].Y;
  
  IF ODD(CursX) THEN CursX := CursX + 1;
  
  SCursDraw;

END;   (* SCursor *)



PROCEDURE DeSelect;

BEGIN
  UpdateBound;
  
  CursX := SignBase + SymX1 + ORD(SymInfo.Xsize) DIV 2;
  CursY := LineBase[Line] + SymY1 + ORD(SymInfo.Ysize) DIV 2;
  
  IF ODD(CursX) THEN CursX := CursX + 1;
  QCursDraw;
  
  SymSelect := FALSE;
  
END;   (* DeSelect *)
	    
	    

PROCEDURE SSave (delta: INTEGER);

BEGIN
  unstate := unsymbol;   (* save sign state for Undo *)

  unmark := SignInx; 
  unlen  := SignNxt - SignInx;
  undif  := delta;


  IF Window THEN   	 (* save following EOF code for WUndo *)
    unlen := unlen + 2;


  MoveL(ADR(buf^[unmark]), ADR(altbuf^[0]), WRD(unlen));

END;  (* SSave *)



PROCEDURE SCopy;

VAR cx1, cx2: sgnInx;
    linx: bufInx;

BEGIN
  UpdateBound;
  
  SignLen := ORD(buf^[SignInx + 1]) * 2;
  SymLast := FALSE;
  
  cx1 := SymX2 + 6; 
  IF ODD(cx1) THEN cx1 := cx1 + 1;
  cx2 := cx1 + ORD(SymInfo.Xsize);
  
  CheckBound(cx2);
  
  SymDraw (buf^[SymInx], buf^[SymInx + 1],
	   SignBase + SymX1, LineBase[Line] + SymY1);
  
  linx := SignInx + 2;

  SSave(4);
  MakeSpace(linx, 4);
  
  MoveL(ADR(buf^[SymInx + 4]), ADR(buf^[linx]), 4);
  buf^[linx + 3] := WRD(cx1 DIV 2);
  
  SignNxt := SignNxt + 4;
  
  SymX1	 := cx1; 
  SymX2	 := cx2;
  SymInx := linx;
  
  QSymDraw (buf^[linx], buf^[linx + 1],
	    SignBase + cx1, LineBase[Line] + SymY1);

END;   (* SCopy *)



PROCEDURE SErase;

BEGIN
  QErase;
  
  SSave(-4);
  DeleteCode(SymInx, 4);
  
  SignNxt := SignNxt - 4;
  SymX2 := 0;
  
  DeSelect;
      
END;   (* SErase *)
	    
	    
	    



  (* ------------------	 Symbol entry  --------------------- *)



PROCEDURE GetDescender (sym, opt: BYTE; VAR descend: INTEGER);

VAR found: BOOLEAN;
    inx:   INTEGER;

BEGIN
  inx := 0;
  descend := 0;
  
  REPEAT
    WHILE sym <> Descender.fsym[inx] DO
      IF Descender.fsym[inx] = 255 THEN RETURN
      ELSE inx := inx + 1;

    found := (opt = Descender.fopt[inx]);
               
    IF NOT found THEN inx := inx + 1;

  UNTIL found;

  descend := ORD(Descender.dval[inx]);

END;   (* GetDescender *)



FUNCTION EntSetup (xsize, ysize, descend: INTEGER; boxup: BOOLEAN;
		   VAR xpos, ypos: sgnInx): BOOLEAN;

VAR xhalf, xhole: INTEGER;
    xbase, ybase: sgnInx;

BEGIN
  xhole := xsize;
  xhalf := xsize DIV 2;
  
  IF ODD(xhole) THEN xhole := xhole + 1;
  IF ODD(xhalf) THEN xhalf := xhalf + 1;
  
  xbase := CursX;  {default}
  ybase := CursY - ysize + 1;

  
  IF CursSelect THEN
    CASE CursLoc OF
      0: xbase := CursX - xhalf;
      
      1: xbase := CursX - xhole;
      
      2: BEGIN
	   xbase := CursX - xhole;
	   ybase := CursY - ysize DIV 2;
	 END;
      
      3: BEGIN
	   xbase := CursX - xhole;
	   ybase := CursY;
	 END;

      4: BEGIN
	   xbase := CursX - xhalf;
	   ybase := CursY;
	 END;
      
      5: ybase := CursY;

      6: ybase := CursY - ysize DIV 2;

      OTHERWISE;
    END;


  ybase := ybase + descend;    (* offset by descender (if any) *)
    
  xbase := xbase - SignBase;
  ybase := ybase - LineBase[Line];
  

  IF (SignBase < 0) OR (xbase < 0) OR 
     (ybase < 0) OR ((ybase + ysize) > (MaxLnHt + 1)) THEN
  BEGIN
    FlashAlertBox(boxup);
    EntSetup := FALSE;

    RETURN;

  END;

  
  IF boxup THEN
  BEGIN
    DrawKeyBoxes(0);
    IF CursSelect THEN SCursDraw ELSE QCursDraw;
  END;


  CheckBound(xbase + xsize);
  
  ClearMarks;

  
  IF buf^[SignInx] = EOF THEN		(* insert NewLine *)
  BEGIN
    IF Line < MaxLn THEN LineStart[Line + 1] := SignInx;
    MakeSpace(SignInx, 2);
    SignNxt := SignNxt + 2;
    buf^[SignInx] := NewLine;
    buf^[SignInx + 1] := 0;
  END;

      
  xpos := xbase;
  ypos := ybase;
  EntSetup := TRUE;

END;   (* EntSetup *)



FUNCTION EnterSym (sym, opt: BYTE; boxup: BOOLEAN): BOOLEAN;

VAR descend: INTEGER;
    next:    sgnInx;

BEGIN
  EnterSym := TRUE;
  
  IF sym = 255 THEN    (* undefined *)
  BEGIN
    EnterSym := FALSE;

    RETURN;

  END;


  GetSign(SignBase, SignInx, SignNxt);	  (* check bounds *)
  GetInfo(sym, opt, SymInfo);
 
  IF TMode = Finger THEN GetDescender(sym, opt, descend)
                    ELSE descend := 0;

  
  SignLen := ORD(buf^[SignInx + 1]) * 2;
  
  IF NOT EntSetup(ORD(SymInfo.Xsize), ORD(SymInfo.Ysize), descend,
                                              boxup, SymX1, SymY1) THEN
  BEGIN
    EnterSym := FALSE;

    RETURN;

  END;
  
  
  SymInx := SignInx + 2;	    (* insert symcode *)
  MakeSpace(SymInx, 4);

  SignNxt := SignNxt + 4;
  
  buf^[SymInx] := sym;
  buf^[SymInx + 1] := opt;
  buf^[SymInx + 2] := WRD(SymY1);
  buf^[SymInx + 3] := WRD(SymX1) DIV 2;
    
  
  QErase;			    (* draw symbol *)

  QSymDraw (sym, opt, 
	    SignBase + SymX1,
	    LineBase[Line] + SymY1);
  
  
  next := SymX1 + ORD(SymInfo.Xsize);

  
  IF (TMode = Finger) OR (TMode = Alpha) THEN   (* advance cursor *)
    
    BEGIN
      IF ODD(next) THEN next := next + 1;
      
      IF next > SignLen THEN
	buf^[SignInx + 1] := WRD(next) DIV 2;
	
      IF TMode = Finger THEN next := next + 6;
      
      IF (SignBase + next) > CursRMar THEN
	SplitLine(TRUE, SignInx, SignBase - 2);
    
      CursX := SignBase + next;
      CursY := LineBase[Line] + SymY1 + ORD(SymInfo.Ysize - 1);
      CursY := CursY - descend;
      
      QCursDraw;
    
    END
  
  ELSE	 (* specify entered symbol as selected *)
    
    BEGIN
      GetAttrs(buf^[SymInx], SymAttrs);
      
      SymX2 := next;
      SymY2 := SymY1 + ORD(SymInfo.Ysize);
      
      SymLast := FALSE;
      SymOpt  := opt;

      IF next > SignLen THEN
      BEGIN
	SignLen2 := SignLen;
	SignLen	 := next;
	SymLast	 := TRUE;
      END;
      
      SymSelect	 := TRUE;
    
    END;
    
    
  CursSelect := FALSE;

END;   (* EnterSym *)
  


FUNCTION EnterObj (entry: InxRec; boxup: BOOLEAN): BOOLEAN;

VAR i, xsize, ysize, nsyms, nbytes, descend: INTEGER;
    pt, next, xmax2, objx, objy:  sgnInx;
    xmin, xmax, ymin, ymax:  sgnInx;
    sinx, linx:	 keyInx;
    lsym, lopt:  BYTE;
    im: Info;

BEGIN
  EnterObj := TRUE;
  
  IF entry.sym.rotval > 127 THEN
    EnterObj := EnterSym(entry.sym.codeval, StockOpt, boxup)
  
  ELSE	 (* compound symbol *)
    
    BEGIN
      GetSign(SignBase, SignInx, SignNxt);
      
      SignLen := ORD(buf^[SignInx + 1]) * 2;
      
      sinx  := entry.inx;
      nsyms := ORD(keymap[sinx]);
      linx  := entry.inx + 1;
      
      lsym  := keymap[linx];
      lopt  := keymap[linx + 1];


      xmin := 100;   xmax := 0;	  xmax2 := 0;	   (* get bounds *)
      ymin := 100;   ymax := 0;
      
      FOR i := 1 TO nsyms DO
      BEGIN
	GetInfo(keymap[linx], keymap[linx + 1], im);
	
	pt := ORD(keymap[linx + 3]); 
	IF pt < xmin THEN xmin := pt;
	
	pt := pt + ORD(im.Xsize); 
	IF pt > xmax THEN 
	BEGIN 
	  xmax2 := xmax; 
	  xmax := pt;
	END
	ELSE IF pt > xmax2 THEN 
	  xmax2 := pt;
	
	pt := ORD(keymap[linx + 2]); 
	IF pt < ymin THEN ymin := pt;
	
	pt := pt + ORD(im.Ysize);
	IF pt > ymax THEN ymax := pt;
	
	linx := linx + 4;
      END;
      
      xsize := xmax - xmin;
      ysize := ymax - ymin;

      IF TMode = Finger THEN GetDescender(lsym, lopt, descend)
                        ELSE descend := 0;

      IF NOT EntSetup(xsize, ysize, descend, boxup,
		      objx, objy) THEN
      BEGIN
	EnterObj := FALSE;
	RETURN;
      END;
      
      
      linx := SignInx + 2;		(* insert symcodes *)
      nbytes := nsyms * 4;
      
      MakeSpace(linx, nbytes);
      SignNxt := SignNxt + nbytes;
      
      MoveL(ADR(keymap[sinx + 1]), ADR(buf^[linx]), WRD(nbytes));
	
      
      QErase;				(* draw symbols *)
      
      FOR i := 1 TO nsyms DO
      BEGIN
	buf^[linx + 3] := Wrd ((Ord (buf^[linx + 3]) - xmin + objx) DIV 2);
	buf^[linx + 2] := Wrd (Ord (buf^[linx + 2]) - ymin + objy);
	
	QSymDraw (buf^[linx], buf^[linx + 1],
		  SignBase +  ORD(buf^[linx + 3]) * 2,
		  LineBase[Line] + ORD(buf^[linx + 2]) );
      
	linx := linx + 4;
      END;
      
      
      next := objx + xsize;
      
      IF TMode = Finger THEN   (* advance cursor *)
	
	BEGIN
	  IF ODD(next) THEN next := next + 1;
 	  IF next > SignLen THEN
	    buf^[SignInx + 1] := WRD(next) DIV 2;
	    
	  next := next + 6;
 	  IF (SignBase + next) > CursRMar THEN
	    SplitLine(TRUE, SignInx, SignBase - 2);
	
	  CursX := SignBase + next;
	  CursY := LineBase[Line] + objy + ysize - 1;
          CursY := CursY - descend;

          QCursDraw;
	
	END
      
      ELSE   (* select last symbol entered *)
	
	BEGIN
	  SymInx  := linx - 4;
	  SymInfo := im;
	  
	  GetAttrs(buf^[SymInx], SymAttrs);
	  SymX1 := ORD(buf^[SymInx + 3]) * 2;
	  SymY1 := ORD(buf^[SymInx + 2]);
	  SymX2 := SymX1 + ORD(SymInfo.Xsize);
	  SymY2 := SymY1 + ORD(SymInfo.Ysize);

	  SymLast := FALSE;
          SymOpt  := buf^[SymInx + 1];
	  
	  IF next > SignLen THEN
	  BEGIN
	    SignLen2 := SignLen;
	    SignLen  := next;
	    
	    IF (objx + xmax2 - xmin) > SignLen2 THEN
	      SignLen2 := objx + xmax2 - xmin;

	    SymLast := SymX2 = SignLen;

	  END;

	  SymSelect  := TRUE;
	
	END;   (* selected *)
      
      
      CursSelect := FALSE;
  
    END;   (* compound symbol *)
  
END;   (* EnterObj *)



FUNCTION GetBox (maxbox: INTEGER): INTEGER;

VAR ch:	  CHAR;
    box:  INTEGER;
    code: INTEGER;

BEGIN
  box := 100;
  
  REPEAT
    GetChar(-1, AllKeys, ch, code);
  
    IF ch = CHR(esc) THEN
      BEGIN
	DrawKeyBoxes(0);
        IF CursSelect THEN SCursDraw ELSE QCursDraw;

	CmdExit;

      END
      
    ELSE IF ch IN [CHR(left), CHR(right), CHR(up), CHR(down)] THEN
      BEGIN
	CursSelect := FALSE;

	IF      ch = CHR(left)  THEN LMoveX(-2)
	ELSE IF ch = CHR(right) THEN LMoveX(+2)
	ELSE IF ch = CHR(up)	THEN LMoveY(-1)
	ELSE  { ch = CHR(down)}	     LMoveY(+1);

      END

    ELSE CASE code OF 
      SC_a:  box := 0;
      SC_s:  box := 1;
      SC_d:  box := 2;
      SC_f:  box := 3;
      
      SC_g:  box := 8;
      SC_h:  box := 9;
      
      SC_j:  box := 4;
      SC_k:  box := 5;
      SC_l:  box := 6;
      SC_x:  box := 7;
      OTHERWISE;

    END;
	
  UNTIL box <= maxbox;

  
  GetBox := box;

  
END;   (* GetBox *)



FUNCTION EnterBox (entry: InxRec; boxup: BOOLEAN): BOOLEAN;

VAR sym: BYTE;
    opt: OptByte;
    at:  Attrs;
    maxrot, size: INTEGER;

BEGIN
  EnterBox := TRUE;

  IF entry.sym.rotval <= 127 THEN        (* compound sym *)
    EnterBox := EnterObj(entry, TRUE)
      
  ELSE				  	 (* unit sym *)
    BEGIN
      sym := entry.sym.codeval;

      KeyToOpt(entry.sym.rotval, opt);
      ReadSize(opt, size);

      GetAttrs(sym, at);
      maxrot := ORD(at.Rotations);
         

      IF (at.SType = Hand) AND (maxrot <= 3) THEN  (* palm-facing kludge *)
        IF (size = 1) OR (size >= 3) THEN maxrot := 7;


      IF maxrot = 0 THEN
	EnterBox := EnterSym(sym, opt, boxup)
      
      ELSE     (* display symbol rotations *)
        BEGIN
          IF boxup THEN EraseKeys
	           ELSE DrawKeyBoxes(1);

          DispRots(sym, opt, maxrot);
      
          REPEAT WriteRot(opt, GetBox(maxrot))
	  UNTIL EnterSym(sym, opt, TRUE);

	END;
	
    END;   (* unit sym *)
    
END;   (* EnterBox *)
    


PROCEDURE EnterKey (code: INTEGER; state: ShiftState);

VAR max, kinx, vinx: INTEGER;
    entry: InxRec;
    junk:  BOOLEAN;
    ch:    CHAR;

BEGIN
  IF (TMode = Alpha) OR DTxtWnd THEN

    BEGIN
      entry := Keyboard^[Alpha, code, state];
      
      IF entry.sym.rotval > 127 THEN   (* live *)
        ch := CHR(entry.sym.codeval)
      ELSE 
        GetDead(entry, ch);            (* dead *)

      IF DTxtWnd THEN
        IF ch IN DNameKeys THEN DAddChar(ch) ELSE {nothing}

      ELSE IF ch <> CHR(255) THEN junk := EnterSym(CharCode, WRD(ch), FALSE);

    END


  ELSE IF TMode = Finger THEN

    BEGIN
      IF state = Shft THEN state := None;
      junk := EnterObj(Keyboard^[Finger, code, state], FALSE)

    END


  ELSE  {TMode = Sign}

    BEGIN
      IF state = Shft THEN state := None;

      entry := Keyboard^[Sign, code, state];
      
      IF entry.sym.rotval > 127 THEN
	junk := EnterBox(entry, FALSE)

	
      ELSE       (* display symbol variations *)
	BEGIN
 	  kinx := entry.inx;

          max  := ORD(keymap[kinx]) - 1;

          IF max = 0 THEN
	    BEGIN
	      MoveL(ADR(keymap[kinx + 1]), ADR(entry), 2);
              junk := EnterObj(entry, FALSE)
            END

          ELSE	   (* display symbol variations *)
	    BEGIN
	      DrawKeyBoxes(1);
	      DispKeys(kinx);
	      
	      REPEAT
                vinx := kinx + 1 + (GetBox(max) * 2);
                MoveL(ADR(keymap[vinx]), ADR(entry), 2);

              UNTIL EnterBox(entry, TRUE);

	    END;
	    
	END;
      
    END; {TMode = Sign}

      
END;  (* EnterKey *)






  (* ------------------	 Edit commands	-------------------- *)



PROCEDURE Space;

VAR sinx: bufInx;
    sbase, lend, ybase: scrInx;

BEGIN
  ClearMarks;
  
  GetSignBound(sbase, sinx);
  GetLineEnd(lend);
  
  MakeSpace(sinx, 2);
  buf^[sinx] := WRD(NewSign + ORD(sbase = 8));
  buf^[sinx + 1] := 0;
  
  QErase;
  
  IF (sbase = 8) OR (buf^[sinx + 2] = NewSign) THEN
    
    IF (CursRMar - lend) >= SpWidth THEN
      BEGIN
	IF buf^[sinx + 2] = NewLine THEN
	  buf^[sinx + 2] := NewSign;
	EraseLine(Line);
	DispLine(Line);
      END
    
    ELSE
      SplitLine(FALSE, sinx + 2, sbase)
    
  ELSE IF (632 - sbase) < (SpWidth + CursIn + 6) THEN  (* wrap *)
    
    BEGIN
      SplitLine(TRUE, sinx, sbase - 2);
      sbase := 8;
    END
      
  ELSE	 (* there's room at the end *)
    
    BEGIN
      SymDraw (Spacer, SpaceOpt,
	       sbase - 2, LineBase[Line] + SpYOff);
      sbase := sbase + SpWidth - 2;
    END;
    
  UpdateGauge;
  
  CursX := sbase + CursIn;
  CursY := LineBase[Line] + CursHomeY;
  
  QCursDraw;

END;   (* Space *)
	    
	    

PROCEDURE Returner;

VAR sinx:  bufInx;
    sbase: scrInx;
     
BEGIN
  ClearMarks;
  
  GetSignBound(sbase, sinx);
  
  IF buf^[sinx] >= NewLine THEN
  BEGIN
    MakeSpace(sinx, 2);
    buf^[sinx] := NewLine;
    buf^[sinx + 1] := 0;
    IF sbase = 8 THEN sinx := sinx + 2;
  END;
  
  QErase;
  SplitLine(TRUE, sinx, sbase - 2);   (* fake wrap *)
  
  UpdateGauge;
  
  CursX := 8 + CursIn;
  CursY := LineBase[Line] + CursHomeY;
  
  QCursDraw;
    
END;   (* Returner *)
	    
	    

PROCEDURE AlphaReturn;

VAR linx: sgnInx;
    xmin, ymax, xpt, ypt: scrInx;

BEGIN
  ClearMarks;
  
  GetSign(SignBase, SignInx, SignNxt);
  
  xmin := 255;
  ymax := 0;
  linx := SignInx + 2;
  
  WHILE buf^[linx] < NewSign DO
  BEGIN

    IF buf^[linx] = CharCode THEN
    BEGIN
      xpt := ORD(buf^[linx + 3]);
      ypt := ORD(buf^[linx + 2]);
    
      IF xpt < xmin THEN xmin := xpt;
      IF ypt > ymax THEN ymax := ypt;
    
    END;
    
    linx := linx + 4;

  END;
      
  IF xmin < 255 THEN  {char in sign}
  BEGIN
    ypt := ymax + CursARet;

    IF ypt > MaxLnHt THEN BEGIN FlashAlertBox(FALSE); RETURN END;

    QErase;
  
    CursX := SignBase + (xmin * 2);
    CursY := LineBase[Line] + ypt;
  
    QCursDraw;
  END;
  
END;   (* AlphaReturn *)
	    


PROCEDURE Deleter;

VAR sym, opt: BYTE;
    descend:  INTEGER;
    im: Info;
    
BEGIN
  GetSign(SignBase, SignInx, SignNxt);
  SymInx := SignInx + 2;


  IF (SignBase < 0) OR  (buf^[SignInx] = EOF) OR
     ((SignInx = 0) AND (buf^[SymInx] = EOF)) THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;


  IF buf^[SymInx] >= NewSign THEN     {on space}
  BEGIN

    IF buf^[SignInx] = NewLine THEN   {lead space... }

      IF buf^[SymInx] >= NewLine THEN {... on blank line}

        BEGIN 
          FlashAlertBox(FALSE);
          CmdExit;
        END

      ELSE buf^[SymInx] := NewLine;   {... on nonblank line}


    ClearMarks;
    
    SSave(-2);
    DeleteCode(SignInx, 2);

    EraseEOL(SignBase - 2);
    DispLine(Line);
    
    CursX := SignBase;
    CursY := LineBase[Line] + CursHomeY;
    QCursDraw;
     
    CmdExit;
   
  END; (* if *)

  
  ClearMarks;

  sym := buf^[SymInx];
  opt := buf^[SymInx + 1];
  
  GetDescender(sym, opt, descend);      (* fingerspell kludge *)
  GetInfo(sym, opt, im);
  
  SymX1 := ORD(buf^[SymInx + 3]) * 2;
  SymX2 := SymX1 + ORD(im.Xsize);
  SymY1 := ORD(buf^[SymInx + 2]);
  SymY2 := SymY1 + ORD(im.Ysize);
  
  QErase;
  FloatSym;
  
  SSave(-4);
  DeleteCode(SymInx, 4);
  SignNxt := SignNxt - 4;
  SymX2 := 0;
  
  UpdateBound;
  
  CursX := SignBase + SymX1;
  CursY := LineBase[Line] + SymY2 - 1;
  
  CursY := CursY - descend;
  
  QCursDraw;
  
END;   (* Deleter *)



PROCEDURE GetBounds (xbase, ybase: sgnInx;
                     VAR xmin, xmax, ymin, ymax: sgnInx);

VAR xpt,  ypt:  sgnInx;
    linx: bufInx;
    im:	  Info;

{ Assert: GetSign(SignBase, SignInx, SignNxt) }

BEGIN
  xmin := 600;	 xmax := 0;
  ymin := 600;	 ymax := 0;
  
  SignLen := 0;
  SymLast := FALSE;
  linx    := SignInx + 2;
    
  WHILE buf^[linx] < NewSign DO
  BEGIN
    GetInfo(buf^[linx], buf^[linx + 1], im);
    
    xpt := ORD(buf^[linx + 3]) * 2;
    ypt := ORD(buf^[linx + 2]);
    
    IF (xpt >= xbase) AND (ypt >= ybase) THEN
      BEGIN
	IF xpt < xmin THEN xmin := xpt;
	
	xpt := xpt + ORD(im.Xsize);
	IF xpt > xmax THEN xmax := xpt;
	
	IF ypt < ymin THEN ymin := ypt;
	
	ypt := ypt + ORD(im.Ysize);
	IF ypt > ymax THEN ymax := ypt;
      
      END
    
    ELSE IF (xpt + ORD(im.Xsize)) > SignLen THEN
      SignLen := xpt + ORD(im.Xsize);
      
    linx := linx + 4;

  END;
      
  IF xmax > SignLen THEN
  BEGIN
    SignLen2 := SignLen;
    SignLen  := xmax;
    SymLast  := TRUE;
  END;


END;  (* GetBounds *)



PROCEDURE SignAdjust (VAR ok: BOOLEAN; xbase, ybase: sgnInx; 
		      	               xoff,  yoff:  INTEGER);

VAR xmin, xmax, ymin, ymax: sgnInx;
    xpt,  ypt:  sgnInx;
    linx: bufInx;

{ Assert: GetSign(SignBase, SignInx, SignNxt) }

BEGIN
  ok := TRUE;

  GetBounds(xbase, ybase, xmin, xmax, ymin, ymax);


		     		    (* check new bounds *)
  IF yoff < 0 THEN   		   

    IF ((ymin + yoff) < ybase) OR (ymin = 600) THEN ok := FALSE
    ELSE {nothing}

  ELSE IF yoff > 0 THEN 

    IF ((ymax + yoff) > MaxLnHt) OR (ymax = 0) THEN ok := FALSE
    ELSE {nothing};
  
  
  IF xoff < 0 THEN

    IF ((xmin + xoff) < xbase) OR (xmin = 600) THEN ok := FALSE
    ELSE {nothing}

  ELSE IF xoff > 0 THEN

    IF xmax = 0 THEN ok := FALSE
    ELSE CheckBound(xmax + xoff);

    
  IF NOT ok THEN 
  BEGIN
    FlashAlertBox(FALSE);

    RETURN;

  END;


				    (* redraw sign *)
  linx := SignInx + 2;
  
  DrawBlock (SignBase + xbase, LineBase[Line] + ybase,
             xmax + 2 - xbase, MaxLnHt + 1 - ybase, 0);

  WHILE buf^[linx] < NewSign DO
  BEGIN
    xpt := ORD(buf^[linx + 3]) * 2;
    ypt := ORD(buf^[linx + 2]);
  
    IF (xpt >= xbase) AND (ypt >= ybase) THEN
    BEGIN
      buf^[linx + 3] := buf^[linx + 3] + WRD(xoff DIV 2);
      buf^[linx + 2] := WRD(ypt + yoff);
      xpt := xpt + xoff;
      ypt := ypt + yoff;
    END;
  
    SymDraw (buf^[linx], buf^[linx + 1], 
             SignBase + xpt, LineBase[Line] + ypt);
  
    linx := linx + 4;

  END;

				    
  SymX2 := xmax + xoff;		    (* update sign bound *)

  UpdateBound;


END;   (* SignAdjust *)



PROCEDURE SignMove (dx, dy: INTEGER);

VAR cx, cy: sgnInx;
    done:   BOOLEAN;

BEGIN
  ClearMarks;
  
  GetSign(SignBase, SignInx, SignNxt);
  
  IF (SignBase < 0) OR (buf^[SignInx] = EOF) THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;

  QErase;

  cx := CursX - SignBase;
  cy := CursY - LineBase[Line];


  SignAdjust(done, 0, 0, dx, dy);

  
  CursX := SignBase + cx;
  CursY := LineBase[Line] + cy;

  QCursDraw;

  IF done THEN
    IF dx <> 0 THEN LMoveX(dx)
	       ELSE LMoveY(dy);
    
END;   (* SignMove *)



PROCEDURE SignCenter;

VAR cx, cy, xmin, xmax, ymin, ymax: sgnInx;
    dx, dy: INTEGER;
    done:   BOOLEAN;

BEGIN
  ClearMarks;
  
  GetSign(SignBase, SignInx, SignNxt);
  
  IF (SignBase < 0) OR (buf^[SignInx] = EOF) THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;

  QErase;

  GetBounds(0, 0, xmin, xmax, ymin, ymax);


  cx := CursX - SignBase;
  cy := CursY - LineBase[Line];

  dx := CursIn - xmin;
  dy := ((MaxLnHt + 1) - (ymax - ymin + 1)) DIV 2 - ymin;


  SignAdjust(done, 0, 0, dx, dy);

  IF done THEN
    BEGIN cx := CursIn; cy := CursHomeY END;


  CursX := SignBase + cx;
  CursY := LineBase[Line] + cy;

  QCursDraw;

    
END;   (* SignCenter *)



PROCEDURE Adjust;

VAR ch:	    CHAR;
    done:   BOOLEAN;
    cx, cy: sgnInx;
    dx, dy, junk: INTEGER;

BEGIN
  ClearMarks;
  
  GetSign(SignBase, SignInx, SignNxt);
  
  IF (SignBase < 0) OR (buf^[SignInx] = EOF) THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;
    
  SSave(0);

  
  QErase;			  (* change cursor *)
  CursLoc := 5;
  SCursDraw;

  cx := CursX - SignBase;
  cy := CursY - LineBase[Line];

  
  WriteCmdName(Msg[69].msg^);     (* display cmd msg *)
  WriteLine(Msg[15].msg^);
  
  UpdateCmdList := TRUE;

  
  REPEAT
    GetChar(-1, [CHR(left), CHR(right), CHR(up), CHR(down), 
	         CHR(esc), CHR(ret)], ch, junk);
      
    
    IF (ch = CHR(esc)) OR (ch = CHR(ret)) THEN	  (* exit *)
    BEGIN
      QErase; 
      QCursDraw;

      CmdExit;

    END;
    
      
    dx := 0; dy := 0;

    IF      ch = CHR(up)    THEN dy := -1
    ELSE IF ch = CHR(down)  THEN dy := +1
    ELSE IF ch = CHR(left)  THEN dx := -2
    ELSE IF ch = CHR(right) THEN dx := +2;


    SignAdjust(done, cx, cy, dx, dy);


    CursX := SignBase + cx;			  (* update cursor *)
    CursY := LineBase[Line] + cy;


    IF done THEN SCursDraw
    	    ELSE WriteCmdName(Msg[69].msg^);

    
  UNTIL FALSE;
  
END;   (* Adjust *)
	    
	    

PROCEDURE Remargin (backup: BOOLEAN; inx: bufInx; stopcode: BYTE;
		    VAR base, stopinx: bufInx);

VAR xpos: scrInx;
    linx: bufInx;
    len:  INTEGER;

BEGIN
  base := inx;
  stopinx := inx;

  IF buf^[inx] = EOF THEN RETURN;
  
  IF backup THEN
    IF inx > 0 THEN
      base := inx + ScanCode(FALSE, -1, ADR(buf^), inx)
    ELSE IF buf^[0] < NewLine THEN
      buf^[0] := NewLine;
  
  xpos := 8 + ORD(buf^[base + 1] * 2) + SpWidth;
  
  linx := base + ScanCode(TRUE, 1, ADR(buf^), base);
  len := ORD(buf^[linx + 1]) * 2;
  
  WHILE (buf^[linx] < stopcode) OR (len > 0) DO
  BEGIN
    xpos := xpos + len + SpWidth;
  
    IF xpos > (CursRMar + SpWidth) THEN
      BEGIN
	buf^[linx] := NewLine;
	xpos := 8 + len + SpWidth;
      END
    ELSE 
      buf^[linx] := NewSign;
    
    linx := linx + ScanCode(TRUE, 1, ADR(buf^), linx);
    len := ORD(buf^[linx + 1]) * 2;
    
  END;
	
  stopinx := linx;

END;   (* Remargin *)
	    
	    

PROCEDURE Margin;

VAR junk: bufInx;

BEGIN
  ClearMarks;
  
  Remargin(FALSE, LineStart[Line], NewLine, junk, junk);
  CNewPage(LineStart[0]);

END;   (* Margin *)
	    
	    

PROCEDURE Copy;

VAR sinx, stop, pbase, rbase: bufInx;
    junk:   scrInx;
    nbytes: INTEGER;

BEGIN
  GetSignBound(junk, sinx);
  
  IF (Lmark = -1) OR (Rmark = -1) THEN
  BEGIN
    Error(Msg[16].msg^);
    CmdExit;
  END;
  
  IF (sinx > Lmark) AND (sinx < Rmark) THEN
  BEGIN
    Error(Msg[17].msg^);
    CmdExit;
  END;
  
  nbytes := Rmark - Lmark;
  
  IF nbytes > (Maxbuf - bufEnd) THEN
  BEGIN
    Error(Msg[18].msg^);
    CmdExit;
  END;
  
  IF sinx <= Lmark THEN
  BEGIN
    Lmark := Lmark + nbytes;
    Rmark := Rmark + nbytes;
  END;
  
  MakeSpace(sinx, nbytes);
  
  MoveL(ADR(buf^[Lmark]), ADR(buf^[sinx]), WRD(nbytes));
  Remargin(TRUE, sinx, NewLine, rbase, stop);
  
  pbase := LineStart[0];
  IF pbase = sinx THEN pbase := rbase;
    
  IF stop < (sinx + nbytes) THEN
    Remargin(TRUE, sinx + nbytes, NewLine, rbase, stop);
  
  unmark  := sinx;
  unlen	  := nbytes;
  unstate := uncopy;
  
  CNewPage(pbase);
  
  UpdateGauge;
  
END;   (* Copy *)
	    
	    

PROCEDURE Eraser (cmd: BOOLEAN);

VAR pbase, ls0, junk: bufInx;
    nbytes: INTEGER;
    oline:  linInx;
    resp:   cmdVal;
    junk2:  CHAR;

BEGIN
  nbytes := Rmark - Lmark;
  
  IF (Lmark = -1) OR (Rmark = -1) THEN
  BEGIN
    Error(Msg[19].msg^);
    CmdExit;
  END;
  
  unmark  := Lmark;
  unlen	  := nbytes;
  unstate := unerase;
  
  IF cmd AND (nbytes > MaxABuf) THEN
  BEGIN
    unmark := -1; 

    DrawAlertBox(1);
    Prompt(20, [], Msg[20].msg^, junk2, resp);
    DrawAlertBox(0);

    IF resp <> C_Yes THEN CmdExit;
  END;
  
  IF cmd AND (unmark <> -1) THEN
    MoveL(ADR(buf^[Lmark]), ADR(altbuf^[0]), WRD(nbytes));
  
  DeleteCode(Lmark, nbytes);
  
  Remargin(TRUE, Lmark, NewLine, pbase, junk);
  
  ls0 := LineStart[0];
   
  IF Rmark <= ls0 THEN
    BEGIN
      pbase := ls0 - nbytes;
      IF (pbase > 0) AND (buf^[pbase] <= NewLine) THEN
	pbase := pbase + ScanCode(FALSE, -1, ADR(buf^), pbase);
    END
    
  ELSE IF Lmark <= ls0 THEN
    IF pbase > 0 THEN
      pbase := pbase + ScanCode(FALSE, -1, ADR(buf^), pbase)
    ELSE {nothing}
    
  ELSE	 (* Lmark > ls0 *)
    pbase := ls0;
  
  Lmark := -1;
  Rmark := -1;
  
  CNewPage(pbase);
  
  oline := Line;
  WHILE LineStart[Line] = -1 DO Line := Line - 1;
  
  IF Line <> oline THEN
  BEGIN
    QErase;
    CursY := LineBase[Line] + CursHomeY;
    QCursDraw;
  END;
  
  UpdateGauge;
  
END;   (* Eraser *)
	    


PROCEDURE LineEraser;

BEGIN
  IF buf^[LineStart[Line]] = EOF THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;


  ClearMarks;

  Lmark := LineStart[Line];
  Rmark := Lmark + ScanCode(FALSE, 1, ADR(buf^), Lmark);

  Eraser(TRUE);

END;   (* LineEraser *)



PROCEDURE Undo;

VAR i: INTEGER;
    pbase, stop, junk: bufInx;

BEGIN
  IF unmark = -1 THEN 
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;

  
  IF unstate = uncopy THEN

    BEGIN
      Lmark := unmark;
      Rmark := unmark + unlen;
      Eraser(FALSE);
    END
    
  ELSE IF unstate = unerase THEN

    BEGIN
      Lmark := -1; Rmark := -1;
      MakeSpace(unmark, unlen);
      
      MoveL(ADR(altbuf^[0]), ADR(buf^[unmark]), WRD(unlen));
      Remargin(TRUE, unmark, NewLine, pbase, stop);
      
      IF stop < (unmark + unlen) THEN
	Remargin(TRUE, unmark + unlen, NewLine, junk, stop);
      
      CursX := 8; CursY := CursHomeY; Line := 0;
      CNewPage(pbase);
      
      UpdateGauge;
    
    END
  
  ELSE (* unstate = unsymbol *)
    
    BEGIN
      IF undif < 0 THEN MakeSpace (unmark, -undif)
                   ELSE DeleteCode(unmark,  undif);

      MoveL(ADR(altbuf^[0]), ADR(buf^[unmark]), WRD(unlen));
  
      QErase;
      FOR i := 0 TO MaxLn DO BEGIN EraseLine(i); DispLine(i) END;
      QCursDraw;

    END;
  
  unmark := -1;
  
END;   (* Undo *)



PROCEDURE CMark;

VAR linx: bufInx;
    cnt:  INTEGER;

BEGIN
  IF (Lmark = -1) AND (Rmark = -1) THEN
  BEGIN
    Error(Msg[42].msg^);
    CmdExit;
  END;
    
  IF (curmark = Lmark) AND (Rmark <> -1) THEN
    curmark := Rmark
  ELSE 
    curmark := Lmark;
  
  IF NOT OnScreen(TRUE, curmark) THEN
  BEGIN
    linx := curmark;
	
    cnt := 1 + ORD(buf^[linx] < NewLine);
    
    WHILE (linx > 0) AND (cnt > 0) DO	 (* back up from mark *)
    BEGIN
      linx := linx + ScanCode(FALSE, -1, ADR(buf^), linx);
      cnt  := cnt - 1;
    END;
    
    CursX := 0;		  (* transitory cursor pos *)
    CursY := CursHomeY;
    CNewPage(linx);
    
  END;
  
  QErase;

  GetSignPos(TRUE, curmark, CursX, CursY, Line);

  IF CursX < CursLMar THEN CursX := CursLMar;
  CursY := CursY + CursHomeY;

  QCursDraw;
  
END;   (* CMark *)



PROCEDURE Jump;

VAR ch:  CHAR;
    cmd: cmdVal;

BEGIN
  WriteCmdName(Msg[68].msg^);
  
  Prompt(4, [], Msg[4].msg^, ch, cmd);
  
  IF ch <> CHR(esc) THEN 
    IF cmd = C_Jbegin THEN CBeginning
    ELSE IF cmd = C_Jend THEN CEnd
    ELSE CMark;
  
END;   (* Jump *)
	    


PROCEDURE Typer;

VAR ch:  CHAR;
    cmd: cmdVal;

BEGIN
  WriteCmdName(Msg[70].msg^);
  
  Prompt(5, [], Msg[5].msg^, ch, cmd);
  
  IF ch = CHR(esc) THEN CmdExit;
  
  TMode := Sign;
    
  IF cmd = C_Tfinger THEN TMode := Finger
  ELSE IF cmd = C_Talfa THEN TMode := Alpha;
      
END;   (* Typer *)
	    


PROCEDURE EGet;

VAR binx,  linx,
    fmark1, fmark2,
    stop, pbase, rbase: bufInx;
    i, byte, 
    cbytes, nbytes,
    curblk, lbytes: INTEGER;
    fmarks: BOOLEAN;
    name:   fName;
    junk:   WORD;
    s:      cString;
    junk2:  scrInx;
    f:	    blkFile;

BEGIN
  WriteCmdName(Msg[67].msg^);
  
  GetString(Msg[21].msg^, NameKeys, 8, TRUE, name);
  
  IF name = NULL THEN CmdExit;
  
  WriteLine(Msg[43].msg^);
  
  OpenFile(f, name, FALSE);

  GetFileInfo(f, name, fmark1, fmark2, nbytes);
  nbytes := nbytes - 2;
  
  fmarks := (fmark1 <> -1) AND (fmark2 <> -1);
  
  IF fmarks THEN nbytes := fmark2 - fmark1;
    
  IF (nbytes < 0) OR (nbytes > (Maxbuf - bufEnd)) THEN
  BEGIN
    Close(f);
    Error(Msg[22].msg^);
    CmdExit;
  END;
  
  
  ClearMarks;
  GetSignBound(junk2, binx);
  
  MakeSpace(binx, nbytes);
  
  lbytes := nbytes;
  linx	 := binx;
  
  IF fmarks THEN
    BEGIN curblk := 1 + (fmark1 DIV 512); byte := fmark1 MOD 512 END
  ELSE 
    BEGIN curblk := 1; byte := 0 END;

  { File pointer at blk 1 after GetFileInfo - move it to curblk } 
  FOR i := 1 TO (curblk-1) DO
    junk := GETUQQ(f, 512, ADS(altbuf^));

  WHILE lbytes > 0 DO
  BEGIN
    junk := GETUQQ(f, 4*512, ADS(altbuf^));
    IF f.errs > 0 THEN
    BEGIN
      DeleteCode(binx, nbytes);
      IOCheck(name, TRUE, f);
    END;
  
    cbytes := 2048 - BYTE;
    IF lbytes < cbytes THEN cbytes := lbytes;
    
    MoveL(ADR(altbuf^[byte]), ADR(buf^[linx]), WRD(cbytes));
    
    lbytes := lbytes - cbytes;
    linx   := linx + cbytes;
    byte   := 0;

  END;
  
  Close(f);
  IF f.errs > 0 THEN 
  BEGIN
    BuildMsg(s, Msg[36].msg^, name);
    Error(s);
  END;

 
  Remargin(TRUE, binx, NewLine, rbase, stop);
  
  pbase := LineStart[0];
  IF pbase = binx THEN pbase := rbase;
    
  IF stop < (binx + nbytes) THEN
    Remargin(TRUE, binx + nbytes, NewLine, rbase, stop);
  
  unmark  := binx;
  unlen	  := nbytes;
  unstate := uncopy;
  
  CNewPage(pbase);
  
  UpdateGauge;
  
END;   (* EGet *)



PROCEDURE Quit;

LABEL QuitExit;

VAR blks, bytes: INTEGER;
    linx: bufInx;
    name: fName;
    ch:	  CHAR;
    cmd:  cmdVal;
    f:	  blkFile;
    
BEGIN
  WriteCmdName(Msg[66].msg^);
  
  Prompt(6, [], Msg[6].msg^, ch, cmd);
  
  IF ch = CHR(esc) THEN CmdExit;

  
  IF cmd = C_Qexit THEN
  BEGIN
    Prompt(23, [], Msg[23].msg^, ch, cmd);

    IF cmd = C_Yes THEN
      BEGIN
        EExit := TRUE;
        UpdateCmdList := FALSE;

        GOTO QuitExit;

      END
    ELSE CmdExit;

  END;


  IF cmd <> C_Qbackup THEN  (* strip trailing blank lines *)

    WHILE (bufEnd >= 4) AND THEN (buf^[bufEnd-4] = NewLine) DO 
    BEGIN
      MoveL(ADR(buf^[bufEnd-2]), ADR(buf^[bufEnd-4]), 2); {move EOF code up}
      bufEnd := bufEnd - 2;
    END;

    
  IF cmd = C_Qwrite THEN  (* prompt for file name *)

    BEGIN
      GetString(Msg[24].msg^, PathKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit ELSE CheckFile(name);
    END
    
  ELSE
    BEGIN
      CopyLST(FileName, name);
      DeleteFile(name);
    END;

  
  IF cmd = C_Qsave   THEN WriteLine(Msg[44].msg^);
  IF cmd = C_Qwrite  THEN WriteLine(Msg[45].msg^);
  IF cmd = C_Qbackup THEN WriteLine(Msg[46].msg^);
  
  OpenFile(f, name, TRUE);
  FillC( ADR(altbuf^), 512, CHR(0));
  
  altbuf^[0] := SWversion;
  altbuf^[1] := SWhost;
  
  MoveL(ADR(Lmark),  ADR(altbuf^[3]), 2);
  MoveL(ADR(Rmark),  ADR(altbuf^[5]), 2);
  MoveL(ADR(bufEnd), ADR(altbuf^[7]), 2);
  
  Eval(PUTUQQ(f, 512, ADS(altbuf^)));
  IOCheck(name, FALSE, f);

  blks	:= 1 + (bufEnd DIV 512);
  bytes := bufEnd MOD 512;
  
  IF bytes > 0 THEN 
    FillC(ADR(buf^[bufEnd]), 512 - WRD(bytes), CHR(0))
  ELSE
    blks := blks - 1;


  Eval(PUTUQQ(f, WRD(512*blks), ADS(buf^)));
  IOCheck(name, FALSE, f);
  
  CloseFile(f, name, TRUE);

  
  IF cmd <> C_Qbackup THEN 	      (* return to file screen *)
  BEGIN
    EExit := TRUE;
    UpdateCmdList := FALSE;
  END;


QuitExit:
  
  IF DictChanged THEN	              (* update dict index *)
  BEGIN
    WriteLine(Msg[86].msg^);
    DictUpdate;
  END;

  IF cmd <> C_Qbackup THEN 	      (* close dictionary *)
    CloseDict;

END;   (* Quit *)
	    
		




  (* ---------------  Window edit commands  -------------- *)

	    
	    
PROCEDURE WCopy;

VAR nbytes: INTEGER;

BEGIN
  IF (wLmark = -1) OR (wRmark = -1) THEN
  BEGIN
    Error(Msg[16].msg^);
    CmdExit;
  END;
  
  IF wRmark > (wLmark + ScanCode(TRUE, 1, ADR(wEdbuf^), wLmark)) THEN
  BEGIN
    Error(Msg[25].msg^);
    CmdExit;
  END;
  
  nbytes := wRmark - wLmark;
  
  IF nbytes > (MaxWBuf - 2) THEN
  BEGIN
    Error(Msg[18].msg^);
    CmdExit;
  END;

  MoveL(ADR(buf^[0]), ADR(altbuf^[0]), WRD(bufEnd));
  
  unmark := 0;
  unlen	 := bufEnd;
  
  MoveL(ADR(wEdbuf^[wLmark]), ADR(buf^[0]), WRD(nbytes));
  
  buf^[nbytes]     := EOF;
  buf^[nbytes + 1] := 0;

  bufEnd := nbytes + 2;
  
  EraseWnd;
  DispLine(Line);
  QCursDraw;
  
END;   (* WCopy *)



PROCEDURE WErase;

BEGIN
  MoveL(ADR(buf^[0]), ADR(altbuf^[0]), WRD(bufEnd));
  
  unmark := 0;
  unlen	 := bufEnd;
  
  DeleteCode(0, BufEnd - 2);
  
  EraseWnd;
  DispLine(Line);
  QCursDraw;
  
END;   (* WErase *)



PROCEDURE WUndo;

BEGIN
  IF unmark = -1 THEN
  BEGIN
    FlashAlertBox(FALSE);
    CmdExit;
  END;
  
  MoveL(ADR(altbuf^[0]), ADR(buf^[0]), WRD(unlen));
  bufEnd := unlen;
  
  unmark := -1;
  
  EraseWnd;
  DispLine(Line);
  QCursDraw;

END;   (* WUndo *)


    
PROCEDURE WindowEdit (wbase: scrInx; wbuf: winPtr);
    
VAR sLineBase, sCursX, sCursY: scrInx;
    sLineStart, sbufEnd: bufInx;
    sLine: linInx;

BEGIN
  wTerm	 := CHR(ret);			     (* default termination *)
  

  wEdbuf := buf;  wLmark := Lmark;  wRmark := Rmark;	(* save env *)
  
  sLineStart := LineStart[MaxLn];  sbufEnd := bufEnd;
  sLineBase  := LineBase[MaxLn];   sLine   := Line;
  
  sCursX := CursX;  sCursY := CursY;
  
  
  Maxbuf := MaxWBuf;	                	    (* set wind env *)
  buf    := Retype(bufPtr, wbuf);

  Lmark := -1;   LineStart[MaxLn] := 0;
  Rmark := -1;   LineBase[MaxLn]  := wbase;

  bufEnd := 2 + ScanCode(FALSE, 1, ADR(wbuf^), 0);
  
  CursX := 8 + CursIn;   CursY  := wbase + CursHomeY;
  Line  := MaxLn;        unmark := -1;

  Window := TRUE;	      
  
  
  DispLine(Line);				    (* edit window *)

  IF DTxtWnd THEN DShowCurs
	     ELSE QCursDraw;

  WGetCommand;
  QErase;
  
  
  buf	 := wEdbuf;	  Lmark := wLmark;	    (* restore env *)
  Maxbuf := MaxEBuf;	  Rmark := wRmark;

  LineStart[MaxLn] := SLineStart;  bufEnd := sbufEnd;
  LineBase[MaxLn]  := SLineBase;   Line	  := sLine;
  
  CursX	 := sCursX;  CursY := sCursY;  unmark := -1;
  
  Window := FALSE;
  
END;   (* WindowEdit *)






    (* --------------  Find & Replace commands	-------------- *)



PROCEDURE Search (part, do_msg: BOOLEAN; targ: winPtr; VAR inx: bufInx);

VAR sinx, sbase, tinx, linx: bufInx;
    slen, tlen:	 INTEGER;
    ok,	  found: BOOLEAN;
    sym,  opt:	 BYTE;

BEGIN
  tlen := ScanCode(FALSE, 1, ADR(targ^), 0) - 2;
  sinx := inx; 
  
  inx := -1;
    
  REPEAT
    tinx := 2 * ORD(tlen >= 0);
    
    sinx := sinx + ScanSym(targ^[tinx], targ^[tinx + 1], ADR(buf^), sinx);

    IF buf^[sinx] = EOF THEN
    BEGIN
      IF do_msg THEN Error(Msg[47].msg^);
      RETURN;
    END;

    sbase := sinx + ScanCode(TRUE, -1, ADR(buf^), sinx) + 2;
    slen  := ScanCode(TRUE, 1, ADR(buf^), sbase);
       
    IF part OR (slen = tlen) THEN
    BEGIN
      ok := TRUE;
      
      buf^[sinx + 2] := buf^[sinx + 2] + 128;	 (* mark found sym *)
      tinx := tinx + 4;
      
      WHILE (targ^[tinx] < NewSign) AND ok DO
      BEGIN
	sym := targ^[tinx];
	opt := targ^[tinx + 1];
	
	linx  := sbase;
	found := FALSE;
      
	WHILE (buf^[linx] < NewSign) AND NOT found DO
	BEGIN
	  found := (buf^[linx	 ] = sym) AND 
		   (buf^[linx + 1] = opt) AND
		   (buf^[linx + 2] < 128);
    
   	  IF found THEN buf^[linx + 2] := buf^[linx + 2] + 128
	  ELSE linx := linx + 4;
	    
	END;
	    
	ok := found;
	tinx := tinx + 4;
      
      END;   (* while *)
	 
      linx := sbase;
      
      WHILE buf^[linx] < NewSign DO    (* unmark all syms *)
      BEGIN
	buf^[linx + 2] := buf^[linx + 2] AND 127;
	linx := linx + 4;
      END;
      
      IF ok THEN inx := sbase - 2;
	 
    END;   (* if *)
       
    sinx := sbase + slen;
    
  UNTIL inx >= 0;
       
END;   (* Search *)



PROCEDURE Find;

VAR   part:   BOOLEAN;
      curpos: bufInx;
      ch:     CHAR;
      cmd:    cmdVal;
      junk:   INTEGER;

BEGIN
  DrawFWnd(1);
  
  WriteCmdName(Msg[64].msg^);
  WriteLine(Msg[26].msg^);
  UpdateCmdList := TRUE;
  
  WindowEdit(FWndBase, fbuf);
  DrawFWnd(0);
  
  IF wTerm = CHR(esc) THEN CmdExit;

  GetSignBound(junk, curpos);
  
  WriteCmdName(Msg[64].msg^);
  WriteLine(Msg[7].msg^);
  UpdateCmdList := TRUE;
  
  part := FALSE;
  
  REPEAT
    Blinkcursor(FALSE);

    GetChar(7, [CHR(esc),CHR(ret)], ch, junk);
    cmd := SearchCmd(7, ch);

    IF cmd = C_Fnext THEN
      
      BEGIN
	QErase;
	Search(part, TRUE, fbuf, curpos);
	QCursDraw;
    
	IF curpos = -1 THEN CmdExit
	ELSE CSign(curpos);
	
	curpos := curpos + ScanCode(TRUE, 1, ADR(buf^), curpos);
	
      END
      
    ELSE IF cmd = C_Fpart THEN
      
      BEGIN
	part := NOT part;
	DrawPart(ORD(part));
      END;
    
  UNTIL (cmd = C_Fstop) OR (ch = CHR(esc)) OR (ch = CHR(ret));
  
END;   (* Find *)



PROCEDURE Replace;

LABEL RepExit;

VAR   fndlen, rplen, delta, junk: INTEGER;
      curpos, base:  bufInx;
      cmd,    lcmd:  cmdVal;
      ch,     lch:   CHAR;
      chnged: BOOLEAN;

BEGIN
  DrawRWnd(1);
  
  WriteCmdName(Msg[65].msg^);
  WriteLine(Msg[26].msg^);
  UpdateCmdList := TRUE;
  
  WindowEdit(FWndBase, fbuf);
  
  IF WTerm = CHR(esc) THEN 
  BEGIN DrawRWnd(0); CmdExit END;
  
  WriteCmdName(Msg[65].msg^);
  WriteLine(Msg[27].msg^);
  UpdateCmdList := TRUE;
  
  WindowEdit(RWndBase, rbuf);
  DrawRWnd(0);
  
  IF WTerm = CHR(esc) THEN CmdExit;
  
  GetSignBound(junk, curpos);
  ClearMarks;
  
  fndlen := ScanCode(FALSE, 1, ADR(fbuf^), 0);
  rplen	 := ScanCode(FALSE, 1, ADR(rbuf^), 0);
  delta	 := rplen - fndlen;
  
  WriteCmdName(Msg[65].msg^);
  WriteLine(Msg[8].msg^);
  UpdateCmdList := TRUE;
  
  junk := curpos;
  Search(FALSE, FALSE, fbuf, junk);
  chnged := junk <> curpos;
  
  REPEAT
    Blinkcursor(FALSE);

    GetChar(8, [CHR(esc),CHR(ret)], ch, junk);
    cmd := SearchCmd(8, ch);

    IF cmd = C_Rchange THEN   (* Change *)
      
      IF NOT chnged THEN
	BEGIN
	  QErase;
	  chnged := TRUE;
 	  IF delta > 0 THEN MakeSpace (curpos,	delta)
		       ELSE DeleteCode(curpos, -delta);
 	  MoveL(ADR(rbuf^[0]), ADR(buf^[curpos]), WRD(rplen));
	  IF rplen > 0 THEN buf^[curpos] := NewSign;
 	  Remargin(TRUE, curpos, NewLine, base, junk);
 	  IF base < LineStart[0] THEN CNewPage(base)
				 ELSE CNewPage(LineStart[0]);
	  CSign(curpos);
	
	END
	
      ELSE  {nothing}
    
    ELSE IF cmd = C_Rnext THEN   (* Next *)
    
      BEGIN
	curpos := curpos + ScanCode(TRUE, 1, ADR(buf^), curpos);
	chnged := FALSE;
	
	QErase;
	Search(FALSE, TRUE, fbuf, curpos);
	QCursDraw;
    
	IF curpos = -1 THEN CmdExit
	ELSE CSign(curpos);
	
      END 
    
    ELSE IF cmd = C_Rall THEN   (* All *)
      
      BEGIN
	QErase;

	Prompt(79, [], Msg[79].msg^, lch, lcmd);

	IF lcmd = C_Yes THEN
	  WriteLine(Msg[80].msg^)
        ELSE
	  BEGIN
	    WriteLine(Msg[8].msg^);
            CYCLE;   
	  END;

	base := curpos;
  
	REPEAT
	  IF NOT chnged THEN
	  BEGIN
	    IF delta > (MaxBuf - BufEnd) THEN
	      BEGIN Error(Msg[12].msg^); GOTO RepExit END
	    ELSE IF delta > 0 THEN
	      MakeSpace(curpos, delta)
	    ELSE
	      DeleteCode(curpos, -delta);
	    
	    MoveL(ADR(rbuf^[0]), ADR(buf^[curpos]), WRD(rplen));
	    IF rplen > 0 THEN buf^[curpos] := NewSign;

 	  END;

 	  curpos := curpos + ScanCode(TRUE, 1, ADR(buf^), curpos);
	  Search(FALSE, chnged, fbuf, curpos);

 	  IF chnged AND (curpos = -1) THEN CmdExit
	  ELSE chnged := FALSE;

 	UNTIL curpos = -1;
	
RepExit:
	Remargin(TRUE, base, EOF, junk, junk);
	
	LineStart[0] := MaxEBuf;   (* force screen update *)
	CSign(base);
	CmdExit;
      
      END;
      
  UNTIL (cmd = C_Rstop) OR (ch = CHR(esc)) OR (ch = CHR(ret));
    
END;   (* Replace *)






  (* ---------------  Dictionary text edit commands  ---------------- *)



PROCEDURE DSwitchWnd;

BEGIN
  QErase;

  DTxtWnd := NOT DTxtWnd;

  IF DTxtWnd THEN DShowCurs 
   	     ELSE QCursDraw;

END;  (* DSwitchWnd *)



PROCEDURE DUpdate (txtdelta, newpos, erasdelta: INTEGER);

VAR i:  INTEGER;
    eq: BOOLEAN;
    ls: DlsT;

BEGIN
  ls := Dls;  				       (* save jiggled line starts *)

  FOR i := (DCursLine + 1) TO DMaxLine DO
    IF ls[i] > 0 THEN ls[i] := ls[i] + txtdelta;


  DWrapText;			       	       (* get new line starts      *)	

  QErase;

  eq := TRUE;
  FOR i := 1 TO DMaxLine DO
    IF ls[i] <> Dls[i] THEN eq := FALSE;       (* compare line starts...   *)

  IF eq THEN              		       (* ...update cur line only  *)	
    BEGIN
      DEraseEOL(DCursLine, DCursCol + erasdelta);
      DDrawLine(DCursLine);

    END
  
  ELSE DWriteText;		       	       (* ...update all text lines *)	


  DPutCursInx(newpos);	       		       (* update cursor position   *)	

  DDrawCurs;


END;  (* DUpdate *)



PROCEDURE DAddChar {ch: CHAR};

VAR inx: INTEGER;

BEGIN
  DSaveCh := CHR(0);

  IF (dtxt.len = Upper(dtxt)) OR
     ((DCursLine = DMaxLine) AND (DCursCol >= DMaxCol)) THEN RETURN;

  DGetCursInx(inx);

  IF inx <= ORD(dtxt.len) THEN Insert(ch, dtxt, inx)
  ELSE
    BEGIN
      dtxt.len := dtxt.len + 1;
      dtxt[dtxt.len] := ch;
    END;

  DUpdate(1, inx + 1, 0);

END;  (* DAddChar *)



PROCEDURE DDeleteChar;

VAR inx: INTEGER;

BEGIN
  IF DCursCol > DLineLen(DCursLine) THEN RETURN;

  DGetCursInx(inx);

  DSaveCh := dtxt[inx];

  Delete(dtxt, inx, 1);

  DUpdate(-1, inx, 0);

END;  (* DDeleteChar *)



PROCEDURE DBackspaceChar;

VAR inx: INTEGER;

BEGIN
  IF DCursCol = 1 THEN RETURN;

  DGetCursInx(inx);

  DSaveCh := dtxt[inx - 1];

  Delete(dtxt, inx - 1, 1);

  DUpdate(-1, inx - 1, -1);

END;  (* DBackspaceChar *)



PROCEDURE DTxtUndo;

BEGIN

 IF DSaveCh <> CHR(0) THEN
   BEGIN DAddChar(DSaveCh); DSaveCh := CHR(0) END;

END;  (* DTxtUndo *)






   (* ------------------  Dictionary command  --------------------- *)



PROCEDURE DictUpdate;

VAR i: INTEGER;

BEGIN
  IOCheck(Msg[91].msg^, FALSE, DictF);
  

  Rewrite(DinxF);		    	      	(* rewrite dict index *)

  FillC(ADR(altbuf^), 512, CHR(0));
  
  altbuf^[0] := SWdict;
  
  MoveL(ADR(DictFree), ADR(altbuf^[1]), 2);
  MoveL(ADR(DictRecs), ADR(altbuf^[3]), 2);
  
  FOR i := 0 TO MaxDV DO
    WITH DictIndex[i] DO
      MoveL(ADR(cnt), ADR(altbuf^[5+(i*2)]), 2);

  MoveSL(ADS(Dcurmap^), ADS(altbuf^[17]), Sizeof(Dcurmap^));


  Eval(PUTUQQ(DinxF, 512, ADS(altbuf^)));

  
  FOR i := 0 TO MaxDV DO
    WITH DictIndex[i] DO
      Eval(PUTUQQ(DinxF, WRD(cnt) * Sizeof(dentry^[0]), ADS(dentry^)));


  IOCheck(Msg[91].msg^, FALSE, DinxF);


END;  (* DictUpdate *)



PROCEDURE GetDictF (recnum: WORD);

BEGIN
  Seek(DictF, recnum);
  Get(DictF);

  IOCheck(Msg[91].msg^, TRUE, DictF);

END;  (* GetDictF *)



PROCEDURE PutDictF (recnum: WORD);

BEGIN
  Seek(DictF, recnum);
  Put(DictF);

  IOCheck(Msg[91].msg^, FALSE, DictF);

END;  (* PutDictF *)



FUNCTION AllocDictRec: WORD;

BEGIN
  IF DictFree = DRecEOL THEN	       		(* append *)
    BEGIN

      IF DictRecs = MaxDRec THEN
        BEGIN Error(Msg[85].msg^); CmdExit END;

      DictRecs := DictRecs + 1;

      AllocDictRec := DictRecs;
  
    END


  ELSE						(* recycle *)
    BEGIN
      AllocDictRec := DictFree;

      GetDictF(DictFree);
      DictFree := DictF^.link;

    END;


END;  (* AllocDictRec *)



PROCEDURE DisposeDictRecs (recnum: WORD);

VAR newhead: WORD;

BEGIN
  newhead := recnum;

  REPEAT
    GetDictF(recnum);
   
    IF DictF^.link <> DRecEOL THEN
      recnum := DictF^.link;   

  UNTIL DictF^.link = DRecEOL;

  DictF^.link := DictFree;
  PutDictF(recnum);
  
  DictFree := newhead;


END;  (* DisposeDictRecs *)



FUNCTION DRead (vol, entry: INTEGER;
		buf: winPtr; VAR txt: DName): INTEGER;

VAR rec, cnt: WORD;
    bufinx:   INTEGER;

BEGIN
  bufinx := 0;

  WITH DictIndex[vol].dentry^[entry] DO
  BEGIN
    rec := recnum;
    CopyLST(key, txt);
  END;


  IF rec <> 0 THEN			(* valid dict entry *)

    REPEAT
      GetDictF(rec);
    
      IF DictF^.len > 0 THEN		(* read sym codes  *)
        BEGIN

          MoveL(ADR(DictF^.data[0]), ADR(buf^[bufinx]), WRD(DictF^.len));
          bufinx := bufinx + ORD(DictF^.len);

        END

      ELSE				(* read text codes *)
        BEGIN
          cnt := WRD(-DictF^.len);

          MoveL(ADR(DictF^.data[0]), ADR(txt[1 + txt.len]), cnt);
          txt.len := txt.len + cnt;

        END;


      IF DictF^.link <> DRecEOL THEN rec := DictF^.link;

    UNTIL DictF^.link = DRecEOL;


  IF bufinx = 0 THEN 			(* empty dict entry *)

    BEGIN
      buf^[0] := NewLine; buf^[1] := 0;
      buf^[2] := EOF;     buf^[3] := 0; 

      bufinx := 4;

    END;


  DRead := bufinx;


END;  (* DRead *)



PROCEDURE DWrite (vol, entry, buflen: INTEGER;
		  buf: winPtr; CONST txt: DName);

VAR rec:	      WORD;
    i, srecs, trecs:  INTEGER;
    bufinx, bufbytes: INTEGER;
    dictrec:          ARRAY [1..15] OF WORD;

BEGIN
  WITH DictIndex[vol] DO

  BEGIN
    IF cnt = 0 THEN RETURN; { empty dict! }

    rec := dentry^[entry].recnum;


    IF rec <> 0 THEN                                 (* flush dict entry  *)
    BEGIN
      DisposeDictRecs(rec);

      dentry^[entry].recnum := 0;

    END;

  END;


  srecs := 1 + (buflen - 1) DIV (MaxData + 1);	     (* sym recs to write *)


  trecs := 0;                     		     (* txt recs to write *)	

  IF txt.len > MaxDK THEN
    trecs := 1 + (ORD(txt.len) - (MaxDK + 1)) DIV (MaxData + 1);

  
  FOR i := 1 TO (srecs + trecs) DO
    dictrec[i] := AllocDictRec;            

    
  bufinx := 0;					     (* write sym recs *)	

  FOR i := 1 TO srecs DO
  BEGIN

    IF buflen > (MaxData + 1) THEN bufbytes := MaxData + 1
    			      ELSE bufbytes := buflen;

    MoveL(ADR(buf^[bufinx]), ADR(DictF^.data[0]), WRD(bufbytes));

    bufinx := bufinx + bufbytes;
    buflen := buflen - bufbytes;
  
    DictF^.len := bufbytes;

    IF (i = srecs) AND (trecs = 0) THEN DictF^.link := DRecEOL
    				   ELSE DictF^.link := dictrec[i+1];

    PutDictF(dictrec[i]);
    
  END;    


  bufinx := MaxDK + 1;           		     (* write txt recs *)
  buflen := ORD(txt.len) - MaxDK;

  FOR i := 1 TO trecs DO
  BEGIN

    IF buflen > (MaxData + 1) THEN bufbytes := MaxData + 1
    			      ELSE bufbytes := buflen;

    MoveL(ADR(txt[bufinx]), ADR(DictF^.data[0]), WRD(bufbytes));

    bufinx := bufinx + bufbytes;
    buflen := buflen - bufbytes;
  
    DictF^.len := -bufbytes;

    IF (i = trecs) THEN DictF^.link := DRecEOL
    		   ELSE DictF^.link := dictrec[srecs+i+1];

    PutDictF(dictrec[srecs+i]);
    
  END;    


  DictIndex[vol].dentry^[entry].recnum := dictrec[1];


END;  (* DWrite *)



PROCEDURE DDelete (vol, entry: INTEGER);

VAR rec, bytes: WORD;

BEGIN

  WITH DictIndex[vol] DO			     (* flush index entry *)	
  BEGIN
    IF entry >= cnt THEN RETURN;  {null dict entry}

    rec := dentry^[entry].recnum;

    bytes := WRD(cnt - entry - 1) * Sizeof(dentry^[0]);

    MoveSL(ADS(dentry^[entry+1]), ADS(dentry^[entry]), bytes);

    cnt := cnt - 1;

  END;


  IF rec <> 0 THEN DisposeDictRecs(rec);	     (* flush dict entry  *)


  DictNewVol := -1;                    (* flush any lingering 'new' entry *)	


END;  (* DDelete *)



PROCEDURE DAdd (vol, entry: INTEGER; buflen: INTEGER);

VAR bytes: WORD;

BEGIN

  WITH DictIndex[vol] DO			     (* add new index entry *)	
  BEGIN

    IF cnt > MaxDE THEN
      BEGIN Error(Msg[85].msg^); CmdExit END;


    bytes := WRD(cnt - entry) * Sizeof(dentry^[0]);

    MoveSR(ADS(dentry^[entry]), ADS(dentry^[entry+1]), bytes);


    MoveSL(ADS(dtxt), ADS(dentry^[entry].key), MaxDK + 1);

    IF dentry^[entry].key.len > MaxDK THEN	     (* build search key   *)
      dentry^[entry].key.len := MaxDK;


    dentry^[entry].recnum := 0;

    cnt := cnt + 1;

    
    DWrite(vol, entry, buflen, dbuf, dtxt);

  END;

  DictOnNew  := FALSE;  			     (* new entry official *)
  DictNewVol := -1;


END;  (* DAdd *)



PROCEDURE DSearchVal (VAR dictvol: INTEGER; VAR name: LSTRING);

VAR i, pos: INTEGER;

BEGIN
  UpCaseStr(name);


  IF (name.len > 1) AND (name[1] = '-') THEN       (* zap lead dash     *)
    Delete(name, 1, 1);


  i := 0;

  WHILE Dcurmap^.pmap[i].pair.len > 0 DO           (* map char pairs    *)

    WITH Dcurmap^.pmap[i] DO	            
    BEGIN
      pos := 1;

      REPEAT
        pos := Positn(pair, name, pos);
  
    	IF pos > 0 THEN
        BEGIN
          Delete(name, pos, 2);
	  Insert(CHR(sval), name, pos);

          pos := pos + 1;

	END;

      UNTIL pos = 0;

      i := i + 1;

    END; {with}


  FOR i := 1 TO ORD(name.len) DO	           (* map single chars *)	
    name[i] := CHR(Dcurmap^.cmap[ORD(name[i])]);


  IF WRD(name[1]) < Dcurmap^.cmap[ORD('M')] THEN   (* return dict vol  *)
    dictvol := 0

  ELSE 
    dictvol := 1;


END;  (* DSearchVal *)



FUNCTION DLookup {CONST name: DName; VAR vol, entry: INTEGER): BOOLEAN};

VAR found:    BOOLEAN;
    srchval,  entryval: DName;
    srchkey,  entrykey: DKey;
    x, lo, hi, junk:    INTEGER;

BEGIN
  srchval := name;

  DSearchVal(vol, srchval);                	   (* map srch name     *)


  MoveL(ADR(srchval), ADR(srchkey), MaxDK + 1);    (* build search key  *)

  IF srchkey.len > MaxDK THEN srchkey.len := MaxDK;


  found := FALSE;
  x := 0;

  WITH DictIndex[vol] DO			   (* search keys       *)
  
    IF cnt > 0 THEN
    BEGIN
      lo := 0;
      hi := cnt - 1;

      REPEAT
        x := (lo + hi) DIV 2;

        entrykey := dentry^[x].key;
						  
        DSearchVal(junk, entrykey);                (* map entry key     *)


        IF srchkey < entrykey THEN		   (* compare keys      *)
          hi := x - 1

        ELSE IF srchkey > entrykey THEN
          BEGIN
            lo := x + 1;
            x := lo; {ret insertion pt}
          END


        ELSE                			   (* keys match...     *)
	  BEGIN
            junk := DRead(vol, x, dbuf, dtxt);	   (* read dict entry   *)	

            entryval := dtxt;
					       	   
            DSearchVal(junk, entryval);            (* map entry name    *)


            IF srchval < entryval THEN	   	   (* compare names     *)	
              hi := x - 1

            ELSE IF srchval > entryval THEN
              BEGIN
                lo := x + 1;
                x := lo; {ret insertion pt}
              END

            ELSE found := TRUE;			   (* names match       *)	

          END;

      UNTIL found OR (lo > hi);


    END; { if cnt > 0 }


  entry := x;

  DLookup := found;


END;  (* DLookup *)



PROCEDURE DCheckUpdate (VAR buflen: INTEGER);

VAR i, n: INTEGER;
    junk: BOOLEAN;
    newname, srchname: DName;

BEGIN
  buflen := ScanCode(FALSE, 1, ADR(dbuf^), 0);

  i := 0;
  n := buflen + 2;  (* match EOF codes *)

  WHILE (i < n) AND (dbuf^[i] = dbufb^[i]) DO
    i := i + 1;


  IF i < n THEN					(* sign changed      *)

  BEGIN
    DictChanged := TRUE;

    IF buflen = 0 THEN				(* sign empty...     *)

      IF NOT DictOnNew THEN
	DDelete(DictCurVol, DictCurEnt)		(* delete old entry  *)

      ELSE
	{dead case: i=n}

    ELSE					(* sign nonempty...  *)
								     
      IF NOT DictOnNew THEN
        DWrite(DictCurVol, DictCurEnt,
	       n, dbuf, dtxt)			(* update old entry  *)

      ELSE
        DAdd(DictCurVol, DictCurEnt, n);	(* add new entry     *)

  END;


  IF (dtxt <> dtxtb) AND 
     (DictOnNew OR (buflen <> 0)) THEN	        (* text changed      *)

  BEGIN
    IF NOT DictOnNew THEN	      		(* delete cur entry  *)
      BEGIN
	DictChanged := TRUE;

        DDelete(DictCurVol, DictCurEnt);

      END

    ELSE DictNewVol := -1;


    DClipText;				     	(* add changed entry *)

    IF dtxt.len > 0 THEN
    BEGIN
      dbufb^ := dbuf^; {save for lookup}
      dtxtb  := dtxt;

      IF DLookup(dtxt, DictCurVol, DictCurEnt) THEN 
        IF NOT DictOnNew THEN
          Error(Msg[101].msg^);			(* dupl entry alert  *)
  
      dbuf^ := dbufb^; {restore}
      dtxt  := dtxtb;

      IF NOT DictOnNew THEN
        DAdd(DictCurVol, DictCurEnt, n)

      ELSE
        BEGIN
          DictName := dtxt;

          DictNewVol := DictCurVol;
          DictNewEnt := DictCurEnt;

        END;

    END;

  END;

	
END;  (* DCheckUpdate *)



PROCEDURE Dictionary;

VAR sinx, pbase, rbase, junk: bufInx;
    xpos:   scrInx;
    nbytes: INTEGER;

BEGIN
  IF NOT DictAvail THEN
    BEGIN Error(Msg[89].msg^); CmdExit END;


  DrawDWnd(1);						(* show window   *)
  
  WriteCmdName(Msg[81].msg^);
  WriteLine(Msg[82].msg^);

  UpdateCmdList := FALSE;

						 
  dtxt := NULL;	     			       		(* get sign name *)

  DWriteWndText;

  DTxtWnd  := TRUE;
  dbuf^[0] := EOF; dbuf^[1] := 0;


  WindowEdit(DWndBase, dbuf);


  DTxtWnd := FALSE;

  DClipText;

  DictName := dtxt;

  IF (wTerm = CHR(esc)) OR (DictName = NULL) THEN
    BEGIN DrawDWnd(0); UpdateCmdList := TRUE; CmdExit END;


  IF DLookup(DictName, DictCurVol, DictCurEnt) THEN 	(* old sign *)
    BEGIN
      DictOnNew := FALSE;

      DictNewVol := -1;
      DictNewEnt := -1;

      dbufb^ := dbuf^;
      dtxtb  := dtxt;

      DWriteWndText;

    END

  ELSE			 				(* new sign *)
    BEGIN
      DictOnNew := TRUE;
      
      DictNewVol := DictCurVol;
      DictNewEnt := DictCurEnt;

      dbuf^[0]  := EOF;  dbuf^[1]  := 0;
      dbufb^[0] := EOF;  dbufb^[1] := 0;

      dtxt  := DictName;
      dtxtb := dtxt;
	
    END;


  	      						
  Dict := TRUE;				 		(* window edit *)

  WriteLine(Msg[83].msg^);
  UpdateCmdList := TRUE;


  WindowEdit(DWndBase, dbuf);

  
  DrawDWnd(0);

  Dict := FALSE;
  DTxtWnd := FALSE;

  
  DCheckUpdate(nbytes);				 	(* update entry *)

  IF wTerm = CHR(esc) THEN CmdExit;



  ClearMarks;		   	                 (* copy sign into text *)
  			   
  GetSignBound(xpos, sinx);
  
  IF nbytes <= 2 THEN
  BEGIN
    Error(Msg[90].msg^);
    CmdExit;
  END;
  
  IF nbytes > (Maxbuf - bufEnd) THEN
  BEGIN
    Error(Msg[18].msg^);
    CmdExit;
  END;
  
  MakeSpace(sinx, nbytes);
  
  MoveL(ADR(dbuf^[0]), ADR(buf^[sinx]), WRD(nbytes));
  Remargin(TRUE, sinx, NewLine, rbase, junk);
  
  pbase := LineStart[0];
  IF pbase = sinx THEN pbase := rbase;
  
  unmark  := sinx;
  unlen	  := nbytes;
  unstate := uncopy;


  UpdateLineStart(pbase);			 (* update screen *) 
	      	
  IF OnScreen(FALSE, sinx) THEN
    CNewPage(pbase);	       

  CSign(sinx);
  CTab;

  UpdateGauge;


END;  (* Dictionary *)



PROCEDURE DWinRedraw;

VAR i:      INTEGER;
    nbytes: INTEGER;
    tname:  DName;

BEGIN
  IF DictOnNew THEN				(* empty buffer *)

    BEGIN
      dbuf^[0]  := EOF;  dbuf^[1]  := 0;
      dbufb^[0] := EOF;  dbufb^[1] := 0;

      dtxt := DictName;

    END

  ELSE						(* read in sign *)
    BEGIN
      nbytes := DRead(DictCurVol, DictCurEnt, dbuf, dtxt);

      MoveL(ADR(dbuf^[0]), ADR(dbufb^[0]), WRD(nbytes));

    END;


  dtxtb := dtxt;

  bufEnd := 2 + ScanCode(FALSE, 1, ADR(dbuf^), 0);

  unmark := -1;

  
  EraseWnd;
  DispLine(Line);				 (* display sign *)

  DWriteWndText;

  IF DTxtWnd THEN DShowCurs
	     ELSE QCursDraw;

END;  (* DWinRedraw *)



PROCEDURE DictNext;

VAR nxtvol, nxtent, nbytes, inc: INTEGER;

BEGIN
  DCheckUpdate(nbytes);					(* update current *)


  IF DictOnNew AND
     (DictCurEnt < DictIndex[DictCurVol].cnt) THEN
  
       DictOnNew := FALSE

  ELSE

    BEGIN
      IF (nbytes = 0) AND NOT DictOnNew THEN inc := 0	(* no advance... *)
					ELSE inc := 1;	(* after delete  *)


      nxtvol := DictCurVol;
      nxtent := DictCurEnt + inc;		(* advance to next *)


      IF (nxtent >= DictIndex[nxtvol].cnt)

         AND NOT ((nxtvol = DictNewVol)  AND 
                  (nxtent = DictNewEnt)) THEN   (* i.e. not new at end *)

      BEGIN
        nxtent := 0;

        REPEAT
          nxtvol := nxtvol + 1;

        UNTIL (nxtvol > MaxDV) OR ELSE

       	      ((DictIndex[nxtvol].cnt > 0) OR
	       (nxtvol = DictNewVol));		(* new in empty vol *)	

   
        IF nxtvol > MaxDV THEN			(* stop at the end *)
          BEGIN FlashAlertBox(FALSE); CmdExit END;


      END;


      DictCurVol := nxtvol;			(* update position *)
      DictCurEnt := nxtent;

      DictOnNew := (DictCurVol = DictNewVol) AND
                   (DictCurEnt = DictNewEnt);

    END; {else}


  DWinRedraw;					(* display new sign *)


END;  (* DictNext *)



PROCEDURE DictPrevious;

VAR nxtvol, nxtent, unused: INTEGER;

BEGIN
  DCheckUpdate(unused);				(* update current  *)


  IF (DictCurVol = DictNewVol) AND      	(* back up to prev *)
     (DictCurEnt = DictNewEnt) AND
     NOT DictOnNew THEN
  
       DictOnNew := TRUE

  ELSE
    BEGIN
      nxtvol := DictCurVol;
      nxtent := DictCurEnt - 1;


      IF (nxtent < 0) THEN                      (* bounds check *)

      BEGIN
        REPEAT
          nxtvol := nxtvol - 1;

        UNTIL (nxtvol < 0) OR ELSE

       	      ((DictIndex[nxtvol].cnt > 0) OR
	       (nxtvol = DictNewVol));		(* new in empty vol *)	


        IF nxtvol < 0 THEN			(* stop at the front *)
          BEGIN FlashAlertBox(FALSE); CmdExit END;


        WITH DictIndex[nxtvol] DO
         
          IF (nxtvol = DictNewVol) AND (cnt = DictNewEnt) THEN
            nxtent := cnt

          ELSE 
            nxtent := cnt - 1;

      END; {if}


      DictCurVol := nxtvol;			(* update position *)
      DictCurEnt := nxtent;

      DictOnNew := (DictCurVol = DictNewVol) AND
                   (DictCurEnt = DictIndex[DictCurVol].cnt);

    END; {else}


  DWinRedraw;					(* display new sign *)


END;  (* DictPrevious *)






   (* ---------------  Window edit command interp --------------- *)



PROCEDURE WCommand;

VAR code:  INTEGER;
    state: ShiftState;
    ch:	   CHAR;
    cmd:   cmdVal;
    save:  BOOLEAN;
    r:     InxRec;

BEGIN
  GetKey(code, state);
    
  IF (state = Alt) OR (state = CtrlShft) THEN  	       (* check for cmds *)
    
    BEGIN
      IF state = CtrlShft THEN               (* get cursor cmd code      *)
        r := Keyboard^[Alpha, code, state]
      ELSE
        r := Keyboard^[Alpha, code, Shft];   (* look for (upcase) ch val *)

      IF NoCharVal(r) THEN EnterKey(code, state)
      
      ELSE BEGIN
        ch  := CHR(r.sym.codeval);
        cmd := GetCmdVal(ch);

        save := CursSelect;
        CursSelect := FALSE;

        IF SymSelect THEN 

          IF NOT ((ch IN [CHR(up), CHR(down), CHR(right), CHR(left)])
                   OR (cmd IN [C_Select, C_Copy, C_Erase]))

          THEN DeSelect ELSE {nothing}
       
        ELSE IF NOT (save OR DTxtWnd) THEN BEGIN  (* update cursor *)
          QErase; 
          QCursDraw;
	END;

        IF SymSelect THEN
      
  	  IF      ch = CHR(right) THEN SMoveX(+8)
  	  ELSE IF ch = CHR(left)  THEN SMoveX(-8)
  	  ELSE IF ch = CHR(down)  THEN SMoveY(+4)
  	  ELSE IF ch = CHR(up)	  THEN SMoveY(-4)

          ELSE CASE cmd OF
    	    C_Copy:   SCopy;
	    C_Erase:  SErase;
	    C_Select: DeSelect;
            OTHERWISE;
          END


	ELSE IF DTxtWnd THEN

          IF (ch = 'W') AND Dict THEN DSwitchWnd

          ELSE CASE cmd OF
      	    C_Undo:     DTxtUndo;
	    C_Next:     IF Dict THEN DictNext;
	    C_Previous: IF Dict THEN DictPrevious;
            C_more,
            C_more2:    IF Dict THEN
	    	        BEGIN 
       	                  altCmd := NOT altCmd;
      	                  IF altCmd THEN WriteLine(Msg[2].msg^)
                                    ELSE WriteLine(Msg[1].msg^);
    	                END;
            OTHERWISE;
          END


        ELSE IF ch <= ' ' THEN

    	  IF      ch = CHR(right) THEN LMoveX(+8)
  	  ELSE IF ch = CHR(left)  THEN LMoveX(-8)
  	  ELSE IF ch = CHR(down)  THEN LMoveY(+4)
  	  ELSE IF ch = CHR(up)    THEN LMoveY(-4)
	  ELSE IF ch = CHR(ret)   THEN AlphaReturn
          ELSE {nothing}


        ELSE IF ch = CHR(del) THEN WErase

        ELSE IF (ch = 'W') AND Dict THEN DSwitchWnd

        ELSE IF ch = 'X' THEN BEGIN TMode := Sign;   UpdateType END
        ELSE IF ch = 'Y' THEN BEGIN TMode := Finger; UpdateType END
        ELSE IF ch = 'Z' THEN BEGIN TMode := Alpha;  UpdateType END
      
        ELSE CASE cmd OF
  	  C_Adjust:   Adjust;
	  C_Copy:     WCopy;
	  C_Erase:    WErase;
	  C_Select:   Select;
	  C_Type:     Typer;
	  C_Undo:     WUndo;
	  C_Next:     IF Dict THEN DictNext;
	  C_Previous: IF Dict THEN DictPrevious;

          C_more,
          C_more2:   BEGIN 
       	               altCmd := NOT altCmd;
      	               IF altCmd THEN WriteLine(Msg[2].msg^)
                                 ELSE WriteLine(Msg[1].msg^);
    	             END;
        
          OTHERWISE  BEGIN CursSelect := save; EnterKey(code, state) END;
        END
      
      END

    END
  
  ELSE	  (* sym key *)
    
    BEGIN
      r := Keyboard^[Alpha, code, state];   (* look for ch val *)

      IF NoCharVal(r) THEN EnterKey(code, state)
      
      ELSE BEGIN
        ch := CHR(r.sym.codeval);
      
        save := CursSelect;
        
        IF CursSelect THEN 
      	  CursSelect := (code = SC_Cursor);
	
        IF SymSelect THEN 

	  IF CtrlCmd(code, state)
             OR (ch IN [CHR(bs)  , CHR(del) , CHR(ret) , CHR(esc), 
		        CHR(pgup), CHR(pgdn), CHR(home), CHR(hend)])

	  THEN DeSelect;
      
        IF SymSelect THEN 
      
	  IF ch = CHR(right)     THEN SMoveX(+2)
	  ELSE IF ch = CHR(left) THEN SMoveX(-2)
	  ELSE IF ch = CHR(down) THEN SMoveY(+1)
	  ELSE IF ch = CHR(up)	 THEN SMoveY(-1)

 	  ELSE CASE code OF
            SC_Change:  IF state = Shft THEN SChange(-1)
                        ELSE IF state = None THEN SChange(+1);

            SC_Mirror:	SMirror;

            SC_Size:    IF state = Shft THEN SSize(FALSE, -1)
                        ELSE IF state = None THEN SSize(FALSE, +1);
	    
            SC_Cursor:  IF state = Shft THEN SCursor(-1)
                        ELSE IF state = None THEN SCursor(+1);
	
            SC_Rotate:  IF state = Shft THEN SRotate(-1)
                        ELSE IF state = None THEN SRotate(+1);
            OTHERWISE;
	  END

      
        ELSE IF CursSelect THEN
      
	  IF state = Shft THEN SCursor(-1)
	  ELSE IF state = None THEN SCursor(+1) ELSE {nothing}


	ELSE IF DTxtWnd THEN

          IF ch IN DNameKeys THEN DAddChar(ch)

          ELSE IF ch = CHR(del) THEN DDeleteChar

          ELSE CASE ch OF
            CHR(bs):    DBackspaceChar;

            CHR(left):  DMoveCursX(-1);
            CHR(right): DMoveCursX(+1);
            CHR(up):    DMoveCursY(-1);
            CHR(down):  DMoveCursY(+1);

            CHR(home):  DMoveCursX(1 - DCursCol);
            CHR(hend):  DMoveCursX(DLineLen(DCursLine) - DCursCol + 1);

            CHR(pgdn):  IF Dict THEN DictNext;
            CHR(pgup):  IF Dict THEN DictPrevious;

            CHR(ret):   WExit := TRUE;
            CHR(esc):   BEGIN wTerm := CHR(esc); WExit := TRUE END;

            OTHERWISE;
	  END


        ELSE IF ch = ' ' THEN
          IF TMode = Alpha THEN LMoveX(+8) ELSE {nothing}
      
        ELSE IF ch < ' ' THEN

          IF CtrlCmd(code, state) THEN

    	    IF      ch = CHR(right) THEN SignMove(+4, 0)
  	    ELSE IF ch = CHR(left)  THEN SignMove(-4, 0)
  	    ELSE IF ch = CHR(down)  THEN SignMove( 0,+2)
  	    ELSE IF ch = CHR(up)    THEN SignMove( 0,-2)
  	    ELSE IF ch = CHR(home)  THEN SignCenter
            ELSE {nothing}
      
	  ELSE 

            CASE ch OF
              CHR(right): LMoveX(+2);
              CHR(left):  LMoveX(-2);
	      CHR(down):  LMoveY(+1);
  	      CHR(up):    LMoveY(-1);
	      CHR(fsp):   LMoveX(+8);
	      CHR(ret):   WExit := TRUE;
	      CHR(bs):    Deleter;
              CHR(home):  CMoveX(CursWMar - CursX);
              CHR(hend):  CEndLine;
              CHR(pgdn):  IF Dict THEN DictNext;
              CHR(pgup):  IF Dict THEN DictPrevious;
  	      CHR(esc):   BEGIN  
	                    wTerm := CHR(esc);
                            WExit := TRUE;
 	                  END;
              OTHERWISE;
            END
      
        ELSE IF ch = CHR(del) THEN Deleter
	
        ELSE BEGIN
	  CursSelect := save;
	  EnterKey(code, state);
        END;
    
      END;

    END;  (* sym key *)
  
END;   (* WCommand *)






   (* --------------  Edit screen command interp  --------------- *)



PROCEDURE ECommand;

VAR code:  INTEGER;
    state: ShiftState;
    ch:	   CHAR;
    cmd:   cmdVal;
    save:  BOOLEAN;
    r:     InxRec;
    
BEGIN
  GetKey(code, state);
    
  IF (state = Alt) OR (state = CtrlShft) THEN   (* check for cmds *)
    
    BEGIN
      IF state = CtrlShft THEN               (* get cursor cmd code *)
        r := Keyboard^[Alpha, code, state]
      ELSE
        r := Keyboard^[Alpha, code, Shft];   (* look for (uppercase) ch val *)

      IF NoCharVal(r) THEN EnterKey(code, state)
      
      ELSE BEGIN
        ch := CHR(r.sym.codeval);
        cmd := GetCmdVal(ch);

        save := CursSelect;
        CursSelect := FALSE;

        IF SymSelect THEN 

          IF NOT ((ch IN [CHR(up), CHR(down), CHR(right), CHR(left)])
                   OR (cmd IN [C_Select, C_Copy, C_Erase]))

          THEN DeSelect ELSE {nothing}
       
        ELSE IF NOT (save OR DTxtWnd) THEN BEGIN  (* update cursor *)
          QErase; 
          QCursDraw;
	END;
    
        IF SymSelect THEN
      
  	  IF      ch = CHR(right) THEN SMoveX(+8)
  	  ELSE IF ch = CHR(left)  THEN SMoveX(-8)
  	  ELSE IF ch = CHR(down)  THEN SMoveY(+4)
  	  ELSE IF ch = CHR(up)	  THEN SMoveY(-4)

          ELSE CASE cmd OF
    	    C_Copy:   SCopy;
	    C_Erase:  SErase;
	    C_Select: DeSelect;
            OTHERWISE;
          END
      
        ELSE IF ch < ' ' THEN

    	  IF      ch = CHR(right) THEN CMoveX(+8)
  	  ELSE IF ch = CHR(left)  THEN CMoveX(-8)
  	  ELSE IF ch = CHR(down)  THEN CMoveY(+4)
  	  ELSE IF ch = CHR(up)    THEN CMoveY(-4)
	  ELSE IF ch = CHR(ret)   THEN AlphaReturn
          ELSE {nothing}


        ELSE IF ch = CHR(del) THEN LineEraser

        ELSE IF ch = ' ' THEN Space

        ELSE IF ch = 'X' THEN BEGIN TMode := Sign;   UpdateType END
        ELSE IF ch = 'Y' THEN BEGIN TMode := Finger; UpdateType END
        ELSE IF ch = 'Z' THEN BEGIN TMode := Alpha;  UpdateType END
      
        ELSE CASE cmd OF
	  C_Adjust:   Adjust;
	  C_Copy:     Copy;
	  C_Dict:     Dictionary;
	  C_Erase:    Eraser(TRUE);
	  C_Find:     Find;
	  C_Get:      EGet;
	  C_Jump:     Jump;
	  C_Margin:   Margin;
	  C_Next:     CMoveY(+57);
	  C_Previous: CMoveY(-57);
	  C_Quit:     Quit;
	  C_Replace:  Replace;
	  C_Select:   Select;
	  C_Type:     Typer;
	  C_Undo:     Undo;
          
          C_more,
          C_more2:    BEGIN 
       	                altCmd := NOT altCmd;
      	                IF altCmd THEN WriteLine(Msg[2].msg^) 
                                  ELSE WriteLine(Msg[1].msg^);
    	              END;
        
          OTHERWISE   BEGIN CursSelect := save; EnterKey(code, state) END;
        END
      
      END

    END

  ELSE	  (* sym key *)
    
    BEGIN
      r := Keyboard^[Alpha, code, state];   (* look for ch val *)

      IF NoCharVal(r) THEN EnterKey(code, state)
      
      ELSE BEGIN
        ch := CHR(r.sym.codeval);
      
        save := CursSelect;
        
        IF CursSelect THEN 
      	  CursSelect := (code = SC_Cursor);
	
        IF SymSelect THEN 

	  IF CtrlCmd(code, state)
             OR (ch IN [' ',       CHR(tab),  CHR(bs),   CHR(del),  
			CHR(ret),  CHR(home), CHR(hend),
		        CHR(pgup), CHR(pgdn), ' '])
	  
	  THEN DeSelect;
      
        IF SymSelect THEN 
      
	  IF ch = CHR(right)     THEN SMoveX(+2)
	  ELSE IF ch = CHR(left) THEN SMoveX(-2)
	  ELSE IF ch = CHR(down) THEN SMoveY(+1)
	  ELSE IF ch = CHR(up)	 THEN SMoveY(-1)

 	  ELSE CASE code OF
            SC_Change:  IF state = Shft THEN SChange(-1)
                        ELSE IF state = None THEN SChange(+1);

            SC_Mirror:	SMirror;

            SC_Size:    IF state = Shft THEN SSize(FALSE, -1)
                        ELSE IF state = None THEN SSize(FALSE, +1);
	    
            SC_Cursor:  IF state = Shft THEN SCursor(-1)
                        ELSE IF state = None THEN SCursor(+1);
	
            SC_Rotate:  IF state = Shft THEN SRotate(-1)
                        ELSE IF state = None THEN SRotate(+1);
            OTHERWISE;
	  END
      
        ELSE IF CursSelect THEN
      
	  IF state = Shft THEN SCursor(-1)
	  ELSE IF state = None THEN SCursor(+1) ELSE {nothing}
      
        ELSE IF ch = ' ' THEN
          IF TMode = Alpha THEN CMoveX(+8) ELSE Space

        ELSE IF ch < ' ' THEN

          IF CtrlCmd(code, state) THEN

    	    IF      ch = CHR(right) THEN SignMove(+4, 0)
  	    ELSE IF ch = CHR(left)  THEN SignMove(-4, 0)
  	    ELSE IF ch = CHR(down)  THEN SignMove( 0,+2)
  	    ELSE IF ch = CHR(up)    THEN SignMove( 0,-2)
  	    ELSE IF ch = CHR(home)  THEN SignCenter
            ELSE {nothing}
      
	  ELSE 
      
	    CASE ch OF
              CHR(right): CMoveX(+2);
	      CHR(left):  CMoveX(-2);
	      CHR(down):  CMoveY(+1);
  	      CHR(up):	  CMoveY(-1);
	      CHR(fsp):	  CMoveX(+8);
	      CHR(tab):	  IF state = Shft THEN CBackTab ELSE CTab;
	      CHR(ret):	  Returner;
	      CHR(bs):    Deleter;
              CHR(home):  CMoveX(CursLMar - CursX);
              CHR(hend):  CEndLine;
              CHR(pgdn):  CPageDown;
              CHR(pgup):  CPageUp;
              OTHERWISE;
             END

        ELSE IF ch = CHR(del) THEN Deleter
	
        ELSE BEGIN
	  CursSelect := save;
	  EnterKey(code, state);
        END;
      
      END;

    END;  (* sym key *)
  
END;   (* ECommand *)






    (* ---------------	File screen commands  ------------------ *)



PROCEDURE LoadFile (VAR f: blkFile; CONST name: fName);

VAR nbytes, junk: INTEGER;

BEGIN
  WriteLine(Msg[55].msg^);
    
  OpenFile(f, name, FALSE);
  GetFileInfo(f, name, Lmark, Rmark, nbytes);
      
  UpdateFileList := TRUE;
      
  Eval(GETUQQ(f, WRD(nbytes), ADS(buf^)));   (* GetFileInfo ate header *)
  IOCheck(name, TRUE, f);
      
  CloseFile(f, name, TRUE);
      
  bufEnd := nbytes;
  unmark := -1;

END;  (* LoadFile *)



PROCEDURE ShowFile (statmsg: INTEGER);

BEGIN
  Lmark := -1;	Rmark := -1;
  
  ClearScreen;
  
  DrawGauge(1);
  WriteLine(Msg[statmsg].msg^);
  
  CursX := 8;  NewX := 8;  CursY := CursHomeY;
  QCursDraw;
  CNewPage(0);
  QErase;

  UpdateFileScrn := TRUE;

END;  (* ShowFile *)



PROCEDURE Edit (new: BOOLEAN);

VAR f:	    blkFile;
    name:   fName;
    nbytes: INTEGER;
    junk:   BOOLEAN;

BEGIN
  IF new THEN  (* create new file *)
    
    BEGIN
      WriteCmdName(Msg[72].msg^);
      
      GetString(Msg[48].msg^, NameKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit ELSE CheckFile(name);
      
      UpdateFileList := TRUE;
      Lmark := -1;  Rmark := -1;  unmark := -1;
      
      buf^[0] := EOF;	   (* empty edit buf *)
      buf^[1] := 0;
      bufEnd  := 2;
      
    END
    
  ELSE  (* edit existing file *)
    
    BEGIN
      WriteCmdName(Msg[71].msg^);
      
      GetString(Msg[49].msg^, NameKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit;
      
      LoadFile(f, name);
      
    END;
  
  CopyLST(name, FileName);
  
  
  IF DictAvail THEN junk := OpenDict;   (* open dictionary *)


  ClearScreen;			        (* set up edit screen *)
  
  DrawBorders;
  
  CursX := 8 + CursIn;  CursY := CursHomeY;  Line := 0;

  SymSelect  := FALSE;	altCmd	:= FALSE;     
  CursSelect := FALSE;	UpdateCmdList := FALSE;
  
  TMode := Sign;

  FScreen := FALSE;

  DictChanged := FALSE;
  
  QCursDraw;
  CNewPage(0);
  
  DrawGauge(1);
  WriteLine(Msg[1].msg^);
  
  EGetCommand;			     (* edit file *)
  

  ClearScreen;			     (* restore file screen *)
  
  FScreen := TRUE;
  
  DrawFileBox;
  DrawGauge(1);
  
END;   (* Edit *)



PROCEDURE Directory;

CONST ChangeDir  = 16#3B;
      ChangeDisk = 16#0E;

VAR name: cString;
    
BEGIN
  WriteCmdName(Msg[75].msg^);

  GetString(Msg[28].msg^, PathKeys, 50, TRUE, name);
  IF name = NULL THEN CmdExit;

  IF name[name.len] = ':' THEN Concat(name,'.');
  Concat(name, CHR(0));

  IF DOSXQQ(ChangeDir, (ADS name[1]).R) <> 0 THEN
  BEGIN
    Error(Msg[60].msg^);
    CmdExit;
  END;

  UpdateFileList := TRUE;

  IF (name.len >= 2) AND THEN (name[2] = ':') THEN
    Eval(DOSXQQ(ChangeDisk, WRD(name[1]) - WRD('A')));

  
  ResetDict;  (* compensate for path change *)


END;  (* Directory *)



PROCEDURE FQuit;

VAR junk: CHAR;
    cmd:  cmdVal;

BEGIN
  WriteCmdName(Msg[66].msg^);
  
  Prompt(58, [], Msg[58].msg^, junk, cmd);
  FExit := (cmd = C_Yes);

END;   (* FQuit *)






    (* -------------------  Print command  --------------------- *)



PROCEDURE SetCmdStr (VAR  cstr: LSTRING;
		     CONST pat: STRING; CONST val: LSTRING);
VAR i: INTEGER;
BEGIN
  i := Positn(pat, cstr, 1);
  Delete(cstr, i, UPPER(pat));
  Insert(val, cstr, i);

END;  (* SetCmdStr *)



PROCEDURE HPCompress (CONST gdata: GrafData; VAR bufbytes: INTEGER);

VAR cb, next, inx, cnt, lastrep: INTEGER;

BEGIN
  inx  := 1;

  cb   := 0;
  next := 1;

  lastrep := -1;
  

  altbuf^[0] := 255;

  WHILE inx <= Upper(gdata) DO
  BEGIN

    cnt := ScanNE(Upper(gdata) - inx + 1, gdata[inx], gdata, inx);

    IF cnt > 1 THEN
      BEGIN
	IF cb <> (next - 1) THEN cb := next;

      	altbuf^[cb    ] := WRD(-(cnt - 1));
     	altbuf^[cb + 1] := WRD(gdata[inx]);
     	altbuf^[cb + 2] := 255;

        lastrep := cb;

        cb   := cb + 2;
        next := cb + 1;

      END

    ELSE
      BEGIN
     	altbuf^[cb  ] := altbuf^[cb] + 1;
     	altbuf^[next] := WRD(gdata[inx]);

        next := next + 1;

      END;

    inx := inx + cnt;

  END;  (* while *)


  IF (lastrep = (next - 3)) THEN 
    IF (altbuf^[lastrep + 1] = 0) THEN  {zap trail blanks}
      next := next - 3;                


  bufbytes := next;


END;  (* HPCompress *)



PROCEDURE PSCompress (CONST gdata: GrafData; VAR bufbytes: INTEGER);

VAR inx, cnt, blankcnt: INTEGER;
    hexstr:   LSTRING(2);
    b:        BYTE;

BEGIN
  inx      := 0;
  blankcnt := 0;

  FOR cnt := 1 TO Upper(gdata) DO
  BEGIN
    b := WRD(gdata[cnt]) XOR #FF;

    IF b = #FF THEN blankcnt := blankcnt + 1
       	       ELSE blankcnt := 0;

    IF Encode(hexstr, b:2:16) THEN;

    altbuf^[inx    ] := WRD(hexstr[1]);
    altbuf^[inx + 1] := WRD(hexstr[2]);

    inx := inx + 2;

  END;

  bufbytes := inx - (blankcnt * 2);   {zap trail blanks}


END;  (* PSCompress *)



PROCEDURE Print;

TYPE lcmdstr = LSTRING(150);

VAR linecmd, colcmd,  prntcmd,  prlncmd,  rulecmd, numstr:  cString;
    i,       colcnt,  maxcol,   junk,     bytes:   INTEGER;
    pgnum,   pgline,  linecnt,  linepart, size:    INTEGER;
    hp,      ps,      nonblank: BOOLEAN;
    pagecmd, cmdstr:  lcmdstr;
    name:    fName;
    gstr:    GrafData;
    f:	     blkFile;
    ch:      CHAR;



    PROCEDURE CheckIO;

    VAR s: cString;

    BEGIN
      IF f.errs > 0 THEN
      BEGIN
        BuildMsg(s, Msg[78].msg^, name);
        Error(s); Close(f); CmdExit;
      END;

    END;  (* CheckIO *)



    PROCEDURE PutString (CONST str: LSTRING);
    
    BEGIN
      Eval(PUTUQQ(f, WRD(str.Len), ADS(str[1])));

    END;  (* PutString *)



    PROCEDURE PutPageNum (pgnum, xpos, ypos: INTEGER);

    VAR pgcmd, str: cString;

    BEGIN
      IF hp THEN CopyLST(HPpgnm, pgcmd)
      	    ELSE CopyLST(PSpgnm, pgcmd);

      IF Encode(numstr, pgnum:1) THEN;
      xpos := xpos - PFcharbias * (ORD(numstr.Len) - 1);  { center # }

      IF Encode(str, xpos:1) THEN;
      SetCmdStr(pgcmd, '!', str);	

      IF Encode(str, ypos:1) THEN;
      SetCmdStr(pgcmd, '%', str);	

      PutString(pgcmd);

      IF hp THEN
        PutString(numstr)

      ELSE {ps}
	BEGIN
      	  CopyLST(PStext, pgcmd);

          SetCmdStr(pgcmd, '#', numstr);	
          PutString(pgcmd);

	END;

    END;  (* PutPageNum *)



    PROCEDURE WritePage;

    VAR i, left: INTEGER;

    BEGIN
      IF PageNum <> NoPgNm THEN   	   (* print page number *)

        IF hp or ps THEN
    
          IF HPpgor[ORD(PageForm)] = '1' THEN  		{wide}

            IF PFncol[ORD(PageForm)] > 1 THEN  		{2-page format}

              BEGIN
                PutPageNum(pgnum, PFpgnmLW, PFpgnmW);

                pgnum := pgnum + 1; 

                PutPageNum(pgnum, PFpgnmRW, PFpgnmW);

      	      END

            ELSE PutPageNum(pgnum, PFpgnmCW, PFpgnmW)


          ELSE PutPageNum(pgnum, PFpgnmCT, PFpgnmT)     {tall}


        ELSE {Epson}

          BEGIN
            IF pgline <> 1 THEN	{skip to end of page}
            BEGIN
              left  := ORD(PageLen) - pgline;

	      FOR i := 1 TO ((left + 1) * (linecnt + 1)) DO
                PutString(EpsRet);

	      FOR i := 1 TO left DO PutString(linecmd);

            END;

            FOR i := 1 TO 8 DO PutString(linecmd);
	    FOR i := 1 TO EpsPgnm DO PutString(' ');
            
	    IF Encode(numstr, pgnum:-3) THEN;
	    PutString(numstr);

          END;


      PutString(pagecmd);

      pgnum := pgnum + 1;

    END;  (* WritePage *)



    PROCEDURE PrintFinish (ok: BOOLEAN);
    BEGIN
      IF ps THEN
        IF ok THEN PutString(PSterm)
	      ELSE PutString(PSesc);	

      Close(f);

    END;  (* PrintFinish *)



BEGIN  (* Print *)

  hp := (PrnType = PrHP);
  ps := (PrnType = PrPS);


  WriteCmdName(Msg[73].msg^);
  
  GetString(Msg[54].msg^, NameKeys, 8, TRUE, name);
  IF name = NULL THEN CmdExit;
  
  LoadFile(f, name);
  ShowFile(56);

  Line	 := 0;
  pgline := 1;
  pgnum  := ORD(PageNum);


  Assign(f, PortName[ORD(PrnPort)]);
  f.trap := TRUE;  
  Rewrite(f);

  CheckIO;


  IF hp OR ps THEN

    BEGIN
      linecnt := MaxLnHt + 1;

      colcnt := 1;
      maxcol := PFncol[ORD(PageForm)];
 
      IF hp THEN

        BEGIN
          CopyLST(HPinit, cmdstr);
          CopyLST(HPcol ,  colcmd);
          CopyLST(HPline, linecmd);
          CopyLST(HPpage, pagecmd);
          CopyLST(HPrule, rulecmd);
      
	  SetCmdStr(cmdstr, '@', HPpgor[ORD(PageForm)]);	
      	  SetCmdStr(cmdstr, '?', HPsize[ORD(PageForm)]);	

          CopyLST(HPprnt, prntcmd);

        END

      ELSE {ps}

        BEGIN
          CopyLST(PSinit, cmdstr);
          CopyLST(PScol ,  colcmd);
          CopyLST(PSline, linecmd);
          CopyLST(PSpage, pagecmd);
          CopyLST(PSrule, rulecmd);

      	  SetCmdStr(cmdstr, '@', PSpgor[ORD(PageForm)]);	
	  SetCmdStr(cmdstr, '#', PSleft[ORD(PageForm)]);	
	  SetCmdStr(cmdstr, '?', PStop [ORD(PageForm)]);	

          CopyLST(PSprnt, prntcmd);
          CopyLST(PSline, prlncmd);

          size := PSsize[ORD(PageForm)];

          IF Encode(numstr, size:1) THEN;
          SetCmdStr(prntcmd, '!', numstr);	

          IF Encode(numstr, (size*2):1) THEN;
          SetCmdStr(prntcmd, '?', numstr);	
          SetCmdStr(prlncmd, '%', numstr);	

        END;


      SetCmdStr(cmdstr, '!', PFleft[ORD(PageForm)]);	
      SetCmdStr(cmdstr, '%', PFtop [ORD(PageForm)]);	

      SetCmdStr( colcmd, '!', PFcol [ORD(PageForm)]);	
      SetCmdStr( colcmd, '%', PFtop [ORD(PageForm)]);	

      SetCmdStr(linecmd, '%', PFline[ORD(PageForm)]);	

      SetCmdStr(pagecmd, '!', PFleft[ORD(PageForm)]);	
      SetCmdStr(pagecmd, '%', PFtop [ORD(PageForm)]);	

      IF PageForm = 7 THEN Insert(rulecmd, pagecmd, 1);


      PutString(cmdstr);

    END	{hp or ps}


  ELSE  {Epson}

    BEGIN
      linecnt := MaxLnHt DIV 4;

      CopyLST(EpsPage, pagecmd);
      CopyLST(EpsLine, linecmd);

      PutString(EpsInit);

    END;



  WHILE buf^[LineStart[Line]] <> EOF DO	    (* print a text line *)
  BEGIN

    FOR linepart := 0 TO linecnt DO
    BEGIN

      IF hp THEN

        BEGIN
          GrafPrep(LineBase[Line] + linepart, ADR(gstr));
     
          HPCompress(gstr, bytes);
          IF Encode(numstr, bytes:1) THEN;

          CopyLST(prntcmd, cmdstr);
          SetCmdStr(cmdstr, '#', numstr);	

          FOR i := 1 TO 2 DO
          BEGIN
            PutString(cmdstr);
      	    Eval(PUTUQQ(f, WRD(bytes), ADS(altbuf^)));
	  END;

  	END
      

      ELSE IF ps THEN

        BEGIN
          GrafPrep(LineBase[Line] + linepart, ADR(gstr));
     
          PSCompress(gstr, bytes);
          IF Encode(numstr, (bytes * 4):1) THEN;

          CopyLST(prntcmd, cmdstr);
          SetCmdStr(cmdstr, '#', numstr);	

          IF bytes > 0 THEN
          BEGIN
            PutString(cmdstr);
            Eval(PUTUQQ(f, WRD(bytes), ADS(altbuf^)));
            PutString(PSpend);
	  END;

          PutString(prlncmd);
	  
  	END


      ELSE {Epson}

        BEGIN
          EpsonPrep(LineBase[Line] + 4 * linepart, ADR(altbuf^), nonblank);

          IF nonblank THEN
          BEGIN
            PutString(EpsPrnt);
            Eval(PUTUQQ (f, 640, ADS(altbuf^)));
          END;

          PutString(EpsRet);

        END;


      IF KeyAvail THEN
        BEGIN
          GetChar(-1, AllKeys, ch, junk);

          IF ch = CHR(esc) THEN
            BEGIN PrintFinish(FALSE); CmdExit END;

        END;


    END; (* for *)


  
    IF pgline = ORD(PageLen) THEN	    (* form feed at page bottom *)	 

      IF (hp OR ps) AND THEN (colcnt < maxcol) THEN
	BEGIN
          pgline := 1;
          colcnt := colcnt + 1;

	  PutString(colcmd);

	END

      ELSE
        BEGIN
          pgline := 1;
          colcnt := 1;

	  WritePage;

        END

    ELSE    				    (* line feed between SW lines *)

      BEGIN
        pgline := pgline + 1;

        PutString(linecmd);

      END;



    IF Line = MaxLn THEN
      BEGIN
	QCursDraw;
	CScrollup(FALSE);
	QErase;
      END

    ELSE Line := Line + 1;

    
    CheckIO;

  END; (* while *)


  IF pgline <> 1 THEN WritePage;            (* form feed after last page *)

  PrintFinish(TRUE);


END;  (* Print *)






    (* -------------------  File command  --------------------- *)



PROCEDURE BuildExName (VAR xname: LSTRING;
		       CONST name: fName; pgnum: INTEGER);

VAR pstr: LSTRING(3);

BEGIN
  pstr.Len := 0;

  IF pgnum > 1 THEN
    IF Encode(pstr, pgnum:1) THEN {nothing};

  CopyLST(name, xname);

  IF (xname.Len + pstr.Len) > 8 THEN
    Delete(xname, ORD(xname.Len - pstr.Len + 1), ORD(pstr.Len));

  Concat(xname, pstr);
  Concat(xname, GrafExt);

END;  (* BuildExName *)



PROCEDURE OpenExFile (VAR f: blkFile; CONST name: LSTRING);

VAR hdr: GrafHeader;

BEGIN
  OpenFile(f, name, TRUE);

  Eval(PUTUQQ(f, Sizeof(hdr), ADS(hdr)));   (* reserve space for header *)
  IOCheck(name, FALSE, f);

END;  (* OpenExFile *)



PROCEDURE CloseExFile (VAR f: blkFile; CONST name: LSTRING;
		       pgnum, pglines: INTEGER);

VAR hdr: GrafHeader;

BEGIN
  FillC(ADR(hdr), Sizeof(hdr), CHR(0));

  WITH hdr DO
  BEGIN
    Manuf   := 10;
    Hard    := 3;   (* no palette info *)
    Encod   := 1;
    BitPx   := 1;
    X2      := 639;
    Y2      := WRD(pglines * 57 - 1);
    Hres    := 640;
    Vres    := 200;
    NPlanes := 1;
    Bplin   := 80;

  END;

  Rewrite(f);

  Eval(PUTUQQ(f, Sizeof(hdr), ADS(hdr)));   (* write header *)
  IOCheck(name, FALSE, f);

  CloseFile(f, name, (pglines > 0) OR (pgnum = 1));

END;  (* CloseExFile *)



PROCEDURE PCXCompress (CONST gdata: GrafData; VAR bufbytes: INTEGER);

VAR inx, binx, cnt, max: INTEGER;

BEGIN
  inx  := 1;
  binx := 0;

  WHILE inx <= Upper(gdata) DO
  BEGIN
    max := Upper(gdata) - inx + 1;
    IF max > 63 THEN max := 63;

    cnt := ScanNE(max, gdata[inx], gdata, inx);

    IF (cnt > 1) OR (ORD(gdata[inx]) DIV 64 = 3) THEN
      BEGIN
     	altbuf^[binx]   := WRD(cnt) OR #C0;
     	altbuf^[binx+1] := WRD(gdata[inx]);
        binx := binx + 2;
      END

    ELSE
      BEGIN
     	altbuf^[binx] := WRD(gdata[inx]);
        binx := binx + 1;
      END;

    inx := inx + cnt;

  END;  (* while *)


  bufbytes := binx;

END;  (* PCXCompress *)



PROCEDURE ExportFile (CONST name: fName);

VAR pgline,   pgnum: INTEGER;
    linepart, bytes: INTEGER;
    pglen,    junk:  INTEGER;
    s:     cString;
    f:	   blkFile;
    gdata: GrafData;
    ch:    CHAR;

BEGIN
  LoadFile(f, name);
  ShowFile(93);

  Line	 := 0;
  pgline := 1;
  pgnum  := 1;

  IF PrnType = PrHP THEN pglen := PFnlin[1]
	 	    ELSE pglen := EpsNlin;	

  BuildExName(s, name, pgnum);
  OpenExFile(f, s);
  

  WHILE buf^[LineStart[Line]] <> EOF DO	    (* export text line *)
  BEGIN
    
    FOR linepart := 0 TO (MaxLnHt + 1) DO
    BEGIN
      GrafPrep(LineBase[Line] + linepart, ADR(gdata));

      PCXCompress(gdata, bytes);

      Eval(PUTUQQ(f, WRD(bytes), ADS(altbuf^)));

      IF KeyAvail THEN
        BEGIN
          GetChar(-1, AllKeys, ch, junk);

          IF ch = CHR(esc) THEN
            BEGIN Close(f); CmdExit END;

        END;

    END;
  

    IF pgline = pglen THEN 
      BEGIN
        CloseExFile(f, s, pgnum, pgline);

	pgnum := pgnum + 1;

        BuildExName(s, name, pgnum);
        OpenExFile(f, s);

	pgline := 1;

      END

    ELSE pgline := pgline + 1;
    

    IF Line = MaxLn THEN
      BEGIN
	QCursDraw;
	CScrollup(FALSE);
	QErase;
      END

    ELSE Line := Line + 1;


  END; (* while *)


  CloseExFile(f, s, pgnum, pgline - 1);


END;  (* ExportFile *)



PROCEDURE Filer;

CONST FileType = FileExt * CHR(0);
   
VAR oldname, newname: LSTRING(13);
    name: fName;
    f:    blkFile;
    ch:   CHAR;
    cmd:  cmdVal;
    s:    cString;

BEGIN
  WriteCmdName(Msg[74].msg^);
  
  Prompt(9, [], Msg[9].msg^, ch, cmd);
  
  IF ch = CHR(esc) THEN CmdExit;


  
  IF cmd = C_Fdelete THEN   			   (* Delete prompt   *)
    
    BEGIN
      GetString(Msg[50].msg^, NameKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit;

      BuildMsg(s, Msg[51].msg^, name);
      Prompt(51, [], s, ch, cmd);

      IF cmd <> C_Yes THEN CmdExit;
      
      DeleteFile (name);
      
      UpdateFileList := TRUE;
      
    END


    
  ELSE IF cmd = C_Frename THEN    		   (* Rename prompt   *)
    
    BEGIN
      GetString(Msg[52].msg^, NameKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit;

      OpenFile (f, name, FALSE);
      CloseFile(f, name, TRUE);
      CopyLST(name, oldname);

      GetString(Msg[53].msg^, NameKeys, 8, TRUE, name);

      IF (name = NULL) OR (name = oldname) THEN CmdExit
					   ELSE CheckFile(name);
      CopyLST(name, newname);
      
      Concat(oldname, FileType);
      Concat(newname, FileType);
      
      RenameFile(oldname, newname);
      
      UpdateFileList := TRUE;
      
    END



  ELSE  (* cmd = C_Fexport *)    		   (* Export prompt   *)
  
    BEGIN
      GetString(Msg[92].msg^, NameKeys, 8, TRUE, name);
      IF name = NULL THEN CmdExit;

      ExportFile(name);

    END;


END;   (* Filer *)






    (* ------------------  Setup command  ---------------------- *)



PROCEDURE DumpMsgs;

VAR tmp: cmdPtr;
    i:   INTEGER;

BEGIN
  FOR i := 1 TO MaxMsg DO
  WITH Msg[i] DO
  BEGIN
    DISMQQ(msg);

    WHILE cmd <> NIL DO
    BEGIN
      tmp := cmd^.nxt;
      Dispose(cmd);
      cmd := tmp;
    END;
  END;

END;  (* DumpMsgs *)



FUNCTION CheckDict (CONST name: LSTRING): BOOLEAN;

VAR f:     blkFile;
    s:     cString;
    dcode: BYTE;
    dct, din: BOOLEAN;

BEGIN
  BuildDTName(s, DicFile, name);        (* check new dict file  *)
  dct := SWOpen(f, s);
  IF dct THEN Close(f);

  BuildDTName(s, DinFile, name);        (* check new dict index *)
  din := SWOpen(f, s);
  IF din THEN
  BEGIN
    dcode := f^;
    Close(f);
  END;


  IF dct <> din THEN		        (* one file missing     *)
  BEGIN
    Error(Msg[87].msg^);
    CmdExit;
  END;


  IF din AND (dcode <> SWdict) THEN     (* old dictionary       *)
  BEGIN
    Error(Msg[102].msg^);
    CmdExit;
  END;


  CheckDict := dct;


END;  (* CheckDict *)



PROCEDURE CreateDict (CONST name: LSTRING);

VAR f:    blkFile;
    s:    cString;
    free: WORD;

BEGIN
  BuildDTName(s, DinFile, name);	    (* dict index file   *)

  OpenFile(f, s, TRUE);


  free := DRecEOL;

  FillC(ADR(altbuf^), 512, CHR(0));

  altbuf^[0] := SWDict;
  MoveL(ADR(free), ADR(altbuf^[1]), 2);

  { nrecs & dict volume counts = 0 }

  MoveSL(ADS(Ccurmap^), ADS(altbuf^[17]), Sizeof(Ccurmap^));


  Eval(PUTUQQ(f, 512, ADS(altbuf^)));
  IOCheck(Msg[91].msg^, FALSE, f);

  CloseFile(f, Msg[91].msg^, TRUE);



  BuildDTName(s, DicFile, name);	    (* (empty) dict file *)	

  OpenFile(f, s, TRUE);

  CloseFile(f, Msg[91].msg^, TRUE);


END;  (* CreateDict *)



PROCEDURE MarkSelect (VAR str: LSTRING; val: INTEGER);

VAR inx: INTEGER;

BEGIN
  inx := 2;
	
  WHILE val > 0 DO
  BEGIN
    val := val - 1;

    inx := inx + ScanEQ(ORD(str.Len) - inx, ' ', str, inx);
    inx := inx + ScanNE(ORD(str.Len) - inx, ' ', str, inx);

  END;
  
  str[inx - 1] := '*';

END;  (* MarkSelect *)



PROCEDURE Setup;

VAR ch:	  CHAR;
    cmd:  cmdVal;
    code: WORD;
    dict: BOOLEAN;
    f:    FILE OF SymHeader;
    s, name:  cString;


  PROCEDURE GetConfig;
  BEGIN
    IF NOT SWOpen(f,SymFile) THEN
      BEGIN Error(Msg[32].msg^); CmdExit END;
    Seek(f, 1);
    Get(f);
    Seek(f, 1);
  END;


  PROCEDURE PutConfig;
  BEGIN
    Put(f);
    IOCheck(SymFile, FALSE, f);
    CloseFile(f, SymFile, TRUE);
  END;


BEGIN  (* Setup *)

  WriteCmdName(Msg[76].msg^);
  
  Prompt(10, [], Msg[10].msg^, ch, cmd);
  
  IF ch = CHR(esc) THEN CmdExit;



  IF cmd = C_Sprnting THEN		    (* Printing command line *)

    BEGIN
      Prompt(11, [], Msg[11].msg^, ch, cmd);
  
      IF ch = CHR(esc) THEN CmdExit;



      IF cmd = C_Sprinter THEN     	    (* Printer prompt        *)
    
        BEGIN
          CopyLST(Msg[96].msg^, s);

	  MarkSelect(s, ORD(PrnType));

          Prompt(96, [CHR(ret)], s, ch, cmd);
          IF (ch = CHR(esc)) OR (ch = CHR(ret)) THEN CmdExit;
  
          PrnType := WRD(cmd);
       

          IF PrnType = PrEpson THEN   	    (* set defaults *) 
            PageLen := EpsNlin

	  ELSE {HP or Postscript}
	    BEGIN
              PageForm := 1;
              PageLen  := WRD(PFnlin[ORD(PageForm)]);
	    END;

  
          GetConfig;
          f^.prType := PrnType;
          f^.PgForm := PageForm;
          f^.PgLen  := PageLen;
          PutConfig;

	END



      ELSE IF cmd = C_Sintface THEN         (* Interface prompt      *)
    
        BEGIN
          CopyLST(Msg[61].msg^, s);

	  MarkSelect(s, ORD(PrnPort));

          Prompt(61, [CHR(ret)], s, ch, cmd);
          IF (ch = CHR(esc)) OR (ch = CHR(ret)) THEN CmdExit;
  
          PrnPort := WRD(cmd);

  
          GetConfig;
          f^.prPort := PrnPort;
          PutConfig;

        END



      ELSE IF cmd = C_Sformat THEN          (* Format prompt         *)
    
        BEGIN
          name.Len := 0;

          IF PrnType = PrEpson THEN
            BEGIN Error(Msg[99].msg^); CmdExit END;

          IF Encode(name, PageForm:1) THEN;
          BuildMsg(s, Msg[97].msg^, name);

          GetString(s, NumKeys, 1, TRUE, name);
          IF name = NULL THEN CmdExit;

          IF Decode(name, code) THEN;

          IF (code < 1) OR (code > MaxPF) THEN
            BEGIN Error(Msg[98].msg^); CmdExit END;

          PageForm := code;
          PageLen  := WRD(PFnlin[ORD(PageForm)]);


          GetConfig;
          f^.PgForm := PageForm;
          f^.PgLen  := PageLen;
          PutConfig;

        END



      ELSE IF cmd = C_Snumbers THEN         (* Numbers prompt        *)
    
        BEGIN
          Prompt(100, [], Msg[100].msg^, ch, cmd);
  	  IF ch = CHR(esc) THEN CmdExit;


          IF cmd = C_No THEN 
            PageNum := NoPgNm
    
	  ELSE
	    BEGIN
	      IF PageNum = NoPgNm THEN PageNum := 1;

              IF Encode(name, PageNum:1) THEN;
              BuildMsg(s, Msg[59].msg^, name);

              GetString(s, NumKeys, 4, TRUE, name);
              IF name = NULL THEN CmdExit;

              IF Decode(name, PageNum) THEN;
	    END;


          GetConfig;
          f^.PgNum := PageNum;
          PutConfig;

        END



      ELSE  (* cmd = C_Slines *)	    (* Lines prompt  	     *)
    
	BEGIN
          name.Len := 0;

          IF Encode(name, PageLen:1) THEN;
          BuildMsg(s, Msg[84].msg^, name);

          GetString(s, NumKeys, 2, TRUE, name);
          IF name = NULL THEN CmdExit;

          IF Decode(name, PageLen) THEN;


          GetConfig;
          f^.PgLen := PageLen;
          PutConfig;

        END


    END	 (* cmd = C_Sprnting *)


  
  ELSE IF cmd = C_Scountry THEN             (* Country prompt        *)
    
    BEGIN
      CCToStr(name, Country);
      BuildMsg(s, Msg[77].msg^, name);

      GetString(s, NumKeys, 3, TRUE, name);
      IF name = NULL THEN CmdExit;
      StrToCC(code, name);
      

      BuildCCName(s, KeyFile, code);        (* check new key file  *)

      IF NOT SWOpen(f, s) THEN
        BEGIN
          Error(Msg[30].msg^);
          CmdExit;
        END
      ELSE Close(f);


      BuildCCName(s, MsgFile, code);	    (* check new msg file  *)

      IF NOT SWOpen(f, s) THEN
        BEGIN
          Error(Msg[57].msg^);
          CmdExit;
        END
      ELSE Close(f);


      dict := CheckDict(name);		    (* check new dict file *)



      GetConfig;			    (* save change *)	
      f^.CCode := code;
      Country  := code;
      
      IF dict THEN
      BEGIN
        f^.DicTit := name;
        DictTitle := name;
      END;

      PutConfig;



      InitKeyFile;                          (* reinit key & msg files *)

      DumpMsgs; 
      InitMsgFile;


      IF dict THEN InitDict;                (* reinit dict files *)


    END



  ELSE  (* cmd = C_Sdiction *)		    (* Dictionary prompt     *)

    BEGIN
      BuildMsg(s, Msg[94].msg^, DictTitle);

      GetString(s, NameKeys, 6, TRUE, name);
      IF name = NULL THEN CmdExit;
      

      IF NOT CheckDict(name) THEN	    (* create new dict? *)
      BEGIN
        BuildMsg(s, Msg[95].msg^, name);

        DrawAlertBox(1);

        Prompt(95, [], s, ch, cmd);

        DrawAlertBox(0);

        IF cmd = C_Yes THEN CreateDict(name)
        ELSE CmdExit;

      END;


      GetConfig;			    (* save change *)	
      f^.DicTit := name;
      DictTitle := name;
      PutConfig;


      InitDict;                             (* reinit dict files *)

    END;


END;  (* Setup *)






     (* -------------  File screen command interp  -------------- *)



PROCEDURE FCommand;

VAR ch:	   CHAR;
    code:  INTEGER;
    state: ShiftState;
    cmd:   cmdVal;
    r:     InxRec;

BEGIN
  GetKey(code, state);
  
  r := Keyboard^[Alpha, code, Shft];
  IF NoCharVal(r) THEN RETURN;
  
  ch  := CHR(r.sym.codeval);
  cmd := SearchCmd(3, ch);

  IF ((state = Alt) OR (State = ShftAlt)) AND (cmd >= 0) THEN
    
    BEGIN
      CASE cmd OF
	C_Edit:  Edit(FALSE);
	C_New:   Edit(TRUE);
	C_Print: Print;
	C_File:  Filer;
	C_Dir:   Directory;
	C_Setup: Setup;
	C_Quit2: FQuit;
        OTHERWISE;
      END;
      
      UpdateCmdList := TRUE;
      
    END
  
  ELSE	 (* sym key *)
      
    IF (ch = CHR(down)) OR (ch = CHR(pgdn)) THEN   (* file list down *)

      IF maxName > (curName + 14) THEN
	BEGIN
	  curName := curName + 15;
	  WriteFileNames(curName);
	END
      ELSE {nothing}

    ELSE IF (ch = CHR(up)) OR (ch = CHR(pgup)) THEN  (* file list up *)

      IF curName > 0 THEN
	BEGIN
	  curName := curName - 15;
	  WriteFileNames(curName);
	END
      ELSE {nothing};
	
END;   (* FCommand *)


  


  
  (* --------------  Command exit procedure  --------------- *)


  
PROCEDURE CmdExit; 

BEGIN
  GOTO EndOfCmd;
END;






  (* ------------------ Command main body  ------------------ *)



BEGIN (* Command *)

  IF Window THEN WCommand
  ELSE IF FScreen THEN FCommand
  ELSE ECommand;

EndOfCmd:

END;



  


  (* ------------------  Main program	--------------------- *)



BEGIN  (* SignWriter *)

  KeyInit;
  CriticalInit;
  
  SWInit;


  FGetCommand;

  
  TextMode;
  KeyTerm;

END.
